/**
* Dropdown extension definition.
* 
* I wrapped the definition of this extension inside a function called axDropDownExtension.
* We don't have to put our extension definition inside a function, however it's a good idea
* to do that because our code declares variables (outside of a function variables are global - so we make
* them local by putting our code inside a function).
*
**/
function DEFINE_EXTENSION_axDropDown()
{
    /**
    * Call the function that defines the drop down item object
    **/
    defineDropdownItemClass();

    /**
    * Define enumeration for the dropdown data source type
    **/
    dataSourceType =
    [
        { value: "xml", text: TEXT.dev("axDropDown.XMLfile") },
        { value: "dynamic", text: TEXT.dev("axDropDown.DynamicTable") },
        { value: "static", text: TEXT.dev("axDropDown.StaticTable") },        
        { value: "fixedValues", text: TEXT.dev("axDropDown.FixedValues") }
    ];

    /**
    * Drop down extension definition
    **/
        var definition_axDropDown =
    {
        /**
        * Programmatic class name of this extension. It's not visible to end user 
        **/
        name: "axDropDown",
        group: "vis",
        /**
        * Type of object this extension is applicable. Can be either "element" or "screen"
        * If this extension applies to both element & screen, we put both in an array like so:
        * type: ["element", "screen"]
        **/
        type: "element",

        /**
        * End-user visible name of this extension
        **/

        display: TEXT.dev("axDropdown.display"),

        /* Name of any temporary fixed values table created */
        
        _tempFixedValuesTable: "",

        /**
        * End-user visible name of this extension
        **/
        properties:
        {
            dropDownStyle: { type: "Style" },
            /**
            * "dataSourceType" property indicates where the dropdown items come from
            * The possible values are what we define earlier in dataSourceType
            **/
            dataSourceType: { type: "String", items: dataSourceType, defaultStatic: "fixedValues", description: TEXT.dev("axDropDown.dataSourceType.description") },

            /** 
            * Table name when the dataSourceType is static or dynamic
            *
            **/
            tableName: { type: "String", description: TEXT.dev("axDropDown.tableName.description") },

            /**
            * "onFillDropDown" event
            * We raised this event to fill the drop down with data
            **/
            onFillDropDown: { type: "Event", defaultDynamic: "", description: TEXT.dev("axDropDown.onFillDropDown.description") },

            /**
            * "xmlFileName" name of the xml file to load the table from.
            **/
            xmlFileName: { type: "String", description: TEXT.dev("axDropDown.xmlFileName.description") },

            /**
            * "fileLocation" where to locate the file (xml or other(not currently implenented)).
            **/
            xmlFileLocation: { type: "String", defaultStatic: "*DEFAULT", description: TEXT.dev("axDropDown.xmlFileLocation.description") },

            /**
            * "fixedValues" values typed by the user when dataSourceType = "fixedValues".
            * We set the "isList" attribute to indicate a list.
            **/
            fixedValues: { type: "axDropDownitem", isList: true },

            /* allows users to add non table entries to a drop down - typically a blank entry  but can be anything, one or multiple entries. In the event the users must */
            /* call add*/
            additionalEntries: { type: "Event", description: TEXT.dev("axDropDown.additionalEntries.description") },

            /**
            * The drop down selected value
            **/
            onSelectValue: { type: "Event", defaultDynamic: "", description: TEXT.dev("axDropDown.onSelectValue.description") },

            /**
            * "onSelectedValueChanged" event
            * We raised this event when the user change the selection in the drop down.
            **/
            onSelectedValueChanged: { type: "Event", defaultDynamic: "FIELD.setValue(ROW.value)", description: TEXT.dev("axDropDown.onSelectedValueChanged.description") },

            /* The name of the query specified in the dynamic table */
            sqlQueryName: { type: "String", description: TEXT.dev("axDropDown.sqlQueryName.description") },

            /* Variables used in the query specified in the dynamic table */
            sqlVariables: { type: "Event", defaultDynamic: "ENV.SQL.SQLVariableLibrary = USERENV.dftSQLDataLibrary", description: TEXT.dev("axDropDown.sqlVariables.description") },

            /* Dynamic table to compare the key with the previous request and not reload if they are the same */
            keepLastKey: {
                type: "Boolean",
                defaultStatic: false,
                description: TEXT.dev("axDropDown.keepLastKey.description")
            },

            onFocus : { 
                type: "Event", 
                description: "Fires when this field receives the focus", 
                defaultDynamic: "" 
            },
            onBlur : { 
                type: "Event", 
                description: "Fires when this field loses focus", 
                defaultDynamic: "" 
            }

        },

        /**
        * This is where we put the code that initialise things (variable, HTML elements) that will remain 
        * the same for the life of the extension.
        *
        * The "field" parameter is the abstraction of the field element and we reference this object
        * when we need to get or set the value of the field
        *
        * The "fieldDiv " parameter is the HTML element that serves as a container for the field.
        * Since it serves as a container for the field and not only for this extension, it might
        * already contain some HTML nodes already when this extension is executed.
        *
        * The reason why each extension does not get their own container is because some extensions
        * do not have their own visualisation (e.g. tooltip) and they only change the behaviour of
        * the associated field.
        *
        * For extensions like dropdown etc that have their own visualisation, we can assume
        * that the fieldDiv will be empty and we can ignore any elements there and create the 
        * elements we need. For the time being though when we are testing our extensions we need to
        * uncheck the "Default Visualisation" extension in order for our extension's visualisation to
        * appear (we might change this later to uncheck automatically not sure yet).
        **/
        init: function(field, fieldDiv)
        {
            this._field = field;
            this._rows = [];
            
            var el = null;
            // use "isInput" property to test if the field is an input or output field
            if (field.isInput)
            {
                // create a SELECT element for input field
                el = document.createElement("select");
                $(el).bind("change", this, this._onChange);
                $(el).bind("focus", this, this._onFocus);
                $(el).bind("blur", this, this._onBlur);
                this.applyTabIndex( el );
            }
            else
            {
                if (AXES.isTS2Engine) {
                    el = document.createElement("span");
                    el.className = "ax-terminal-textField";
                    if (field instanceof AXES.Terminal.Field.TextField) {
                        el.className += " ax_x" + field.attribute.substring(1).toUpperCase() + "o";
                    }
                } else {
                    el = this.createTextElement();
                }
            }

            /* Listen for changes in the field value */
            field.addListener("ValueChanged", this._onFieldValueChanged, this);
            if (AXES.isTS2Engine) this._field.addListener("AddedToDOM", this._handleAddedToDOM, this);
            
            // save the reference to the element so we can use it later
            this._element = el;

            // append to the container
            fieldDiv.appendChild(el);
        },

        /**
        * We should put things that are impacted by change in properties here.
        *
        * The list of items is coming from the "fixedValues" property so the population
        * of the dropdown will happen in this method
        **/
        render: function(field, fieldDiv)
        {
            // el is the SELECT html element
            var el = this._element;

            var baseStyle = false;
            //if (!AXES.isTS2Engine) 
            //{
                var size = field.getSize();
                // set the size of the element to occupy the whole fieldDiv container
                baseStyle = {
                    width: size.width,
                    height: size.height
                };
            //}
            this.applyStyle(this.getPropertyValue("dropDownStyle"), el, baseStyle);

            var sSourceType = this.getPropertyValue("dataSourceType");
            var sTableName = "";
            var oTable = null;

            if (!this.isDesignMode())
            {
                switch (sSourceType)
                {
                    case "fixedValues":
                        var item;
                        var items = this.getPropertyValue("fixedValues");

                        AXES.TableManager.deleteTable(this._tempFixedValuesTable); /* Drop anything hanging around */  
                        oTable = AXES.TableManager.getTable(); /* Force creation of a brand new table */
                        this._tempFixedValuesTable = oTable.name;
                        sTableName = oTable.name;

                        for (var i in items)
                        {
                           item = items[i];
                           oTable.insertChild({ value: item.value, text: item.text });
                        }
                                                
                        break;

                    case "static":
                        sTableName = this.getPropertyValue("tableName");
                        oTable = AXES.TableManager.getTable(sTableName);
                        AXES.TableManager.loadStaticTables(USERENV.staticTablesFile, {}, false);
                        break;

                    case "dynamic":
                        var env = { SQL: {} };
                        this.raiseEvent("sqlVariables", env);

                                                var bKeepLastKey = false ;
                                                bKeepLastKey = this.getPropertyValue("keepLastKey"); 
                                      
                        sTableName = this.getPropertyValue("sqlQueryName");
                        AXES.TableManager.loadDynamicTable(sTableName, USERENV.dynamicTablesFile, env.SQL, bKeepLastKey);
                        oTable = AXES.TableManager.getTable(sTableName);
                        break;

                    case "xml":
                        var sPath = this.getPropertyValue("xmlFileLocation");
                        if (sPath == "*DEFAULT")
                        {
                            sPath = AXES.TableManager.getDefaultPath();
                        }

                        var sFullFileName = sPath + "/" + this.getPropertyValue("xmlFileName");

                        sTableName = sFullFileName;
                        oTable = AXES.TableManager.getTable(sTableName);

                        AXES.TableManager.loadFromXML(sFullFileName, oTable, false);
                        break;
                }

                var oRow, sDropDownText, env;
                var bSelected = false;
                var sFldValue = field.getValue().trim();
                /* store away the table name for the addTableROW method to use */
                this.TableName = sTableName;
                this._rows.length = 0;
                
                if (field.isInput)
                {
                    // clear all existing item elements first
                    el.options.length = 0;
                    var oOption, optParent = el;

                    if (!(oTable.additionalEntriesAdded))
                    {
                        this.raiseEvent("additionalEntries");
                        oTable.additionalEntriesAdded = true; 
                    }
                    
                    env = { ROW: null, FIELD: null };
                    
                    var iChildCount = oTable.childCount();
                    
                    this._rows.length = iChildCount;
                    
                    for (var i = 0; i < iChildCount; i++)
                    {
                        oRow = oTable.child(i);
                        env.ROW = oRow;
                        env.FIELD = field;

                        sDropDownText = this.getPropertyValue("onFillDropDown", env);
                        if (sDropDownText === undefined) sDropDownText = oRow.text;
                        
                        if (sDropDownText != null)
                        {
                            oOption = document.createElement("option");
                            oOption.appendChild(document.createTextNode(sDropDownText)); 
                            
                            // The table is a centralised resource that may get changed by other scripts (shouldn't, but might)
                            // So, to make sure we can reliably retrieve them when required, we keep our own array of oRow objects.
                            oOption.value = i;
                            this._rows[i] = oRow;
                            
                            if (!bSelected) /* No point on keep comparing when a value has already been selected */
                            {
                                bSelected = this.getPropertyValue("onSelectValue", env);
                                if (bSelected == undefined) bSelected = oRow.value;
                                
                                /* The type returned by onSelectValue will depend on whether the user has specified the property as a simple expression or a more complex one */
                                /* For example, the default value of onSelectValue is ROW.value. This is interpreted as the user wanting to select the drop down entry when the */
                                /* the value of the field associated with the drop down is equal to the current ROW.value. However, the user may have specified something like: */
                                /* ROW.value == (something or other). Hence, we inspect the type of the value returned by the property. If its a boolean it means the user specified */
                                /* a complex expression that yields a true/false. Otherwise we compare the property with the field's value */
                                if (bSelected !== undefined && typeof (bSelected) != "boolean")
                                {
                                    bSelected = (bSelected.toString() == sFldValue);
                                }
                                if (bSelected) oOption.selected = true;
                            }
                            optParent.appendChild(oOption);
                            
                        }
                        else
                        {
                            var groupLabel = oRow.optgroup;

                            if (groupLabel != null)
                            {
                                var oOptGroup = document.createElement("optgroup");
                                oOptGroup.label = groupLabel;
                                el.appendChild(oOptGroup);
                                optParent = oOptGroup;
                            }
                        }
                    } /* end for */
                }  /* end if input */
                else /* it's an output field */
                {
                    for (var i = 0; i < oTable.childCount(); i++)
                    {
                        oRow = oTable.child(i);
                        this._rows.push(oRow);
                        env = { ROW: oRow, FIELD: field };

                        bSelected = this.getPropertyValue("onSelectValue", env);
                        if (bSelected == undefined) bSelected = oRow.value;
                        
                        if (bSelected !== undefined && typeof (bSelected) != "boolean")
                        {
                            bSelected = (bSelected.toString() == sFldValue);
                        }

                        if (bSelected)
                        {
                            var sDropDownText = this.getPropertyValue("onFillDropDown", env);
                            if (sDropDownText === undefined) sDropDownText = oRow.text;
                            $(el).text(sDropDownText);
                            break;
                        }
                    } /* end for */
                } /* end if input */
            } /* end if design mode */
        },

        addTableROW: function(oRow, bAtBegin)
        {
            var oTable = AXES.TableManager.getTable(this.TableName);
            oTable.insertChild(oRow, bAtBegin);
        },

        /**
        * Do opposite of init
        **/
        destroy: function(field, fieldDiv)
        {
            // delete any temporary table created - delete will ignore null/empty names
              AXES.TableManager.deleteTable(this._tempFixedValuesTable);  
            this._tempFixedValuesTable = "";    
            // remove the element referenced by this._element from the DOM
            var el = this._element;
            if (el != null)
            {
                if (el.options) el.options.length = 0;
                $(el).remove();
                delete this._element;
            }

            this._field = null;
            this._rows.length = 0;
            
            /* Remove any listeners */
            field.removeListener("ValueChanged", this._onFieldValueChanged);
            field.removeListener("AddedToDOM", this._handleAddedToDOM);
        },

        /**
        * Handles the onchange event of the dropdown
        **/
        _onChange: function(e)
        {
            var self = e.data;
            if (!self.isDesignMode()) 
            {
                self._noFieldValueChangedEvent = true;
                var option = self._element.options;
                var index = option.selectedIndex;
                if ( 0 > index ) return;
                
                self.raiseEvent( "onSelectedValueChanged", { ROW: self._rows[ option[ index ].value ], FIELD: self._field } );
            }
        },

        /* Handles the change in field value */
        _onFieldValueChanged: function(event)
        {
            var self = this;

            if (!(self._field.isInput)) return;

            if (!self._noFieldValueChangedEvent)
            {
                var env, bSelected, newSI = -1;
                var sFldValue = TRIM(self._field.getValue());

                for (var i=0,oRow; oRow=this._rows[i]; i++)
                {
                    env = { ROW: oRow, FIELD: self._field };
                    bSelected = self.getPropertyValue("onSelectValue", env);
                    if (bSelected == undefined) bSelected = oRow.value;
                    if (bSelected !== undefined && typeof (bSelected) != "boolean")
                    {
                        bSelected = (bSelected.toString() == sFldValue);
                    }
                    if (bSelected) 
                    {
                        newSI = i;
                        break;
                    }
                }
                self._element.selectedIndex = i;
            }
            else
            {
                delete self._noFieldValueChangedEvent;
            }
        },
        _handleAddedToDOM: function(event)
        {
            // To improve rendering performance, The TS2 engine doesn't add a field to the DOM until all
            // the extensions have been rendered. Elements do not have dimensions until they are added
            // to the DOM so we use this event to do anything that requires knowledge of the size of things.
            // For best rendering performance, avoid using this unless absolutely necessary.
            var size = this._field.getSize();
            var sizeCSS = {
                width: size.width + "px",
                height: size.height + "px"
            };
            this.applyStyle2(sizeCSS, this._element, false);
        },

        _onFocus: function(e)
        {
            var self = e.data;
            if (!self.isDesignMode()) 
            {
                var option = self._element.options;
                var index = option.selectedIndex;
                if ( 0 > index ) return;
                self.raiseEvent( "onFocus", { ROW: self._rows[ option[ index ].value ], FIELD: self._field } );                
            }
        },
        
        _onBlur: function(e)
        {
            var self = e.data;
            if (!self.isDesignMode()) 
            {
                var option = self._element.options;
                var index = option.selectedIndex;
                if ( 0 > index ) return;
                self.raiseEvent( "onBlur", { ROW: self._rows[ option[ index ].value ], FIELD: self._field } );                
            }
        }
    };

    /**
    * Register the extension with AXES
    **/
    AXES.Extensions.add(definition_axDropDown);

    /**
    * The function belows defines the drop down item object.
    * Each drop down item has a "value" and "text" property.
    * The "value" property is what will be used as the actual 5250 field value and
    * the "text" property is what will be displayed on the drop down.
    **/
    function defineDropdownItemClass()
    {
        /** 
        * constructor function for the drop down item object
        **/
        var itemClass = function()
        {
            this.value = "";
            this.text = "";
        };

        /** 
        * "toString" function for the item object.
        * This will be used by the designer when it needs to display a simple representation
        * of the drop down item.
        **/
        itemClass.prototype.toString = function()
        {
            var value = this.getPropertyValue("value");
            var text = this.getPropertyValue("text");

            return value + " - " + text;
        };

        /** 
        * Properties of drop down item object.
        **/
        var itemClassDef =
        {
            value: { type: "String" },
            text: { type: "String" }
        };

        /** 
        * We need to register the constructor function and property list with AXES.
        * We want to name the item object class "axDropDownitem".
        **/
        AXES.Extensions.definePropertyType("axDropDownitem", { objectClass: itemClass, objectDefinition: itemClassDef });
    }
}

/**
* Make sure we call the function we just defined.
**/
DEFINE_EXTENSION_axDropDown();