
/**
 * RadioButton extension definition.
 * 
 * I wrapped the definition of this extension inside a function called axRadioButtonExtension.
 * We don't have to put our extension definition inside a function, however it's a good idea
 * to do that because our code declares variables (outside of a function variables are global - so we make
 * them local by putting our code inside a function).
 *
 **/
function DEFINE_EXTENSION_axRadioButton()
{
    /**
     * Call the function that defines the drop down item object
     **/
    defineRadioButtonItemClass();

    /**
     * Define enumeration for the dropdown data source type
     **/
    dataSourceType = [{
        value: "xml",
        text: TEXT.dev("axRadioButton.XMLfile")
    }, {
        value: "dynamic",
        text: TEXT.dev("axRadioButton.DynamicTable")
    }, {
        value: "static",
        text: TEXT.dev("axRadioButton.StaticTable")
    }, {
        value: "fixedValues",
        text: TEXT.dev("axRadioButton.FixedValues")
    }];

    /**
     * Define enumeration for the dropdown data source type
     **/
    orientation = [{
        value: "V",
        text: TEXT.dev("axRadioButton.Vertical")
    }, {
        value: "H",
        text: TEXT.dev("axRadioButton.Horizontal")
    }];

    /**
     * Drop down extension definition
     **/
    var definition_axRadioButton = {
        /**
         * Programmatic class name of this extension. It's not visible to end user
         **/
        name: "axRadioButton",
        group: "vis",
        /**
         * Type of object this extension is applicable. Can be either "element" or "screen"
         * If this extension applies to both element & screen, we put both in an array like so:
         * type: ["element", "screen"]
         **/
        type: "element",

        /**
         * End-user visible name of this extension
         **/
        display: TEXT.dev("axRadioButton.display"),

        /* The name of any temporary table created */

        _tempFixedValuesTable: "",

        /**
         * End-user visible name of this extension
         **/
        properties: {
            radioButtonStyle: {
                type: "Style"
            },
            captionStyle: {
                type: "Style"
            },

            /**
             * "dataSourceType" property indicates where the dropdown items come from
             * The possible values are what we define earlier in dataSourceType
             **/
            dataSourceType: {
                type: "String",
                items: dataSourceType,
                defaultStatic: "fixedValues",
                description: TEXT.dev("axRadioButton.dataSourceType.description")
            },

            orientation: {
                type: "String",
                items: orientation,
                defaultStatic: "V",
                description: TEXT.dev("axRadioButton.orientation.description")
            },
            /** 
             * Table name when the dataSourceType is static or dynamic
             **/
            tableName: {
                type: "String",
                description: TEXT.dev("axRadioButton.tableName.description")
            },

            /**
             * The drop down selected value
             **/
            onSelectRadio: {
                type: "Event",
                defaultDynamic: "ROW.value",
                description: TEXT.dev("axRadioButton.onSelectRadio.description")
            },

            /**
             * "onCreateRadioButton" event
             * We raised this event to fill the drop down with data
             **/
            onCreateRadioButton: {
                type: "Event",
                defaultDynamic: "ROW.text",
                description: TEXT.dev("axRadioButton.onCreateRadioButton.description")
            },

            /**
             * "xmlFileName" name of the xml file to load the table from.
             **/
            xmlFileName: {
                type: "String",
                description: TEXT.dev("axRadioButton.xmlFileName.description")
            },

            /**
             * "fileLocation" where to locate the file (xml or other(not currently implenented)).
             **/
            xmlFileLocation: {
                type: "String",
                defaultStatic: "*DEFAULT",
                description: TEXT.dev("axRadioButton.xmlFileLocation.description")
            },

            /**
             * "fixedValues" values typed by the user when dataSourceType = "fixedValues".
             * We set the "isList" attribute to indicate a list.
             **/
            fixedValues: {
                type: "axRadioButtonitem",
                isList: true
            },

            /* The name of the query specified in the dynamic table */
            sqlQueryName: {
                type: "String",
                description: TEXT.dev("axRadioButton.sqlQueryName.description")
            },

            /* Variables used in the query specified in the dynamic table */
            sqlVariables: {
                type: "Event",
                defaultDynamic: "ENV.SQL.SQLVariableLibrary = USERENV.dftSQLDataLibrary",
                description: TEXT.dev("axRadioButton.sqlVariables.description")
            },

            /**
             * "onRadioButtonClicked" event
             * We raised this event when the user change the selection in the drop down.
             **/
            onRadioButtonClicked: {
                type: "Event",
                defaultDynamic: "FIELD.setValue(ROW.value)",
                description: TEXT.dev("axRadioButton.onRadioButtonClicked.description")
            },

            /* Dynamic table to compare the key with the previous request and not reload if they are the same */
            keepLastKey: {
                type: "Boolean",
                defaultStatic: false,
                description: TEXT.dev("axRadioButton.keepLastKey.description")
            },

            onChange: { 
                type: "Event", 
                description: "Fires when this element changes state",
                defaultDynamic: ""
            },
            onFocus: {
                type: "Event", 
                description: "Fires when this element receives the focus", 
                defaultDynamic: "" 
            },
            onBlur: { 
                type: "Event", 
                description: "Fires when this element loses focus", 
                defaultDynamic: "" 
            }
        },

        _focusedChrome: false,
        _changeEventOwner: null,

        /**
         * This is where we put the code that initialise things (variable, HTML elements) that will remain
         * the same for the life of the extension.
         *
         * The "field" parameter is the abstraction of the field element and we reference this object
         * when we need to get or set the value of the field
         *
         * The "fieldDiv " parameter is the HTML element that serves as a container for the field.
         * Since it serves as a container for the field and not only for this extension, it might
         * already contain some HTML nodes already when this extension is executed.
         *
         * The reason why each extension does not get their own container is because some extensions
         * do not have their own visualisation (e.g. tooltip) and they only change the behaviour of
         * the associated field.
         *
         * For extensions like dropdown etc that have their own visualisation, we can assume
         * that the fieldDiv will be empty and we can ignore any elements there and create the
         * elements we need. For the time being though when we are testing our extensions we need to
         * uncheck the "Default Visualisation" extension in order for our extension's visualisation to
         * appear (we might change this later to uncheck automatically not sure yet).
         **/
        init: function( field, fieldDiv )
        {
            this._field = field;
            var oContainer = null;
            if ( field.isInput ) { 
                // create container for radio button
                oContainer = this._container = document.createElement("span");
                oContainer.style.overflow = "auto";
            } else {
                // create container output text
                if (AXES.isTS2Engine) {
                    oContainer = this._container = document.createElement("span");
                    oContainer.className = "ax-terminal-textField";
                    if (field instanceof AXES.Terminal.Field.TextField) {
                        oContainer.className += " ax_x" + field.attribute.substring(1).toUpperCase() + "o";
                    }
                } else {
                    oContainer = this.createTextElement();
                }
            }
            fieldDiv.appendChild(oContainer);
            /* Listen for changes in the field value */
            field.addListener("ValueChanged", this._onFieldValueChanged, this);
            field.addListener("SetFocus", this.setFocus, this);
        },

        /**
         * We should put things that are impacted by change in properties here.
         *
         * The list of items is coming from the "fixedValues" property so the population
         * of the dropdown will happen in this method
         **/
        render: function(field, fieldDiv)
        {
            // render text output
            if (!field.isInput) {
                this._container.textContent = field.getValue();
                return;
            } 
            // render radio button
            var sSourceType = this.getPropertyValue("dataSourceType");
            var sTableName = "";
            var oTable = null;

            $(this._container).empty();

            var sName = "axRB_" + ((AXES.isTS2Engine) ? field.id : field.getId());

            if (this.isDesignMode()) 
            {
                var oRB = document.createElement("input");
                oRB.name = sName;
                oRB.id = sName;
                oRB.type = "radio";

                var oLabel = document.createElement("label");
                $(oLabel).attr("for", sName).text(TEXT.dev("axRadioButton.sampleRadioButtonCaption"));

                this._container.appendChild(oRB);
                this._container.appendChild(oLabel);

                this.applyStyle(this.getPropertyValue("radioButtonStyle"), oRB, true);
                this.applyStyle(this.getPropertyValue("captionStyle"), oLabel, true);
            }
            else
            {
                var oRBStyle = this.getPropertyValue("radioButtonStyle");
                var oSpanStyle = this.getPropertyValue("captionStyle");

                switch (sSourceType)
                {
                    case "fixedValues":
                        var item;
                        var items = this.getPropertyValue("fixedValues");

                          AXES.TableManager.deleteTable(this._tempFixedValuesTable); /* Drop anything hanging around */  
                        oTable = AXES.TableManager.getTable(); /* Force creation of a brand new table */ 
                        this._tempFixedValuesTable = oTable.name; 
                        sTableName = oTable.name; 

                        for (var i in items) 
                        {
                           item = items[i];
                           oTable.insertChild({value:item.value,text:item.text}); 
                        }

                        break;

                    case "static":
                        sTableName = this.getPropertyValue("tableName");
                        oTable = AXES.TableManager.getTable(sTableName);
                        AXES.TableManager.loadStaticTables(USERENV.staticTablesFile, {}, false);
                        break;

                    case "dynamic":
                        var env = {
                            SQL: {}
                        };
                        this.raiseEvent("sqlVariables", env);

                                                var bKeepLastKey = false ;
                                                bKeepLastKey = this.getPropertyValue("keepLastKey"); 
                        
                        sTableName = this.getPropertyValue("sqlQueryName");
                        AXES.TableManager.loadDynamicTable(sTableName, USERENV.dynamicTablesFile, env.SQL, bKeepLastKey);
                        oTable = AXES.TableManager.getTable(sTableName);
                        break;

                    case "xml":
                        
                        var sPath = this.getPropertyValue("xmlFileLocation");
                        if (sPath == "*DEFAULT") {
                            sPath = AXES.TableManager.getDefaultPath();
                        }

                        var sFullFileName = sPath + "/" + this.getPropertyValue("xmlFileName");
                        sTableName = sFullFileName;
                        oTable = AXES.TableManager.getTable(sTableName);
                        
                        AXES.TableManager.loadFromXML(sFullFileName, oTable, false);
                        break;
                }

                var oRow, env;
                var bDisabled = (field.isInput) ? false : true;
                var sFldValue = field.getValue().trim();
                var oRadioB, oLabel, oButtonWrapper, sCaption;
                var bChecked = false;

                var sWrapperDisplay = (this.getPropertyValue("orientation") == "H") ? "inline" : "block";

                for (var i = 0; i < oTable.childCount(); i++) 
                {
                    oRow = oTable.child(i);
                    env = {
                        ROW: oRow,
                        FIELD: field
                    };

                    sCaption = this.getPropertyValue("onCreateRadioButton", env);
                    
                    if (sCaption != null) 
                    {
                        var rbid = sName + "_" + i;

                        oButtonWrapper = document.createElement("span");
                        oButtonWrapper.style.display = sWrapperDisplay;

                        var br = jQuery.browser;
                        if (br.msie && br.version < 8) 
                        {
                            // IE6 and IE7 have a bug in which setting the name attribute does
                            // not work. Without a name, the button cannot be set. So, use an IE-only method to create the element.
                            oRadioB = document.createElement('<input type="radio" id="' + rbid + '" name="' + sName + '">');
                        } else
                        {
                            oRadioB = document.createElement("input");
                            oRadioB.id = rbid;
                            oRadioB.type = "radio";
                            oRadioB.name = sName;
                        }
                        oRadioB.disabled = bDisabled;
                        $(oRadioB).bind("click", this, this._onClick).data("RowRef", oRow);
                        $( oRadioB ).bind( "change", this, this._onChange );
                        $( oRadioB ).bind( "focus", this, this._onFocus );
                        $( oRadioB ).bind( "blur", this, this._onBlur );

                        oLabel = document.createElement("label");
                        $(oLabel).attr("for", rbid).text(sCaption);

                        oButtonWrapper.appendChild(oRadioB);
                        oButtonWrapper.appendChild(oLabel);
                        this._container.appendChild(oButtonWrapper);

                        this.applyTabIndex( oRadioB, true );

                        this.applyStyle(this.getPropertyValue("radioButtonStyle"), oRadioB, true);
                        this.applyStyle(this.getPropertyValue("captionStyle"), oLabel, true);

                        oRadioB.checked = false;
                        if (!bChecked) 
                        {
                            bChecked = this.getPropertyValue("onSelectRadio", env);
                            
                            /* The type returned by onSelectRadio will depend on whether the user has specified the property as a simple expression or a more complex one */
                            /* For example, the default value of onSelectRadio is ROW.value. This is interpreted as the user wanting to select the radio button entry when the */
                            /* the value of the field associated with the radio button is equal to the current ROW.value. However, the user may have specified something like: */
                            /* ROW.value == (something or other). Hence, we inspect the type of the value returned by the property. If its a boolean it means the user specified */
                            /* a complex expression that yields a true/false. Otherwise we compare the property with the field's value */
                            if (typeof(bChecked) != "boolean") 
                            {
                                bChecked = (bChecked == sFldValue);
                            }
                            
                            oRadioB.checked = bChecked;
                        }
                        
                    } /* end if (sCaption != null) */
                } /* end for */
            }
        },

        /**
         * Do opposite of init
         **/
        destroy: function(field, fieldDiv)
        {
            // delete any temporary table created - delete will ignore null/empty names
            AXES.TableManager.deleteTable(this._tempFixedValuesTable);   
            this._tempFixedValuesTable = "";     
            // Remove visual parts 
            $(this._container).remove();
            this._field = null;
            /* Remove any listeners */
            field.removeListener("ValueChanged", this._onFieldValueChanged);
            field.removeListener("SetFocus", this.setFocus);
        },
        /* 
        * Field focus
        */
        setFocus: function(event)
        {
        // May not be able to set the focus if the field is invisible
            try 
            {
                if (event.field.type == AXES.Terminal.Field.Types.INPUT)
                {
                    if (event.offset != undefined) {
                        this.offsetOnFocus = event.offset;
                    }
                    var el = $(this._container).find("input,radio").get(event.offset ? event.offset : 0);
                    el.focus();
                }
            } catch(e) { }
        },
        /**
         * Handles the onclick event of the radio button
         **/
        _onClick: function( e )
        {
            var self = e.data;
            // raise the user event
            self._noFieldValueChangedEvent = true;
            var env =
            {
                ROW: $( this ).data( "RowRef" ),
                FIELD: self._field
            };

            if ( $.browser.webkit || $.browser.safari )
            {
                if ( self._focusedChrome )
                    self._focusedChrome = false;
                else
                    this.focus();
            }

            self.raiseEvent("onRadioButtonClicked", env);
        },

        /* Handles the change in field value */
        _onFieldValueChanged: function(event)
        {
            var self = this;
            if (!(self._field.isInput)) return;
            if (!self._noFieldValueChangedEvent) 
            {
                var sName = "axRB_" + ((AXES.isTS2Engine) ? this._field.id : this._field.getId());
                var sFldValue = TRIM(self._field.getValue());
                var env;

                $("input:radio", this._container).each(function()
                {
                    env = {
                        ROW: $(this).data("RowRef"),
                        FIELD: self._field
                    };
                    var bChecked = self.getPropertyValue("onSelectRadio", env);
                    if (typeof(bChecked) != "boolean") 
                    {
                        bChecked = (self.getPropertyValue("onSelectRadio", env) == sFldValue);
                    }
                    this.checked = bChecked;
                    if (bChecked) return false; // stop the loop
                });
            }
            else 
            {
                delete self._noFieldValueChangedEvent;
            }
        },

        _onChange: function( e )
        {
            var self = e.data;

            if ( self.isDesignMode() )
                return;

            // simulate the element to receive focus on chrome and safari
            if ( $.browser.webkit || $.browser.safari )
            {
                this.focus();
                self._focusedChrome = true;
            }
            
            if ( AXES.isTS2Engine && $.browser.msie && "7.0" == $.browser.version )
            {
                if ( self._changeEventOwner != this )
                {
                    self.raiseEvent( "onChange", { ROW: $( this ).data( "RowRef" ), FIELD: self._field } );
                    self._changeEventOwner = this;
                }
            }
            else
                self.raiseEvent( "onChange", { ROW: $( this ).data( "RowRef" ), FIELD: self._field } );
        },

        _onFocus: function( e )
        {
            var self = e.data;
            if ( !self.isDesignMode() ) 
                self.raiseEvent( "onFocus", { ROW: $( this ).data( "RowRef" ), FIELD: self._field } );
        },

        _onBlur: function( e )
        {
            var self = e.data;
            if ( !self.isDesignMode() ) 
                self.raiseEvent( "onBlur", { ROW: $(this).data( "RowRef" ), FIELD: self._field } );
        }
    };

    /**
     * Register the extension with AXES
     **/
    AXES.Extensions.add(definition_axRadioButton);

    /**
     * The function belows defines the drop down item object.
     * Each drop down item has a "value" and "text" property.
     * The "value" property is what will be used as the actual 5250 field value and
     * the "text" property is what will be displayed on the drop down.
     **/
    function defineRadioButtonItemClass()
    {
        /** 
         * constructor function for the drop down item object
         **/
        var itemClass = function()
        {
            this.value = "";
            this.text = "";
        };

        /** 
         * "toString" function for the item object.
         * This will be used by the designer when it needs to display a simple representation
         * of the drop down item.
         **/
        itemClass.prototype.toString = function()
        {
            var value = this.getPropertyValue("value");
            var text = this.getPropertyValue("text");
            
            return value + " - " + text;
        };

        /** 
         * Properties of drop down item object.
         **/
        var itemClassDef = {
            value: {
                type: "String"
            },
            text: {
                type: "String"
            }
        };

        /** 
         * We need to register the constructor function and property list with AXES.
         * We want to name the item object class "axRadioButtonitem".
         **/
        AXES.Extensions.definePropertyType("axRadioButtonitem", {
            objectClass: itemClass,
            objectDefinition: itemClassDef
        });
    }
}

/**
 * Make sure we call the function we just defined.
 **/
DEFINE_EXTENSION_axRadioButton();