/*********************************************************************
 *                                                                   *
 * Copyright (C) LANSA Group. All rights reserved                    *
 *                                                                   *
 *********************************************************************/

/**
 * @fileOverview Contains AXES.Layout and AXES.UI
 */

/**
 * Handles the visual layout of the terminal
 * @namespace
 * @augments AXES.EventsBase
 * @augments AXES.Extensions.ExtendableObjectBase
 * @property {Object} characterSize The size, in pixels, of one character block in the terminal grid.
 * @property {Integer} _type The type of the current layout: 0=none/hidden, 1=normal, 2=custom html page.
 */
AXES.Layout = {
    _focusElement: null,
    _lastFocusElement: null,
    _layoutCleared: false,
    extensions: [],
    userFields: [],
    customTabOrder: [],
    characterSize: {width: 0, height:0},
    _type: 0,
    _zoomSetting: "auto",
    _zoom: 1,
    _messages: [],
    horizPad: 2, //measured in pixels
    vertPad: 2, //measured in pixels
    spaceBetweenRows: 3, // measured in pixels
    spaceBetweenCols: 0, // measured in pixels
    show5250Cursor: false,
    moveableWindows: true,
    _isCustomized: false,
    _customUserElements : []
}

AXES.Layout.init = function() {
    this.displayArea = $("#terminalInner");
    this._terminalWrapper = $("#terminalWrapper");
    this._terminalContent = $("#terminalContent");

    $(document).keydown(AXES.Keyboard.keydownHandler)
        .keypress(AXES.Keyboard.keypressHandler)
        .keyup(AXES.Keyboard.keyupHandler);

        //
        if (!AXES.AppHost)
        {
            window.onbeforeunload = function(e)
            {
                var e = e || window.event;
                if (AXES.Terminal.isConnected) 
                {
                    if (AXES.EXTHOST_TSInterface)
                    {
                        return;
                    }
                    var msg = TEXT.Core("axTerm.Msg001");
                    if (e) e.returnValue = msg;
                    return msg;
                }
            }
        }

        //
        $(window).unload(function(e)
        {
            AXES.closeSession(true);
            AXES.Application.stopExtensions();
        });
    
        // Focus/blur events do not bubble so we can't use standard event delegation 
        // to handle them. IE has a focusin event that does bubble. For other
        // browsers we need to use event capturing.
        var daDiv = this.displayArea[0];
        if (daDiv.addEventListener) {
            daDiv.addEventListener("focus", AXES.Layout.elementFocus, true);
            daDiv.addEventListener("blur", AXES.Layout.elementBlur, true);
        } else {
            daDiv.attachEvent("onfocusin", AXES.Layout.elementFocus);
            daDiv.attachEvent("onfocusout", AXES.Layout.elementBlur);
        }
        // Suppress the IE help on F1
        $("body").bind("help", function(e) {
            e.preventDefault();
        });
	
	//suppress context menu
	$("body").bind("contextmenu", function(e) {
		return false;
	});
	
    $("#terminalRefreshButton").click(function(e) {
        AXES.Terminal.doSysPost("REDRAW");
    });
    $("#terminalCancelButton").click(function(e) {
        AXES.Terminal.doSysPost("SYSREQ2");
    });
    $("#terminalClearButton").click(function(e) {
        AXES.Layout.clearMessages();
    });
    
    /**
     * Indicates that the zoom level for the terminal display has changed.
     * @name AXES.Layout#ZoomChanged
     * @event
     * @param {Object} e Event Parameters.
     * @param {Number} e.zoomAmount The zoom amount as a percentage.
     */
    
    /**
     * Indicates that the layout is about to start rendering. This event is fired after the screen has been processed to load customizations and AutoGUI eXtensions and just before rendering to the display.
     * It is not fired for screens that use a custom html page such as the login screen.
     * @name AXES.Layout#PreRender
     * @event
     */
    
    /**
     * Indicates that the layout has completed rendering. This event is fired just prior to doing auto-zoom calculations.
     * It is not fired for screens that use a custom html page such as the login screen.
     * @name AXES.Layout#RenderComplete
     * @event
     */
    
    jQuery.extend(true, this, AXES.EventsBase, {
        _eventNames: {
            ZoomChanged: true, RenderComplete: true, PreRender: true
        }, _listeners: {}
    });
    
    jQuery.extend(true, this, AXES.Extensions.ExtendableObjectBase, {extensions: []});
    
    this.AutoGUI.init();
    
    this.alertAudio = false;
    try 
    {
        if (Modernizr.audio) 
        {
            if (Modernizr.audio.mp3) 
            {
                this.alertAudio = new Audio("error.mp3");
            }
            else if (Modernizr.audio.wav) 
            {
                this.alertAudio = new Audio("error.wav");
            }
        }
    } catch(e) {}
}
/**
 * Handles focus events at the displayArea level to track the currently active element
 * @param {Event} e
 */
AXES.Layout.elementFocus = function(e) {
    e = jQuery.event.fix(e);
    AXES.Layout._focusElement = e.target;
}
/**
 * Handles blur events at the displayArea level to track the currently active element
 * @param {Event} e
 */
AXES.Layout.elementBlur = function(e) {
    e = jQuery.event.fix(e);
    AXES.Layout._lastFocusElement = AXES.Layout._focusElement
    AXES.Layout._focusElement = null;
}
/**
 * Sets the CSS background color of the terminal area
 * @param {String} color Any valid CSS color value.
 */
AXES.Layout.setBackgroundColor = function(color) 
{
    $("#terminalInner").css({
        "background-color": color,
        "background-image": "none"
    });
}
/**
 * Returns the CSS background color of the terminal area. The value should be an
 * rgb or rgba value of the form "rgb(r,g,b)" or "rgba(r,g,b,a)" however some 
 * browsers may vary and return a hex value or a color name.
 * @returns {String}
 */
AXES.Layout.getBackgroundColor = function() 
{
    return $("#terminalInner").css("background-color");
}

/**
 * Adjusts the zoom of the terminal area. Possible values and their meanings are:
 * "auto": Automatically adjust each screen to fit in the available are. Continue to adjust if the screen is resized
 * "fit": Set the zoom value so that the current screen fits in the available area.
 * "+": Add 0.05 to the current zoom value.
 * "-": Subtract 0.05 from the current value.
 * value: Set the zoom to the supplied value.
 * The zoom value cannot be less than 0.1 (10%) or greater than 5 (500%). Any operation that would result in a value outside this range will be limited to keep the value valid.
 * @param {String | number} zoomValue The zoom setting. Allowed values are "auto", "fit", "+", "-" or a number between 0.1 and 5.
 */
AXES.Layout.setZoom = function(zoomValue) {
    if (AXES.Layout._designMode) return;
    var val = NaN, oldZoom = this._zoom;
    this._zoomSetting = "";
    switch (zoomValue)
    {
        case "fit":
        case "auto":
            this._zoomSetting = zoomValue;
            this.applyZoom();
            break;
        case "+":
            val = this._zoom + 0.05;
            break;
        case "-":
            val = this._zoom - 0.05;
            break;
        default:
            val = new Number(zoomValue);
    }
    if (!isNaN(val)) 
    {
        this._zoom = (val < 0.1) ? 0.1 : ((val > 5) ? 5 : val);
    }
    this.applyZoom();
    if (this._zoom != oldZoom) 
    {
        this.fire("ZoomChanged");
    }
}
/**
 * Gets the current zoom value as a percentage.
 * @returns (Number) The current zoom level. A negative value indicates that the zoom is not applied either because a custom page is being displayed or the browser does not support it.
 */
AXES.Layout.getZoom = function() {
    return Math.round(this._zoom * 100);
}
AXES.Layout.getZoomLock = function() 
{
    var zo = this.autoZoomOverride;
    return zo ? zo.lockZoom : false;
}
/**
 * Gets the current zoom setting. If the zoom has been set to a fixed value this will be an empty string, otherwise it
 * will be "auto" or "fit".
 * @returns {String} The current zoom setting.
 */
AXES.Layout.getZoomSetting = function() 
{
    return this._zoomSetting;
}
/**
 * Sets the zoom of an element to a specified amount
 * @param {Number} zoomAmount The zoom amount as a fraction (e.g. 0.5 = 50%)
 * @param {jQuery Objec} $termContentDiv A jQuery object pointing to the element to set the zoom on.
 * @returns {Boolean} True if successful (may fail if the browser does not support this functionality).
 */
AXES.Layout._setZoomCSS = function( zoomAmount, $elemToZoom )
{
    if ( this._zoomNotSupported )
    {
        return false;
    }

    var css = {};
    var s = document.body.style;
    if ( s[ "transform" ] != undefined || s[ "-ms-transform" ] != undefined || s[ "MozTransform" ] != undefined || s[ "-webkit-transform" ] != undefined || s[ "-o-transform" ] != undefined )
    {
        var trVal = "scale(" + zoomAmount + "," + zoomAmount + ")";
        css =
        {
            "transform"        : trVal,
            "-moz-transform"   : trVal,
            "-webkit-transform": trVal,
            "-ms-transform"    : trVal,
            "-o-transform"     : trVal
        };
    }
    else if ( s[ "zoom" ] != undefined )
    {
        css.zoom = zoomAmount;
        if ( $( 'html' ).is( '.ie8' ) )
            css.overflow = "hidden";
    }
    else 
    {
        return false;
    }

    $elemToZoom.css( css );
    return true;
}
/**
 * Applies the current zoom setting to the display
 * @private
 */
AXES.Layout.applyZoom = function()
{
    var tc = this._terminalContent, zoom = this._zoom, zo = this.autoZoomOverride,
    cH, cW, cs, maxW=0, maxH=0;
    var setZoomOK = this._setZoomCSS(1, tc);
    
    cH = this.characterSize.height;
    cW = this.characterSize.width;
    cs = this.contentSize = AXES.Layout.characterDimensionsToPixels({
        width: AXES.Terminal.cols,
        height: AXES.Terminal.rows
    }, false, true);
    cs.width += (this.horizPad * 2);
    cs.height += (this.vertPad * 2);
    
    // Work out the actual content size (need to do this even for zoom=1 because
    // the terminal wrapper and overlay divs are sized at the end of this.
    if (!zo || (zo.height == 0) || (zo.width == 0)) 
    {
        tc.children(":not(.ui-widget-overlay):not(#axDesignGrid)").each(function()
        {
            maxW = Math.max(this.offsetLeft + this.offsetWidth, maxW);
            maxH = Math.max(this.offsetTop + this.offsetHeight, maxH);
        });
        maxW += this.horizPad;
        maxH += this.vertPad;
        
        cs.width = Math.max(cs.width, maxW);
        cs.height = Math.max(cs.height, maxH);
    }
    
    if (zo) 
    {
        if (zo.width) cs.width = zo.width;
        if (zo.height) cs.height = zo.height;
    }
    
    if (!setZoomOK) 
    {
        this._zoomNotSupported = true;
        zoom = -1;
    }
    else if (this._type == 1) 
    {
        // Calculate the zoom
        zoom = Math.abs(zoom);

        var zs = this._zoomSetting;
        if ((zs == "auto") || (zs == "fit")) 
        {
            var displaySize = {
                width: this.displayArea.innerWidth(),
                height: this.displayArea.innerHeight()
            },
            scale1 = displaySize.width / cs.width,
            scale2 = displaySize.height / cs.height;
            
            zoom = Math.floor(Math.min(scale2, scale1) * 100) / 100;
            if (zs == "fit") this._zoomSetting = "";
        }
        
        this._setZoomCSS(zoom, tc);
    }
    else 
    {
        zoom *= -1;
    }
    this._terminalWrapper.width(Math.ceil(cs.width * Math.abs(zoom)))
        .height(Math.ceil(cs.height * Math.abs(zoom)));
        
    tc.find(".ui-widget-overlay").width(cs.width).height(cs.height);
    this._zoom = zoom;
}
AXES.Layout.reapplyZoom = function() 
{
    var tc = $("#terminalContent");
    if ((tc.length > 0) && (this._zoom > 0)) 
    {
        this._setZoomCSS(this._zoom, tc);
    }
}
/**
 * Handles resize events from the layout pane containing the terminal.
 */
AXES.Layout.onResize = function() {
    var oldZoom = this._zoom;
    this.applyZoom();
    if (this._zoom != oldZoom) 
    {
        this.fire("ZoomChanged");
    }
}
AXES.Layout.displayHostBusyMessage = function(callbackOnDone)
{
    AXES.UI.clearDialogs();
    var dialogOptions = {
        closeOnEscape: false,
        bgiframe: true,
        modal: true,
        position: "top",
        buttons: {}
    }
    dialogOptions.buttons[TEXT.Core("Yes")] = function()
    {
        $(this).dialog('close').dialog('destroy');
        callbackOnDone(true);
    };
    dialogOptions.buttons[TEXT.Core("No")] = function()
    {
        $(this).dialog('close').dialog('destroy');
        callbackOnDone(false);
    };
    
    $("<div></div>").attr("title", TEXT.Core("Attention"))
    .text(TEXT.Core("The application is busy. Do you want to continue waiting?")).dialog(dialogOptions);
    this.playAlert();
}
AXES.Layout.displayServerMessage = function(theMessage)
{
    AXES.UI.clearDialogs();
    var dialogOptions = {
        closeOnEscape: false,
        bgiframe: true,
        modal: true,
        position: "top",
        buttons: {}
    }
    dialogOptions.buttons[TEXT.Core("OK")] = function()
    {
        $(this).dialog('close').dialog('destroy');
    };
    
    $("<div></div>").attr("title", TEXT.Core("Attention"))
    .text(theMessage).dialog(dialogOptions);
    this.playAlert();
}
AXES.Layout.displayErrorAlert = function(theMessage)
{
    if (this._errorAlertDialog) {
        // An alert is already displayed. Queue this one.
        if (!this._errorAlertQueue) this._errorAlertQueue = [];
        this._errorAlertQueue.push(theMessage);
        return;
    }
    AXES.UI.clearDialogs();
    var dialogOptions = {
        closeOnEscape: true,
        bgiframe: true,
        modal: true,
        position: "top",
        close: function()
        {
            delete AXES.Layout._errorAlertDialog;
            if (AXES.Layout._errorAlertQueue && (AXES.Layout._errorAlertQueue.length > 0)) 
            {
                window.setTimeout(function()
                {
                    AXES.Layout.displayErrorAlert(AXES.Layout._errorAlertQueue.shift());
                }, 10);
            }
        },
        buttons: {}
    }
    dialogOptions.buttons[TEXT.Core("OK")] = function()
    {
        $(this).dialog('close').dialog('destroy');
    };
    this._errorAlertDialog = $("<div></div>").attr("title", TEXT.Core("Attention"))
    .html('<span class="ui-icon ui-icon-alert" style="float:left; margin:0 7px 50px 0;"></span>' + theMessage)
    .dialog(dialogOptions);
    this.playAlert();
}
AXES.Layout.displayVerifyUserDialog = function(header, message, user, callback) {
    var dialogOptions = {
        closeOnEscape: false,
        bgiframe: true,
        modal: true,
        position: "top",
        buttons: {}
    };

    dialogOptions.buttons[TEXT.Core("Verify")] = function()
    {
        $(this).dialog('close').dialog('destroy');
        callback({
            user: $("#ax-terminalVerifyName").val(),
            password: $("#ax-terminalVerifyPassword").val()
        });
    };
    dialogOptions.buttons[TEXT.Core("Cancel")] = function()
    {
        $(this).dialog('close').dialog('destroy');
        callback(null);
    };

    $("#ax-terminalVerifyUserMessage").text(message);
    $("#ax-terminalVerifyName").val(user);
    $("#ax-terminalVerifyPassword").val("");
    this._verifyUserDialog = $("#ax-terminalVerifyUserDialog")
    .attr("title", header)
    .bind("keydown keypress keyup", function(e) { e.stopPropagation(); })
    .dialog(dialogOptions);
    this.playAlert();
}
/**
 * A utility method that sets the selection to a specified position within an
 * input field.  The field must already have focus.
 * @param {Element} elem The element. Must be a text|password type INPUT or a TEXTAREA.
 * @param {Number} startPos The position of the first character in the selection
 * @param {Number} [endPos] The position of the last character in the selection
 */
AXES.Layout.setFieldSelection = function(elem, startPos, endPos) {
    if (!elem) return;
    if (!startPos) startPos = 0;
    if (!endPos) endPos = startPos;
    if (elem.selectionStart >= 0) 
    {
        elem.setSelectionRange(startPos, endPos);
    }
    else if (elem.createTextRange) 
    {
        var range = elem.createTextRange();
        range.moveStart("character", startPos);
        range.moveEnd("character", (endPos - elem.value.length));
        range.select();
    }
}
/**
 * A utility method that gets the position of the selected text or caret
 * in an input field. The field must already have focus.
 * @param {Element} elem The element
 */
AXES.Layout.getFieldSelection = function(elem) {
    var retVal = {start: 0, end: 0};
    if (!elem) return retVal;
    if(elem.selectionStart >= 0) {
        retVal.start = elem.selectionStart;
        retVal.end = elem.selectionEnd;
    } else if (elem.createTextRange) {
        var selRange = document.selection.createRange();
        if (selRange.parentElement() === elem) 
        {
            // IE makes this difficult
            var elemRange = selRange.duplicate();
            // Select all text
            if (elem.nodeName == "INPUT") 
            {
                elemRange.expand("textedit");
            }
            else 
            {
                elemRange.moveToElementText(elem);
            }
            // Now move 'dummy' end point to end point of original range
            elemRange.setEndPoint('EndToEnd', selRange);
            // Now we can calculate start and end points
            retVal.start = elemRange.text.length - selRange.text.length;
            retVal.end = retVal.start + selRange.text.length;
        }
    }
    return retVal;
}
/**
 * Returns the HTML element within the terminal that has the focus, or null if nothing has the focus.
 * @returns {HTMLElement}
 */
AXES.Layout.getActiveElement = function() 
{
    var ae = document.activeElement;
    // Make sure it is a child of the terminal div.
    if ($(ae).parents("#terminalInner").length) 
    {
        return ae;
    }
    else 
    {
        return this._lastFocusElement;
    }
}
/**
 * Returns the AXES.Terminal.Field associated with the active element (if there is one).
 * @returns {AXES.Terminal.Field.Base}
 */
AXES.Layout.getActiveField = function()
{
    return this.getFieldFromHTMLElement(this.getActiveElement());
}
/**
 * Finds and returns the Terminal Field (5250 or UserField) associated with the supplied DOM element.
 * @param {HTMLElement} element
 * @returns {AXES.Terminal.Field.Base} The associated Field or null if not found.
 */
AXES.Layout.getFieldFromHTMLElement = function(element)
{
    var fldDiv = $(element).parents("div.ax-terminal-field");
    var id = fldDiv.attr("id");
    if (fldDiv.length == 0) return null;
    var fld = AXES.Terminal.getFieldById(id);
    if (!fld) 
    {
        for (var i = 0; fld = this.userFields[i]; i++) 
        {
            if (fld.id == id) break;
        }
    }
    return (fld) ? fld : null;
}
/**
 * Moves the focus to the next field.
 * @returns {Boolean} Returns false if unable to set the next field (such as for custom pages).
 */
AXES.Layout.nextField = function()
{
    /*
     * This is quite a complex business. AXES.Terminal.sequencedFields contains a list of
     * fields in the topmost window, sorted by screen order. This is the default tab order.
     * The host may modify the tab order by specifying the "tabNext" property.
     * The host may also link a group of fields together with the "ctd", "ctd_to" and
     * "ctd_from" properties.
     * Add to this, UserFields added to the screen by an aXes developer
     * And finally, the aXes developer may apply a custom tabIndex to a host field
     * or user field.
     */
    if (AXES.Terminal.isLogin) return false;
    var nextField = null;
    var termFields = AXES.Terminal.sequencedFields;
    var ctoInfo;
    
    // Is there a currently active element
    var ae = this.getActiveElement();
    if (ae) 
    {
        // See if it is in the custom tabIndex sequence.
        var ctoInfoIndex;
        for (ctoInfoIndex = 0; ctoInfo = this.customTabOrder[ctoInfoIndex]; ctoInfoIndex++) 
        {
            if (ctoInfo.htmlElement === ae) break;
        }
        if (ctoInfo) 
        {
            var ctoNext = this.customTabOrder[ctoInfoIndex+1];
            if (ctoNext && (ctoNext.field === ctoInfo.field)) 
            {
                // The next in the render sequence belongs to the same field so just go there.
                ctoNext.htmlElement.focus();
                return true;
            }
            if (!ctoNext) ctoNext = this.customTabOrder[0];
            if (ctoInfo.tabIndex > 0) 
            {
                // It is part of the user defined tab sequence, just return the next field
                // added check if nextfield is visible
                do 
                {
                    if (++ctoInfoIndex >= this.customTabOrder.length) ctoInfoIndex = 0;
                    nextField = this.customTabOrder[ctoInfoIndex].field;
                    var nfldExt = nextField.getExtensions()[0];
                    var visible = nfldExt.getPropertyValue('visible',this) 
                }    while (!visible);                
            }
            else 
            {
                var fld = ctoInfo.field;
                // TabIndex is 0. That means follow the host defined order or rendering order.
                // First, see if the host gave any instructions
                if (fld.tabNext) 
                {
                    // The host has specified the next field in the tab sequence
                    if (fld.tabNext <= termFields.length) nextField = termFields[fld.tabNext - 1];
                }
                else if (fld.ctd_to) 
                {
                    // The field is part of a cto sequence
                    var nextID = "ifld" + fld.ctd_to;
                    nextField = AXES.Terminal.getFieldById(nextID);
                }
                else 
                {
                    // The host hasn't specified anything. Pick the next in the rendering sequence
                    // if the next field is not visible look for the next visible field.    
                    do 
                    {                
                        if (++ctoInfoIndex >= this.customTabOrder.length) ctoInfoIndex = 0;                
                        nextField = this.customTabOrder[ctoInfoIndex].field;
                        var nfldExt = nextField.getExtensions()[0];
                        var visible = nfldExt.getPropertyValue('visible',this) 
                    }    while (!visible);        
                }
                if (nextField) 
                {
                    while (nextField.ctd_from) 
                    {
                        // If the next field is in the middle of a ctd sequence
                        // we need to go back to the beginning (unless the field
                        // we came from is the previous field in the same 
                        // sequence).
                        var prevID = "ifld" + nextField.ctd_from;
                        if (prevID == fld.id) break;
                        nextField = AXES.Terminal.getFieldById(prevID);
                    }
                }
            }
        }
    }
    
    if (!nextField && this.customTabOrder.length) nextField = this.customTabOrder[0].field;
    if (nextField) nextField.setFocus();
    return true;
}
/**
 * Moves the focus to the previous field.
 * @returns {Boolean} Returns false if unable to set the next field (such as for custom pages).
 */
AXES.Layout.prevField = function()
{
    /*
     * This is quite a complex business. AXES.Terminal.sequencedFields contains a list of
     * fields in the topmost window, sorted by screen order. This is the default tab order.
     * The host may modify the tab order by specifying the "tabNext" property.
     * The host may also link a group of fields together with the "ctd", "ctd_to" and
     * "ctd_from" properties.
     * Add to this, UserFields added to the screen by an aXes developer
     * And finally, the aXes developer may apply a custom tabIndex to a host field
     * or user field.
     */
    if (AXES.Terminal.isLogin) return false;
    var prevField = null;
    var termFields = AXES.Terminal.sequencedFields;
    var ctoInfo;
    
    // Is there a currently active element
    var ae = this.getActiveElement();
    if (ae) 
    {
        // See if it is in the custom tabIndex sequence.
        var ctoInfoIndex;
        for (ctoInfoIndex = 0; ctoInfo = this.customTabOrder[ctoInfoIndex]; ctoInfoIndex++) 
        {
            if (ctoInfo.htmlElement === ae) break;
        }
        if (ctoInfo) 
        {
            var ctoPrev = this.customTabOrder[ctoInfoIndex-1];
            if (ctoPrev && (ctoPrev.field === ctoInfo.field)) 
            {
                // The next in the render sequence belongs to the same field so just go there.
                ctoPrev.htmlElement.focus();
                return true;
            }
            if (!ctoPrev) ctoPrev = this.customTabOrder[this.customTabOrder.length-1];
            if (ctoInfo.tabIndex > 0) 
            {
                // It is part of the user defined tab sequence, just return the previous field
                // added check if prevfield is visible
                do 
                {                
                    if (--ctoInfoIndex < 0) ctoInfoIndex = this.customTabOrder.length-1;
                    prevField = this.customTabOrder[ctoInfoIndex].field;
                    var nfldExt = prevField.getExtensions()[0];
                    var visible = nfldExt.getPropertyValue('visible',this) 
                }    while (!visible);                                                
            }
            else 
            {
                // TabIndex is 0. That means follow the host defined order or rendering order.
                // First, see if the host gave any instructions
                var fld = ctoInfo.field;
                if (fld.ctd_from) 
                {
                
                    // If the field is part of a ctd sequence, just get the previous field in the sequence    
                    var prevID = "ifld" + fld.ctd_from;
                    prevField = AXES.Terminal.getFieldById(prevID);
                }
                else 
                {
                    // See if there is a field who's tabNext points to this
                    var fldIndex = -1;
                    var l = termFields.length;
                    for (var i = 0; i < l; i++) 
                    {
                        if (termFields[i].tabNext && termFields[termFields[i].tabNext - 1] === fld) 
                        {
                            prevField = termFields[i];
                            fldIndex = -1;
                            break;
                        }
                    }
                }
                if (!prevField) 
                {
                    // The host hasn't specified anything. Pick the next in the rendering sequence
                    // if the prev field is not visible look for the prev visible field.                    
                    do 
                    {                
                        if (--ctoInfoIndex < 0) ctoInfoIndex = this.customTabOrder.length-1;
                        prevField = this.customTabOrder[ctoInfoIndex].field;
                        var nfldExt = prevField.getExtensions()[0];
                        var visible = nfldExt.getPropertyValue('visible',this) 
                    }    while (!visible);                        
                }
                if (prevField) 
                {
                    while (prevField.ctd_to) 
                    {
                        // If the next field is in the middle of a ctd sequence
                        // we need to go to the end (unless the field
                        // we came from is the next field in the same 
                        // sequence).
                        var prevID = "ifld" + prevField.ctd_to;
                        if (prevID == fld.id) break;
                        prevField = AXES.Terminal.getFieldById(prevID);
                    }
                }
            }
        }
    }
    
    if (!prevField && this.customTabOrder.length) prevField = this.customTabOrder[this.customTabOrder.length-1].field;
    if (prevField) prevField.setFocus();
    return true;
}
/**
 * Moves the focus to the first field on the next line.
 * @returns {Boolean} Returns false if unable to set the next field (such as for custom pages).
 */
AXES.Layout.newLine = function() {
    if (AXES.Terminal.isLogin) return false;
    
    var flds = AXES.Terminal.sortedFields, nextField, fld, l, i,
    types = AXES.Terminal.Field.Types, t1=types.INPUT, t2=types.SELECT;
    
    if (flds.length == 0) return false;
    nextField = flds[0];
    fld = AXES.Layout.getActiveField();
    if (fld) 
    {
        l = flds.length - 1;
        if (fld.row == flds[l].row) 
        {
            nextField = flds[0];
        }
        else 
        {
            for (i = 0; i < l; i++) 
            {
                nextField = flds[i];
                if ((nextField.type == t1 || nextField.type == t2) && !fld.byp && flds[i].row > fld.row) 
                {
                    nextField = flds[i];
                    break;
                }
            }
        }
    }
    if (nextField) nextField.setFocus();
    return true;
}
AXES.Layout.playAlert = function()
{
    if (this.alertAudio) 
    {
        this.alertAudio.play();
    }
    else 
    {
        var audioObj = document.getElementById("ax-sounds-error-bg");
        if (audioObj) audioObj.src = "error.wav";
    }
}
/**
 * Converts a supplied set of dimensions in character coordinates to pixel coordinates.
 * The returned values will include the space between columns and rows as specified
 * by the AXES.Layout.spaceBetweenRows and AXES.Layout.spaceBetweenCols properties.
 * @param {Object} dimensions The set of dimensions to convert. This object may have one or more of the following properties: top, left, right, bottom, width, height. Position values are 1-indexed.
 * @param {Boolean} [appendPx=false] If true, "px" will be appended to each value
 * @param {Boolean} [includeGap=true] The terminal layout includes a gap between columns and rows. If includeGap is true, this gap is added to the width and height values (not right or bottom).
 * @param {Boolean} [includePadding=true] The terminal contains 2 pixels of padding around the edge but windows do not. Set this parameter to false if calculating positions within a window.
 * @returns {Object} An object containing the converted properties.
 */
AXES.Layout.characterDimensionsToPixels = function(dimensions, appendPx, includeGap, includePadding)
{
    // If you modify this, remember to modify the version in axscreens.mjs
    var cH = this.characterSize.height + this.spaceBetweenRows;
    var cW = this.characterSize.width + this.spaceBetweenCols;
    if (includeGap === undefined) includeGap = true;
    if (includePadding === undefined) includePadding = true;
    var aX = (includeGap) ? 0: -this.spaceBetweenCols;
    var aY = (includeGap) ? 0: -this.spaceBetweenRows;
    var xPad = (includePadding) ? this.horizPad : 0;
    var yPad = (includePadding) ? this.vertPad : 0;
    var px = (appendPx) ? "px" : 0;
    var result = {};
    if (typeof dimensions.top == "number") result.top = Math.floor(((dimensions.top-1) * cH) + yPad) + px;
    if (typeof dimensions.left == "number") result.left = Math.floor(((dimensions.left-1) * cW) + xPad) + px;
    if (typeof dimensions.bottom == "number") result.bottom = Math.floor(((dimensions.bottom * cH) + yPad)) + px;
    if (typeof dimensions.right == "number") result.right = Math.floor(((dimensions.right * cW) + xPad)) + px;
    if (typeof dimensions.height == "number") result.height = Math.floor((dimensions.height * cH) + aY) + px;
    if (typeof dimensions.width == "number") result.width = Math.floor((dimensions.width * cW) + aX) + px;
    return result;
}
/**
 * Finds a field (either terminal or custom) with the supplied name.
 * @param {String} name Name of the field to find.
 * @param {Integer} [index] Index number of field to return (if it is a subfile field).
 */
AXES.Layout.getNamedField = function(name, index)
{
    var fld = AXES.Terminal.getFieldByName(name);
    if (!fld) 
    {
        name = name.toLowerCase();
        for (var i = 0, fld; fld = this.userFields[i]; i++) 
        {
            if (fld.symbolicName.toLowerCase() == name) break;
        }
    }
    if (!fld) 
    {
        // Named field not found, try id
        fld = AXES.Terminal.getFieldById(name);
        if (!fld) 
        {
            for (var i = 0, fld; fld = this.userFields[i]; i++) 
            {
                if (fld.id == name) break;
            }
        }
    }
    if (fld) 
    {
        if (index != undefined && (typeof index == "number" && !isNaN(index)) && fld.isIndexable()) 
        {
            fld = fld.getIndexedSubfileField(index);
        }
    }
    return (fld && !fld.byp) ? fld : null;
}
/**
 * Show or hide the Function Key buttons created by the AutoGUI. The optional kKeys parameter allows a subset of the buttons to be specified.
 * @param {Array} [fKeys] An array of function key names (e.g. ["F1","F3"]). If supplied, only the buttons for these keys will be affected.
 * @param {Boolean} show Indicates if the buttons should be shown or hidden
 */
AXES.Layout._showHideFKeyButtons = function(fKeys, show)
{
    // Find the required eXtensions
    var fkeyButtons;
    if (jQuery.isArray(fKeys)) 
    {
        fkeyButtons = [];
        for (var i = 0, fKey; fKey = fKeys[i]; i++) 
        {
            for (var j = 0, buttonInfo; buttonInfo = this.AutoGUI._fKeyButtons[j]; j++) 
            {
                if (buttonInfo.key == fKey) 
                {
                    fkeyButtons.push(buttonInfo);
                    break;
                }
            }
        }
    }
    else 
    {
        fkeyButtons = this.AutoGUI._fKeyButtons;
    }
    for (var j = 0, buttonInfo; buttonInfo = fkeyButtons[j]; j++) 
    {
        var ex = buttonInfo.extension;
        ex.setPropertyValue("hidden", {
            v: !show
        });
        ex.render();
    }
}
/**
 * Shows the Function Key buttons created by the AutoGUI. The optional kKeys parameter allows a subset of the buttons to be specified.
 * @param {Array} [fKeys] An array of function key names (e.g. ["F1","F3"]). If supplied, only the buttons for these keys will be affected.
 */
AXES.Layout.showFKeyButtons = function(fKeys) 
{
    this._showHideFKeyButtons(fKeys, true);
}
/**
 * Hides the Function Key buttons created by the AutoGUI. The optional kKeys parameter allows a subset of the buttons to be specified.
 * @param {Array} [fKeys] An array of function key names (e.g. ["F1","F3"]). If supplied, only the buttons for these keys will be affected.
 */
AXES.Layout.hideFKeyButtons = function(fKeys) 
{
    this._showHideFKeyButtons(fKeys, false);
}
AXES.Layout.renderBackgroundImages = function() 
{
    // Add any background images defined by the AutoGUI
        if (this.AutoGUI.autoGUIOn) 
        {
            for (var i = 0, img; img=this.AutoGUI.backgroundImages[i]; i++) 
            {
                if (!img.parentNode) this._terminalContent.prepend(img);
            }
        } else 
        {
            $("#terminalInner .ax-terminal-backgroundImg").remove();
        }
}

AXES.Layout.doLogin = function (info)
{
    var term = AXES.Terminal;
    var userNameField = term.getFieldById("loginUser");
    var passwordField = term.getFieldById("loginPwd");
    var programField = term.getFieldById("loginPgm");
    var curlibField = term.getFieldById("loginCurlib");

    //
    userNameField.setValue(info.user);
    passwordField.setValue(info.password);
    programField.setValue(info.program);
    curlibField.setValue(info.library);

    //
    AXES.Terminal.doPost("Enter");
}

AXES.Layout.loadLoginPage = function() 
{
    if (!this._layoutCleared) this.clearLayout();
    this._type = 2;
    var tw = this._terminalWrapper.get(0);
    if (tw.parentNode) 
    {
        //Remove the terminal wrapper from the DOM
        tw.parentNode.removeChild(tw);
    }
    var loginPage = AXES.urlParameters["login"] || "login.html",
    dotPos = loginPage.lastIndexOf("."),
    loginCSS = loginPage.substring(0,(dotPos > 0 ? dotPos : undefined)) + ".css";
    loginPage = (dotPos < 0) ? loginPage += ".html" : loginPage;
    $("#axCustomPageCSS").attr("href", AXESBUILD.updateURL(loginCSS));
    this.displayArea.load(AXESBUILD.updateURL(loginPage), null, function(responseText, textStatus, XMLHttpRequest) {
        if (textStatus == "success" || textStatus == "notmodified") 
        {
            var fe = $(document).data("focusElement");
            if (fe) 
            {
                $(document).removeData("focusElement");
                $(fe).focus().focus();
            }
        }
        else 
        {
            var msg = TEXT.core('Error %s1 (%s2) loading login page.', XMLHttpRequest.status, XMLHttpRequest.statusText);
            AXES.Debug.error(msg);
            AXES.Layout.displayArea.html("<p>" + msg + "</p>");
        }
    });
    document.title = AXES.Terminal.systemName + " - aXesTS";
    this._titleSet = false;
}
/**
 * Renders the terminal layout.
 */
AXES.Layout.doLayout = function( useDef )
{
    var term = AXES.Terminal;
    
    //AXES.UI.clearDialogs();
    this.clearMessages();
    this.setBusy( false );
    
    if ( term.isLogin )
    {
        if ( AXES.AppHost )
        {
            var args = {};
            if ( AXES.Terminal.messages.length )
                args.message = AXES.Terminal.messages[ 0 ];
            AXES.AppHost.exec( "Terminal", "login", args );
        }
        else 
            this.loadLoginPage();
    }
    else 
    {
        if ( !this._titleSet )
        {
            document.title = [
                term.systemName,
                term.deviceName,
                term.jobUser,
                term.jobNum
            ].join( " - " );
            this._titleSet = true;
        }

        if ( 1 == term.windows.length || !( AXES.urlParameters.getInt( "popups", 3 ) & 2 ) || 1 != this._type )
            this.clearLayout();
        var firstWin = 0;
        if ( !this._layoutCleared )
        {
            var self = this;
            firstWin = term.windows.length - 1;
            this._terminalContent
                .find( "div.ax-terminal-window, div.ax-terminal-pulldown" )
                .filter(
                    function( index )
                    {
                        return index >= ( firstWin - 1 );
                    }
                )
                .each(
                    function()
                    {
                        self.clearWindow( this );
                    }
                )
                .remove();
            this._terminalContent.find( "div.ui-widget-overlay" ).remove();
        }

        var extensions         = this.extensions;
        extensions.length      = 0;
        this.userFields.length = 0;
        this.customTabOrder    = [];
        this._isCustomized     = false;
        this._type             = 1;
        var terminalContainer  = this._terminalContent.get( 0 );
        terminalContainer.setAttribute( "id", "terminalContent" );
        
        if ( !this.characterSize.locked )
        {
            var charSize = $( '<span id="ax-terminal-charactersize" class="ax-terminal-textField" style="position:absolute;visibility:hidden;">M</span>' );
            terminalContainer.appendChild( charSize[ 0 ] );
            this.characterSize.width = charSize.width();
            this.characterSize.height = charSize.height() + 2;
            // Some browsers (Firefox) allow more height for their characters.
            if ( this.characterSize.browserHeightAdjustment == undefined )
            {
                charSize.css( "font-size", "10px" );
                this.characterSize.browserHeightAdjustment = charSize.height() - 10;
            }
            this.characterSize.height -= this.characterSize.browserHeightAdjustment;
            terminalContainer.removeChild( charSize[ 0 ] );
        }

        this._isCustomized = AXES.Extensions.getExtensionsForScreen( useDef );

        // Initialise screen extensions
        for ( var i = 0, ex; ex = extensions[ i ]; i++ )
        {
            if ( ex.renderStage && "preRender" != ex.renderStage )
                continue;

            ex._field = null;
            ex.init( terminalContainer );
            ex.render( terminalContainer );
        }

        var focusFld     = null;
        var container    = terminalContainer;
        var cH           = this.characterSize.height;
        var cW           = this.characterSize.width;
        var xOffset      = 0;
        var yOffset      = 0;
        this.contentSize =
        {
            width : ( AXES.Terminal.cols * cW ) + ( this.horizPad * 2 ),
            height: ( AXES.Terminal.rows * cH ) + ( this.vertPad * 2 )
        };

        this.AutoGUI.generate( firstWin );

        this.fire( "PreRender", [] );

        // Add any background images defined by the AutoGUI
        this.renderBackgroundImages();

        var topWin = term.windows.length - 1;
        var windows = term.windows;
        for ( var j = firstWin, win; win = windows[ j ]; j++ )
        {
            if ( j > 0 )
            {
                // A 5250 window has top/bottom borders of 1 character and 
                // Left/right borders of 3 characters (blank either side of 
                // the line). The position supplied in the transaction is the
                // position of the blank character to the left of the border.
                // The height and width supplied in the transaction are the
                // size of the inner area of the window.

                // For compatibility with TS1, we need to make the window div this full size.

                container = document.createElement( "div" );
                container.setAttribute( "id", win.id );
                container.className = win.pulldown ? "ax-terminal-pulldown" : "ax-terminal-window ui-corner-all";

                var outerDimen =
                {
                    top   : win.row,
                    left  : win.col,
                    width : win.width + 6,
                    height: win.height + 2
                };
                var innerDimen =
                {
                    top   : 1.5,
                    left  : 2,
                    width : win.width + 4,
                    height: win.height + 1
                };
                xOffset = win.col - 1;
                yOffset = win.row - 1;

                var winBackground = document.createElement( "div" );
                winBackground.className = "ax-terminal-window-background ui-corner-all";
                if ( win.header )
                {
                    var header             = document.createElement( "div" );
                    header.className       = "ax-terminal-window-header ui-corner-top " + win.header.attribute;
                    header.style.textAlign = win.header.align;
                    header.innerHTML       = "<div>" + win.header.text + "</div>";
                    winBackground.appendChild( header );
                    innerDimen.height     += 0.6;
                    outerDimen.top        -= 0.6;
                    outerDimen.height     += 0.6;
                    yOffset               -= 0.6;
                }
                if ( win.footer )
                {
                    var footer             = document.createElement( "div" );
                    footer.className       = "ax-terminal-window-footer ui-corner-bottom " + win.footer.attribute;
                    footer.style.textAlign = win.footer.align;
                    footer.innerHTML       = "<div>" + win.footer.text + "</div>";
                    winBackground.appendChild( footer );
                    innerDimen.height     += 0.6;
                    outerDimen.height     += 0.6;
                }

                var dimen = this.characterDimensionsToPixels( outerDimen, true, false, true );
                var st = container.style;
                st.zIndex = 100 * j;
                for ( var p in dimen )
                    st[ p ] = dimen[ p ];
                container.appendChild( winBackground );

                if ( j == topWin && win.capCursor )
                {
                    // This is the top-most "window". Place an overlay below it to dim the background
                    // and prevent interaction.
                    var overlay = document.createElement( "div" );
                    overlay.className = "ui-widget-overlay";
                    overlay.style[ "zIndex" ] = ( 100 * j ) - 1;
                    terminalContainer.appendChild( overlay );
                }

                terminalContainer.appendChild( container );

                dimen = this.characterDimensionsToPixels( innerDimen, false, false, false );
                var $w = $( winBackground );
                dimen.left -= ( $w.outerWidth() - $w.innerWidth() ) / 2;
                dimen.top -= ( $w.outerHeight() - $w.innerHeight() ) / 2;
                st = winBackground.style;
                for ( var p in dimen )
                    st[ p ] = dimen[ p ] + "px";

                if ( this.moveableWindows && !win.pulldown )
                    $( container ).draggable(
                        {
                            containment: terminalContainer,
                            iframefix  : true
                        }
                    );
            }

            if ( j == topWin )
            {
                // Initialise "windowStart" screen extensions
                var exArrayCopy = [];
                for ( var i = 0, ex; ex = this.extensions[ i ]; i++ )
                {
                    exArrayCopy.push( ex );
                    if ( ex.renderStage == "windowStart" )
                    {
                        ex._field = null;
                        ex.init( container );
                        ex.render( container );
                    }
                }
                $( container ).data( "extensions", exArrayCopy );
            }

            var curRow       = 0;
            var curRowAdjust = 0;
            var yAdjust      = 0;
            var fields       = win.fields;
            var newEl, pos;

            for ( var i = 0, fld; fld = fields[ i ]; i++ )
            {
                // Skip fields with a byp flag.
                if ( fld.byp )
                    continue;

                // Check if any HTMLKW fields in the previous row have forced
                // everything else to be pushed down.
                if ( fld.row > curRow )
                {
                    yAdjust     += curRowAdjust;
                    curRow       = fld.row;
                    curRowAdjust = 0;
                }

                newEl = document.createElement( "div" );
                newEl.setAttribute( "id", fld.id );
                newEl.className = "ax-terminal-field " + ( fld.attribute ? fld.attribute : "" );
                pos = this.characterDimensionsToPixels(
                    {
                        top : fld.row - yOffset,
                        left: fld.col - xOffset
                    },
                    false,
                    true,
                    ( j == 0 )
                );
                newEl.style.top  = ( pos.top + yAdjust ) + "px";
                newEl.style.left = pos.left + "px";
                if ( fld instanceof AXES.Terminal.Field.HtmlField )
                    newEl.style.width = ( ( AXES.Terminal.cols - fld.col ) * cW ) + "px";
                //fix highlight selection of autogui fkey buttons
                if ( fld.extensions.length > 2 )
                    newEl.style.width = ( fld.width  * cW ) + "px";

                fld.terminalDiv = newEl;

                this.drawField( fld );

                if ( !focusFld && j == topWin && term.cursor.isInField( fld ) )
                    focusFld = fld;

                container.appendChild( newEl );

                fld.addedToDOM();

                // Track the tallest HTMLKW field in each row and, if it is taller than
                // one character, push all fields on subsequent rows down accordingly.
                if ( fld instanceof AXES.Terminal.Field.HtmlField )
                {
                    var h = $( newEl ).outerHeight();
                    if ( h > cH )
                    {
                        var newAdj = h - cH;
                        if ( newAdj > curRowAdjust )
                            curRowAdjust = newAdj;
                    }
                }
            }
            if ( j == topWin )
            {
                for ( var i = 0, fld; fld = this.userFields[ i ]; i++ )
                {
                    var newEl = document.createElement( "div" );
                    newEl.setAttribute( "id", fld.id );
                    newEl.className = "ax-terminal-field";
                    fld.terminalDiv = newEl;
                    this.drawField( fld );
                    container.appendChild( newEl );
                    fld.addedToDOM();
                }

                // Initialise "windowEnd" screen extensions
                for ( var i = 0, ex; ex = this.extensions[ i ]; i++ )
                    if ( ex.renderStage == "windowEnd" )
                    {
                        ex._field = null;
                        ex.init( container );
                        ex.render( container );
                    }
            }
        }

        if ( !this._isCustomized || !this.Cursor._limitedCursorSupport )
            this.Cursor.hide();
        else this.Cursor.show();

        this._layoutCleared = false;
        this.fire( "RenderComplete", [] );
    }

    this.customTabOrder.sort(
        function( a, b )
        {
            var retVal = a.tabIndex - b.tabIndex;
            if ( retVal == 0 )
                retVal = a.createdIndex - b.createdIndex;
            return retVal;
        }
    );

    var oldZoom = this._zoom;
    this.applyZoom();
    if ( oldZoom != this._zoom )
        this.fire( "ZoomChanged" );
    this.displayMessages();

    if (AXES.Dev && AXES.Dev.screenLocked) return;

    // focusFld is currently the field under the cursor. Now check if the aXes developer
    // has overridden this
    var firstElToFocus = this.getExtensionProperty( "firstElementToFocus" );
    if ( firstElToFocus && firstElToFocus.v != "default" && this.customTabOrder.length )
        focusFld = this.customTabOrder[ 0 ].field;

    if ( focusFld )
        window.setTimeout(
            function()
            {
                focusFld.setFocus();
            },
            10
        );
}

/**
 * Display a message in the terminal status bar. Any existing messages are removed.
 * @param {String} theMsg The message to display
 * @param {Boolean} clearFirst If true, any existing messages are cleared.
 * @param {Boolean} playAlert If true, an alert sound is played
 */
AXES.Layout.displayMessage = function(theMsg, clearFirst, playAlert)
{
    if (clearFirst) this._messages.length = 0;
    if (theMsg) this._messages.push({
        text: theMsg,
        alert: playAlert
    });
    if (!this.isBusy()) 
    {
        var sel = $("#terminalMessages").get(0);
        if (sel && sel.options) 
        {
            if (clearFirst) sel.options.length = 0;
            if (theMsg) 
            {
                var opt = document.createElement("option");
                opt.appendChild(document.createTextNode(theMsg));
                if (sel.options.length == 0) 
                {
                    sel.appendChild(opt);
                }
                else 
                {
                    sel.insertBefore(opt, sel.options[0]);
                }
                sel.selectedIndex = 0;
                if (playAlert) this.playAlert();
            }
            sel.disabled = (sel.options.length == 0);
        }
    }
}
AXES.Layout.displayMessages = function()
{
    // Combine error messages from the terminal with any messages added to the layout
    if (this._type == 0) return;
    var playAlert = false;
    var sel = $("#terminalMessages").get(0);
    if (sel && sel.options) 
    {
        sel.options.length = 0;
        var msgs = AXES.Terminal.messages;
        var l = msgs.length;
        var opt;
        for (var i = 0; i < l; i++) 
        {
            opt = document.createElement("option");
            opt.appendChild(document.createTextNode(msgs[i]));
            sel.appendChild(opt);
            playAlert = true;
        }
        l = this._messages.length;
        for (var i=0, msg; msg=this._messages[i]; i++) 
        {
            opt = document.createElement("option");
            opt.appendChild(document.createTextNode(msg.text));
            sel.appendChild(opt);
            playAlert = playAlert || msg.alert;
        }
        
        sel.selectedIndex = 0;
        sel.disabled = (sel.options.length == 0);
        if (playAlert) this.playAlert();
    }
}
AXES.Layout.onPostFailed = function(aidKey, reason) {
    this.displayMessage(reason, true, true);
    //TODO. Axes 1.3 displayed the error in a dialog and offered the
    //user the option to post anyway. Not sure if this is really a good
    //idea but, if we want to impliment it, this is where we would do it
    //(Terminal.doPost will need an option to ignore errors too)
}
/**
 * Renders a field eXtension
 * @param {Object} fld The field being drawn
 * @param {Object} ex The eXtension being drawn
 */
AXES.Layout.drawFieldExtension = function(fld, ex) {
    ex._field = fld;
    try 
    {
        ex.init(fld, fld.terminalDiv);
        ex.render(fld, fld.terminalDiv);
    } 
    catch (e) 
    {
        var fldName = (fld.symbolicName) ? fld.symbolicName : fld.id;
        var msg = TEXT.Dev("JavaScript error initialising eXtension %s1 on field %s2: %s3.", ex.name, fldName, e.message);
        AXES.SysTrace.error(msg);
    }
}
/**
 * Creates and renders all field extensions
 * @param {Object} fld
 */
AXES.Layout.drawField = function(fld)
{
    $(fld.terminalDiv).data("TermField", fld);
    for (var i = 0, ex; ex = fld.extensions[i]; i++) 
    {
        this.drawFieldExtension(fld, ex);
    }
}
/**
 * Removes and re-draws a field contents
 * @param {Object} fld
 */
AXES.Layout.redrawField = function(fld) {
    for (var i = 0, ex; ex = fld.extensions[i]; i++) 
    {
        ex.destroy();
    }
    $(fld.terminalDiv).empty();
    this.drawField(fld);
}

/**
 * Clears the layout. Calls the destroy() method on each extension
 * before deleting it.
 */
AXES.Layout.clearLayout = function()
{
    
    // Remove all remaining DOM elements.
    if (this._type == 0) 
    {
        $("#loadingIndicator").remove();
    }
    else if (this._type == 2) 
    {
        $("#axCustomPageCSS").attr("href", "");
        this.displayArea.empty();
        this.displayArea.append(this._terminalWrapper);
    }
    else 
    {
        var self = this;
        this._terminalContent.find("div.ax-terminal-window, div.ax-terminal-pulldown").each(function()
        {
            self.clearWindow(this);
        }).remove();
        this.clearWindow(this._terminalContent);
        this._terminalContent.empty();
    }
    
    this.displayArea.addClass("axdefault");
    this.extensions.length = 0;
    this.userFields.length = 0;
    this.customTabOrder = [];
    this._isCustomized = false;
    
    this.clearMessages();
    this._type = 0;
    this._layoutCleared = true;
}
AXES.Layout.clearWindow = function(windowDiv)
{
    // Remove all field eXtensions
    var $w = $(windowDiv);
    $w.find(".ax-terminal-field").each(function(e)
    {
        var fld = $(this).data("TermField");
        if (fld) 
        {
            for (var j = 0, ex; ex = fld.extensions[j]; j++) 
            {
                try 
                {
                    if (ex.destroy) ex.destroy(fld, fld.terminalDiv);
                } catch(e) {
                    AXES.SysTrace.error(TEXT.Dev("JavaScript error destroying eXtension %s1: %s2.", ex.name, e.message));
                }
                delete ex._field;
                fld.extensions[j] = null;
            }
            delete fld.terminalDiv;
            fld.extensions.length = 0;
        }
        
    }).remove();
    var exArray = $w.data("extensions");
    if (exArray) 
    {
        for (var i = 0, ex; ex = exArray[i]; i++) 
        {
            try 
            {
                if (ex.destroy) ex.destroy(this);
            } catch(e) {
                AXES.SysTrace.error(TEXT.Dev("JavaScript error destroying eXtension %s1: %s2.", ex.name, e.message));
            }
        }
    }
}
AXES.Layout.clearMessages = function() {
    var sel= $("#terminalMessages").get(0);
    if (sel && sel.options) sel.options.length = 0;
    sel.disabled = true;
    this._messages.length = 0;    
}
AXES.Layout.setBusy = function(busyOn) 
{
    if (this._hostBusyDialog) {
        this._hostBusyDialog.dialog('close').dialog('destroy');
        delete this._hostBusyDialog;
    }
    if (busyOn) 
    {
        var $cover = $("#terminalLockedCover");
        if ($cover.length == 0) 
        {
            $("#terminalStatusIndicator").addClass("status-busy");
            $cover = $(document.createElement("div")).attr("id", "terminalLockedCover");
            this.displayArea.append($cover);
            $cover.show();
        }    
    } else {
        $("#terminalLockedCover").remove();
        $("#terminalStatusIndicator").removeClass("status-busy");
    }
}
AXES.Layout.isBusy = function() 
{
    return $("#terminalStatusIndicator").hasClass("status-busy");
}
/**
 * Indicates if the current screen has been customized.
 * @returns {Boolean}
 */
AXES.Layout.isCustomized = function() {
    return this._isCustomized;
}
/**
 * Creates an HTML element. This is exactly the same as calling the DOM method document.createElement.
 * It should be used in TS1 to ensure the correct document object is used to create elements that
 * will be added to the terminal area. It is provided in TS2 for backwards compatability.
 * @param {String} element The tag name of the element to create.
 * @returns {HTMLElement} The new element.
 */
AXES.Layout.createUserElement = function(element)
{
    if (AXES.Utils.isString(element)) 
    {
        return document.createElement(element.toLowerCase());
    }
    return null;
}
/**
 * Adds an HTML element to the terminal area. Optionally creates the element as well so you don't need to
 * call createUserElement. Generally you should use createUserElement first. Performance is much better
 * if you do as much manipulation of element properties and contents as possible before adding them
 * to the DOM. Elements do not have a size until after they have been added to the DOM so anything
 * that requires querying an elements size should be done after it has been attached (or avoided if possible).
 * @param {String|HTMLElement} element Either a string containing a tag name or an HTML element object.
 * @returns {HTMLElement} The element added to the terminal area.
 */
AXES.Layout.attachUserElement = function(element)
{
    if (!element) return null;
    if (AXES.Utils.isString(element)) 
    {
        element = document.createElement(element.toLowerCase());
    }
    if (element && element.nodeType && element.nodeType == 1) 
    {
        if (element.style) element.style.position = "absolute";
        this._terminalContent.append(element);
        //if (jQuery.inArray(element, this._customUserElements) == -1) this._customUserElements.push(element);
        return element;
    }
    return null;
}
/**
 * Removes an element from the terminal area. You should use this method instead of the DOM removeChild method in order to minimise the risk of memory leaks.
 * @param {HTMLElement} element
 */
AXES.Layout.removeUserElement = function(element) 
{
    if (element && element.nodeType) 
    {
        //var i = jQuery.inArray(element, this._customUserElements);
        //if (i > -1) 
        //{
        //    this._customUserElements.splice(i, 1);
        //}
        AXES.Utils.discardElement(element);
    }
}

/**
 * 
 * @namespace
 */
AXES.Layout.Cursor = {
    _isShowing: false,
    _limitedCursorSupport: false,
    _win : AXES.Terminal.windows,
    _xOffset: 0,
    _yOffset: 0,
    _calculateOffset: function ()
    {
        var topWin = this._win[this._win.length - 1];
        if ( this._win.length > 1 ) {            
            this._xOffset = topWin.col - 1;
            this._yOffset = topWin.row - 1;
            if ( topWin.header ) this._yOffset -= 0.6;
        }
    },
    init: function(limitedCursorSupport)
    {
        this._limitedCursorSupport = limitedCursorSupport;
    },
    show: function()
    {
        if (!AXES.Layout._isCustomized || !this._limitedCursorSupport)
            return;
        
        var winId = this._win.length > 1 ?
            "#" + this._win[this._win.length - 1].id :
            "#terminalContent";         
        var cursorDiv = $("#axcursor");
        if (cursorDiv.length == 0) {  
            cursorDiv = document.createElement("div");
            cursorDiv.id = "axcursor";
        }
        if (this._limitedCursorSupport) {
            cursorDiv.className = "Cursor";
        } else {
            cursorDiv.className = "CursorNav";
        }
        this._calculateOffset();
        $(cursorDiv).css(AXES.Layout.characterDimensionsToPixels({
                top: AXES.Terminal.cursor.row - this._yOffset,
                left: AXES.Terminal.cursor.col - this._xOffset,
                width: 1,
                height: 1
            }, true,false)).prependTo(winId);
        $(winId).bind("click", this, this._handleClick);
        this._isShowing = true;
    },
    hide: function()
    {
        if (!this._isShowing) return;
        
        this._isShowing = false;
        var winId = AXES.Terminal.windows.length > 1 ?
            "#" + AXES.Terminal.windows[ AXES.Terminal.windows.length - 1 ].id :
            "#terminalContent";
        $(winId).unbind("click", this._handleClick);
        if ( $("#axcursor").hasClass("Cursor") ) 
            $("#axcursor").removeClass("Cursor");
        $("#axcursor").remove();
    },
    update: function()
    {
        if (!this._isShowing) return;
        
        this._calculateOffset();
        var pos = AXES.Layout.characterDimensionsToPixels({
            top: AXES.Terminal.cursor.row - this._yOffset,
            left: AXES.Terminal.cursor.col - this._xOffset
        }, true);
        var $c = $("#axcursor").css(pos);
        if ($c.length == 0) this.show(); // Cursor isn't there anymore
    },
    checkFieldAtPosition: function(position, e)
    {
        var fld = AXES.Terminal.getFieldAtPosition(position);
        var afld = AXES.Layout.getActiveField();
        if ( (fld!==null) && (afld !== null)) {
            if (fld !== afld) {
                fld.setFocus(position);
            } else {
                var element = e.target;
                if ((afld.type == "IField") && (element.tagName=="INPUT" && element.type=="text")) {
                    var sel = AXES.Layout.getFieldSelection(element);                    
                    position.col = afld.col + sel.end;
                    AXES.Terminal.cursor.set(position);
                } 
            }
        }
    },
    _handleClick: function(e)
    {
        if (!AXES.Layout._isCustomized || !e.data._limitedCursorSupport )
            return;
        
        var winId = e.data._win.length > 1 ?
                    "#" + e.data._win[this._win.length - 1].id :
                    "#terminalContent";
        var l = AXES.Layout;
        var cW = l.characterSize.width + l.spaceBetweenCols;
        var cH = l.characterSize.height + l.spaceBetweenRows;
        var terminalPos = $(winId).offset();        
        var col = Math.ceil((((e.pageX - terminalPos.left) / l._zoom) - l.horizPad) / cW );
        var row = Math.ceil((((e.pageY - terminalPos.top) / l._zoom) - l.vertPad) / cH);
        e.data._calculateOffset();
        AXES.Terminal.cursor.set({
            row: row - e.data._yOffset,
            col: col - e.data._xOffset
        });
    },
    _handleKeyDown: function(e)
    {
        // This is called by the function handling keyboard events at the document level
        if (!AXES.Layout._isCustomized) return;
        
        if (!this._isShowing && !this._limitedCursorSupport ) 
            return;
        
        var cursor = AXES.Terminal.cursor;
        var newPos = {
            row: cursor.row,
            col: cursor.col
        };
        var kbn = AXES.Keyboard.keysByName;
        switch (e.which)
        {
            case (kbn.Home):
                // set cursor to home cursor
                newPos = AXES.Terminal.cursorHome;
                break;
            case (kbn.LeftArrow):
                newPos.col--;
                break;
            case (kbn.RightArrow):
                newPos.col++;
                break;
            case (kbn.UpArrow):
                newPos.row--;
                break;
            case (kbn.DownArrow):
                newPos.row++;
                break;
        }
        cursor.set(newPos);
        this.checkFieldAtPosition(newPos, e);
    }
}

AXES.Layout.UserField = AXES.Terminal.Field.Base.extend(
/**
* @lends AXES.Layout.UserField.prototype
*/
{
    /** 
     * Represents an user defined field
     * @constructs
     * @augments AXES.Terminal.Field.Base
     * @param {String} id The id of the new field
     */
    init: function(id)
    {
        this._super(null);
        this.id = id;
        this.type = AXES.Terminal.Field.Types.OUTPUT;
        this.isInput = false;
    },
    // Override a few methods
    adjustedId: function() {return this.id;},
    isIndexable: function() {return false;},
    ddsReset: function(resetMdt) {},
    setCursor: function(offset) {},
    getSize: function()
    {
        var $td = $(this.terminalDiv);
        return {
            width: $td.length ? $td.width() : 0,
            height: $td.length ? $td.height() : 0
        };
    }
});

AXES.Layout.addUserField = function()
{
    var windows = AXES.Terminal.windows, win = windows[windows.length-1], fld, newEl, i, newId = 1;
    for (i=0; fld=AXES.Layout.userFields[i]; i++) {
        newId = Math.max(newId, parseInt(fld.id) + 1);
    }
    fld = new AXES.Layout.UserField(newId.toString());
    AXES.Layout.userFields.push(fld);
    fld.extensions.push(AXES.Extensions.createNewExtension("axuex", {"style":{"v":{"left":"20px","top":"20px","width":"100px","height":"30px"}}}, null, true));
    newEl = document.createElement("div");
    newEl.setAttribute("id", fld.id);
    newEl.className = "ax-terminal-field";
    fld.terminalDiv = newEl;
    AXES.Layout.drawField(fld);
    return fld;
}

/**
 * Handles AutoGUI generation
 * @namespace
 */
AXES.Layout.AutoGUI = {
    autoGUIOn: true,
    enabled: true,
    rules: {},
    backgroundImages: [],
    _excludeFromSubfileProcessing: {},
    _fKeyButtons: [],
    _subfileOptionMenu: [],
    _notifyInitDone: true,
    init: function()
    {
        AXES.registerInitTask("LoadAutoGUIRules");
        this.loadRules(false);
    },
    _loadDone: function() {
        if (this._notifyInitDone) {
            AXES.initTaskCompleted("LoadAutoGUIRules");
            this._notifyInitDone = false;
        }
    },
    /**
     * Turns AutoGUI mode on or off.
     * @param {Boolean} autoGUIOn The new state of the autoGUI setting
     * @param {Boolean} [refreshLayout=true] Indicates whether or not to redraw the display
     */
    setAutoGUI: function(autoGUIOn, refreshLayout)
    {
        if (!this.enabled) return;
        if (refreshLayout === undefined) refreshLayout = true;
        if (this.autoGUIOn != autoGUIOn) 
        {
            this.autoGUIOn = autoGUIOn;
            if (refreshLayout) 
            {
                for (var i = 0, fld; fld = AXES.Terminal.fields[i]; i++) 
                {
                    fld.screenUnloaded();
                }
                AXES.Layout.clearLayout();
                AXES.Layout.doLayout();
            }
            AXES.UI.MenuBar.refreshMenus();
        }
    },
    _loadRules: function(filesToTry, refreshLayout) 
    {
        var self = this, fileInfo = filesToTry.pop();
        jQuery.ajax({
            url: fileInfo.url,
            cache: !AXES.devMode,
            dataType: fileInfo.type,
            success: function(data, textStatus)
            {
                if (fileInfo.type == "xml") 
                {
                    self.processXMLRules(data);
                    self.processRules(refreshLayout);
                } else 
                {
                    self.rules =  data;
                    self.processRules(refreshLayout);
                }
                self._loadDone();
            },
            error: function(XMLHTTPRequest, textStatus, errorThrown)
            {
                var msg;
                if (errorThrown) 
                {
                    msg = TEXT.Core("Error retrieving AutoGUI rules: %s1", errorThrown);
                } else 
                {
                    msg = TEXT.Core("HTTP error %s1 retrieving AutoGUI rules: %s2", XMLHttpRequest.status, XMLHttpRequest.statusText);
                }
                AXES.SysTrace.error(msg);
                if (filesToTry.length) self._loadRules(filesToTry, refreshLayout);
                else self._loadDone();
            }
        });
    },
    loadRules: function(refreshLayout)
    {
        var bp="/ts/", ds=AXES.definitionSet, filesToTry = [];
        filesToTry.push({url:bp + "rules.xml", type:"xml"});
        bp += "screens/";
        filesToTry.push({url:bp + "axautoguirules.json", type:"json"});
        if (ds) filesToTry.push({url:bp + ds + "/axautoguirules.json", type:"json"});
        
        this._loadRules(filesToTry, refreshLayout);
    },
    processXMLRules: function(xmlData)
    {
        this.rules = {};
        var val, rule;
        var ruleNodes = $("RULE", xmlData);
        var l = ruleNodes.length;
        for (var i = 0; i < l; i++) 
        {
            rule = $(ruleNodes[i]);
            val = undefined;
            switch (rule.attr("type")) {
                case "checkbox":
                    val = (rule.attr("value").toLowerCase() == "true");
                    break;
                case "numeric":
                    val = Number(rule.attr("value"));
                    if (val === NaN) {
                        val = 0
                    }
                    break;
                default:
                    val = rule.attr("value");
            }
            this.rules[rule.attr("id")] = val;
        }
    },
    processRules: function(refreshLayout)
    {
        var a = [];
        this.regExps = {};
        
        // Build the Regular Expression for Detecting function keys
        var keyTypes = [];
        if (this.rules.FKeyRule) keyTypes.push("F");
        if (this.rules.PFKeyRule) keyTypes.push("PF");
        if (this.rules.FPKeyRule) keyTypes.push("FP");
        if (this.rules.CFKeyRule) keyTypes.push("CF");
        
        var defaultFKeyREStr = "(" + keyTypes.join("|") + ")\\d{1,2} *=";
        var customFKeyREStr = this.rules.KeyCustomRule ? this.rules.KeyCustomRule.trim() : "";
        var fkeyREStr;
        
        if (defaultFKeyREStr.length && customFKeyREStr.length) 
        {
            fkeyREStr = "(" + defaultFKeyREStr + ")|(" + customFKeyREStr + ")";
        } else if (customFKeyREStr.length) 
        {
            fkeyREStr = customFKeyREStr;
        } else 
        {
            fkeyREStr = defaultFKeyREStr;
        }
        this.regExps.fKey = null;
        if (fkeyREStr.length) 
        {
            try 
            {
                this.regExps.fKey = new RegExp(fkeyREStr, "g");
            } catch (e) 
            {
                AXES.Trace.error(TEXT.Core("DefaultMessage.AXS8861"));
                this.regExps.fKey = defaultFKeyREStr.length ? new RegExp(defaultFKeyREStr, "g") : null;
            }
        }
        
        // Build the Regular Expression for Detecting menus;
        
        // Looking for [number.text]. Require space after dot if spaceRequired is true, 
        // otherwise don't match on [number.number].
        this.regExps.menu = this.rules["MenuSpaceRule"] ? /^( *\d+)\. {1,2}(.*)$/ : /^( *\d+)\.(?!\d+)(.*)$/;
        
        // Build the Regular Expressions for detecting page up/down indicators
        try 
        {
            this.regExps.upDownButtons = null;
            this.regExps.upButton = null;
            var reSplit = /\, */;
            a.length = 0;
            if (this.rules.UpDnRule.length) a = this.rules.UpDnRule.split(reSplit);
            this.regExps.upDownButtons = (a.length) ? new RegExp("^\\s*(" + a.join("|") + ")\\s*$", "i") : null;
            
            a.length = 0;
            if (this.rules.UpRule.length) a = this.rules.UpRule.split(reSplit);
            this.regExps.upButton = (a.length) ? new RegExp("^\\s*(" + a.join("|") + ")\\s*$", "i") : null;
        } catch (e) 
        {
            AXES.Trace.error(TEXT.Core("DefaultMessage.AXS8863"));
            this.regExps.upDownButtons = null;
            this.regExps.upButton = null;
        }
            
        // Build the regular expression for detecting subfile options
        // A full RegExp, including custom delimiter looks like this:
        // (\d+|[a-z]) *(?:=|:)(.*?)(?=(?:\b(?:\d+|[a-z]) *(?:=|:)| \.\.\.$|$))( \.\.\.)?
        // Group 1: \b(\d+|[a-z]) - One or more digits or a
        // single letter followed by zero or more spaces and a delimiter.
        // Group 2: ( *(?:=|:) - Some optional spaces followed by a delimiter.
        // Group 3: (.*?) - All the following text up to the next match. The lazy star makes it
        // stop at the first "next match".  The "next match" is defined by the "look ahead 
        // assertion" (?=...) - means match this but don't include in the match.
        // The look ahead matches 3 possibles: The code=text pattern we've just used, three
        // dots preceeded by a word boundary (more indicator) or the end of the string.
        // Group 4: ( \.\.\.)? - Optional more indicator.
        // There is some optional text before the first group. This is not built into the
        // RexExp because Firefox and Google do not handle it properly after the first match.
        this.regExps.subfileOption = null;
        var codePart = "";
        a.length = 0;
        if (this.rules.NumOptRule) a.push("\\d+");
        if (this.rules.CharOptRule) a.push("[a-z]");
        if (a.length) 
        {
            var codePart = a.join("|");
            var customDelimiter = this.rules.SubOptCustomRule.trim();
            var delimiter = (customDelimiter.length) ? "=|" + customDelimiter : "=";
            
            var subOptREStr = "(%1)( *(?:%2))(.*?)(?=(?:\\b(?:%1) *(?:%2)| \\.\\.\\.$|$))( \\.\\.\\.)?";
            
            try 
            {
                this.regExps.subfileOption = new RegExp(AXES.Utils.strargs(subOptREStr,[codePart,delimiter]), "gi");
            } catch (e) 
            {
                AXES.Trace.error(TEXT.Core("DefaultMessage.AXS8862"));
                this.regExps.subfileOption = new RegExp(AXES.Utils.strargs(subOptREStr,[codePart,"="]), "gi");
            }
        }
        
        this.prepareBackgroundImages();
        
        if (refreshLayout) AXES.Layout.doLayout();
    },
    prepareBackgroundImages: function(recursionCount)
    {
        // Create the background images
        // Only create an image if its class exists in the theme stylesheet
        this.backgroundImages.length = 0;
        recursionCount = recursionCount || 1;
        if (recursionCount > 10) return;
        
        var bgImagesStr = this.rules["BackgroundImage"];
        if (bgImagesStr) 
        {
            // Get the rules array from the CSS
            var linkTag = $("#axThemeCSS");
            try 
            {
                var sheet = (linkTag.length) ? linkTag[0].sheet || linkTag[0].styleSheet : null;
                var rules = sheet.cssRules ? sheet.cssRules : sheet.rules;
            } catch (e) 
            {
            
            } finally 
            {
                if (!rules || (sheet && sheet.owningElement && sheet.owningElement.readyState != "complete")) 
                {
                    // CSS may still be loading. Try again after a timeout
                    window.setTimeout(function()
                    {
                        AXES.Layout.AutoGUI.prepareBackgroundImages(recursionCount + 1);
                    }, 200);
                    return;
                }
            }
            
            $("#terminalInner .ax-terminal-backgroundImg").remove();
            this.backgroundImages.length = 0;
            var reImg = /\{src:(.*?);id:(.*?)\}/gi;
            var matchArray = reImg.exec(bgImagesStr);
            while (matchArray) 
            {
                if (matchArray[1] && matchArray[2]) 
                {
                    var cssSelector = "." + matchArray[2].toLowerCase(), lr = rules.length;
                    for (var j = 0; j < lr; j++) 
                    {
                        var rule = rules[j];
                        if (rule.selectorText.toLowerCase() == cssSelector) 
                        {
                            var bgImg = document.createElement("img");
                            bgImg.id = "BackgroundImage" + this.backgroundImages.length;
                            bgImg.className = matchArray[2] + " ax-terminal-backgroundImg";
                            bgImg.src = AXESBUILD.updateURL(matchArray[1]);
                            this.backgroundImages.push(bgImg);
                            break;
                        }
                    }
                }
                matchArray = reImg.exec(bgImagesStr);
            }
            AXES.Layout.renderBackgroundImages();
        }
    },
    generate: function( firstWin )
    {
        firstWin                           = firstWin || 0;
        this._excludeFromSubfileProcessing = {};
        this._fKeyButtons                  = [];
        this._subfileOptionMenu            = [];

        var fields;
        var term    = AXES.Terminal;
        var windows = term.windows;

        for ( var i = firstWin, win; win = windows[ i ]; i++ )
        {
            fields = win.fields;
            for ( var j = 0, fld; fld = fields[ j ]; j++ ) 
            {
                if ( fld.extensions.length )
                    continue;

                // All fields have a base properties eXtension. If not already added, add it.
                var baseExType = fld instanceof AXES.Layout.UserField ? "axuex" : "axex";
                AXES.Extensions.createNewExtension( baseExType, {}, fld, true );
            }

            for ( var j = 0, fld; fld = fields[ j ]; j++ )
            {
                if ( this.autoGUIOn &&
                    fld.extensions.length == 1 /* Only process fields that don't already have eXtensions */ &&
                    !fld.isHidden() /* Only process visible fields */ )
                {
                    var nextField;
                    try
                    {
                        nextField = term.sortedFields[ ( i ? windows[ 0 ].length : 0 ) + j + 1 ];
                    }
                    catch ( e )
                    {
                        nextField = null;
                    }

                    // Make sure nextField is visible and immediately following current field
                    if ( !nextField || nextField.row != fld.row || nextField.col != ( fld.col + fld.width + 1 ) || nextField.isHidden() )
                        nextField = null;

                    if ( fld.type == AXES.Terminal.Field.Types.OUTPUT &&
                        fld.getValue().length >= 1  /* Only process fields containing more than one character */ )
                    {
                        if ( this.checkForMenuItem( fld, nextField ) ||
                            this.checkForFKey( fld, nextField ) ||
                            this.checkForSubfileOption( fld ) ||
                            this.checkForPageUpDown( fld ) ||
                            this.checkForHyperlink( fld ) )
                            continue;

                        var prevField;
                        try
                        {
                            prevField = j > 0 ? fields[ j - 1 ] : null;
                        }
                        catch ( e )
                        {
                            prevField = null;
                        }

                        if ( prevField && ( prevField.row != fld.row ) )
                            prevField = null;

                        if ( this.checkForPromptButton( fld, prevField ) )
                            continue;
                    }
                }

                if ( fld.extensions.length == 1 )
                    // Add the default extension
                    AXES.Extensions.createNewExtension( "axdv", {}, fld, true );
            }
        }

        if ( this.autoGUIOn )
        {
            this.checkForSubfiles();
            this.checkForFKeyPanel();
            this.checkForSubfileOptionMenu();
        }
    },
    checkForSubfiles: function()
    {
        var ruleOn = this.rules["SubFileGUI"], osr = this._overrideSubfileRule;
        if (osr && (osr[AXES.Terminal.symbolicName] != undefined)) ruleOn = osr[AXES.Terminal.symbolicName];
        if (!ruleOn) return false;
        
        var cW = AXES.Layout.characterSize.width, cH = AXES.Layout.characterSize.height, isFirstRF = true;
        for (var i = 0, rf; rf = AXES.Terminal.recordFormats[i]; i++) 
        {
            if (rf instanceof AXES.Terminal.RecordFormat.Subfile) 
            {
                // Check is there is already a Subfile Enhancer already
                // applied to this recordformat.
                for (var j=0, ex; ex = AXES.Layout.extensions[j]; j++) 
                {
                    if (ex.name == "axSubfileEnhancer") 
                    {
                        var sfName = ex.getPropertyValue("subfileName");
                        if ((sfName == rf.name) || ((sfName == "FIRST") && isFirstRF)) 
                        {
                            // This subfile already has a Subfile Enhancer applied to it
                            return false;
                        }
                    }
                }
                isFirstRF = false;
                
                if (this._excludeFromSubfileProcessing[rf.name]) continue;
                
                if ( (!this.rules["SubFileAllRowsRule"] || this.rules["SubFileAllRowsRule"] != "true") &&
                    rf.startLine < this.rules["SubFileStartRowRule"] &&
                    (rf.endLine) > this.rules["SubFileEndRowRule"] ) return false;
                
                props = {
                        subfileName: {v: rf.name},
                        headerHeightRows: {v: rf.rowsInHeader},
                        columnFirst: {v: 0},
                        columnLast: {v: 0},
                        highlightRows: {v: true}
                    }
                AXES.Extensions.createNewExtension("axSubfileEnhancer", props, AXES.Layout, true);
            }
        }
    },
    _fieldIsInRuleArea: function(field, rulePrefix) {    
        if (!this.rules[rulePrefix + "AllRowsRule"] || this.rules[rulePrefix + "AllRowsRule"] != "true") 
                if ( (field.row < this.rules[rulePrefix + "StartRowRule"]) ||
            ((field.row + field.height - 1) > this.rules[rulePrefix + "EndRowRule"]) )
                  return false;        
        
        if (!this.rules[rulePrefix + "AllColsRule"] || this.rules[rulePrefix + "AllColsRule"] != "true")
            if ( (field.col < this.rules[rulePrefix + "StartColRule"]) ||
                     ((field.col + field.width - 1) > this.rules[rulePrefix + "EndColRule"]) )
            return false;
            
        return true;                
    },
    checkForMenuItem: function(field, nextField)
    {
        if (!this.rules["MenuItemGUI"]) return false;
        if (this.regExps.menu == null) return false;
        
        if (!this._fieldIsInRuleArea(field, "Menu")) return false;

        var nextFieldIncluded = false;
        var testText = field.getValue();
        if (this.rules["MenuCombineRule"] && nextField && /^\s*\d+.$/.test(testText)) 
        {
            testText += " " + nextField.getValue();
            if (nextField.extensions.length > 1) return false;
            nextFieldIncluded = true;
        }
        var matchArray = this.regExps.menu.exec(testText);
        if (matchArray !== null) 
        {
            var props = {
                value: {
                    v: matchArray[1]
                },
                displayText: {
                    v: matchArray[2]
                },
                useImage: {
                    v: this.rules["MenuImgRule"] || false
                },
                imageURL: {
                    v: this.rules["MenuImgPathRule"]
                }
            };
            AXES.Extensions.createNewExtension("ax5250menu", props, field, true);
            if (nextFieldIncluded) AXES.Extensions.createNewExtension("axnoop", {}, nextField, true);
            if (field.recordFormat) 
            {
                this._excludeFromSubfileProcessing[field.recordFormat.name] = true;
            }
            return true;
        }
        return false;
    },
    checkForFKey: function(field, nextField)
    {
        if (!this.rules["FunctionKeyGUI"]) return false;
        if (this.regExps.fKey == null) return false;
        
        if (!this._fieldIsInRuleArea(field, "Key")) return false;
        
        var testText = field.getValue().trim();

        var nextFieldIncluded = false;
        if (nextField && /^(F|PF|FP|CF|Cmd)\d{1,2}$/i.test(testText)) 
        {
            testText = (field.getValue() + " " + nextField.getValue()).trim();
            if (nextField.extensions.length > 1) return false;
            nextFieldIncluded = true;
        }
        
        // Fields matching this pattern can be ignored
        // (Not exactly sure what this is but 1.3.4 does it so reproducing)
        if (/(\.[\s\S]+\,)|(\,[\s\S]+\.)|(^\S*\.$)/.test(testText)) return false;
        
        // Because we don't know how the custom RegExp may be constructed, we can't rely on the RegExp
        // to split the string into the bits we need.  We will assume that it does identify the FKey
        // string from the beginning up to and including the delimiter.
        // 1.3.3 didn't do this. It searched for a delimiter or the first non-digit after the number.
        // This new approach may introduce some backwards compatability issues with custom patterns
        // but it is simpler, offers greater flexibility in defining the delimiter and fixed custom
        // patterns will still work in 1.3.4.
        var re = this.regExps.fKey, matchArray = re.exec(testText), fKeys = [], s, e, p=0, prevIndex=0;
        while (matchArray) 
        {
            if (!this.rules["KeyMatchMutipleRule"] && (fKeys.length > 1)) return false;
            // Any digits found in the matched string are the FKey number
            var fKeyNum = matchArray[0].match(/\d{1,2}/);
            if (fKeyNum === null) break; //No number, can't process as an FKey.
            // Everything from the end of the match to the beginning of the next
            // match (or end of string) is the button text.
            s = re.lastIndex;
            p += AXES.Utils.getDBCSLength(testText.substring(prevIndex,matchArray.index));
            prevIndex = matchArray.index;
            matchArray = re.exec(testText);
            e = matchArray ? matchArray.index : testText.length;
            fKeys.push({
                keyNum: fKeyNum[0],
                keyName: "F" + fKeyNum[0],
                keyText: testText.substring(s, e).trim(),
                keyPosition: p
            });
        }
        
        var ex;
        for (var i = 0, key; key = fKeys[i]; i++) 
        {
            var keyCaption = (this.rules["FKeyOnButton"]? AXES.Utils.strargs("%1=%2", [key.keyName,key.keyText]) : key.keyText);
            var props = {
                onClick: {
                    v: "SENDKEY('" + key.keyName + "');",
                    s: true
                },
                caption: {
                    v: keyCaption
                },
                tooltip: {
                    v: AXES.Utils.strargs("%1=%2", [key.keyName,key.keyText])
                },
                sizeToField: {
                    v: false
                }
            };
            if (i > 0) 
            {
                props.style = {
                    v: {
                        position: "absolute",
                        left: (AXES.Layout.characterSize.width * key.keyPosition) + "px"
                    }
                };
            }
            ex = AXES.Extensions.createNewExtension("axButton", props, field, true);
            this._fKeyButtons.push({key: key.keyName, extension: ex});
        }
        if (fKeys.length && nextFieldIncluded) AXES.Extensions.createNewExtension("axnoop", {}, nextField, true);
        return (fKeys.length > 0);
    },
    checkForHyperlink: function(field)
    {
        if (!this.rules["HyperlinkGUI"]) return false;
        
        if (!this._fieldIsInRuleArea(field, "Hyperlink")) return false;
        
        var testText = field.getValue().trim();
        
        var linkRE = /(.*)\b(?:(http[s]?:\/\/|ftp:\/\/|telnet:\/\/|gopher:\/\/|wais:\/\/|news:\/\/|nntp:\/\/|mailto:|tel:|sms:)(\S+[^\s.,:;]))(.*)/i;
        var mailRE = /(.*?)()([A-Z0-9._%+-]+@[A-Z0-9.-]+\.[A-Z]{2,4})\b(.*)/i;
        var webRE = /(.*)\b()(www\.[A-Z0-9.-]+\.[A-Z]{2,4})\b(.*)/i;
        
        var urlPrefix = "";
        var match = linkRE.exec(testText);
        if (match) urlPrefix = match[2];
        if (!match) 
        {
            urlPrefix = "http://";
            match = webRE.exec(testText);
        }
        if (!match) 
        {
            urlPrefix = "mailto://";
            match = mailRE.exec(testText);
        }

        if (!match)
        {
            return false;
        }

        var props = {
            textBeforeLink: {
                v: match[1]
            },
            textAfterLink: {
                v: match[4]
            },
            caption: {
                v: match[3]
            },
            onClick: {
                v: AXES.Utils.strargs("window.open('%1%2', 'aXes')",[urlPrefix,match[3]]),
                s: true
            },
            style: {v: { "white-space":"pre"}}
        };
        
        AXES.Extensions.createNewExtension("axHyperLink", props, field, true);
        return true;
    },
    checkForPageUpDown: function(field)
    {
        if (!this.rules["PageUpDnGUI"]) return false;
        if (this.regExps.upDownButtons == null) return false;
        if (this.regExps.upButton == null) return false;
        
        if (!this._fieldIsInRuleArea(field, "UpDn")) return false;
        
        var testText = field.getValue().trim();

        var doDown = this.regExps.upDownButtons.test(testText);
        var doUp = doDown || this.regExps.upButton.test(testText);
        
        if (doUp) 
        {
            var props = {
                scrollMarkers: {
                    v: this.rules.UpDnRule.replace(/\\/gi,"").split(/\, */)
                }
            };
            AXES.Extensions.createNewExtension("axSubfileScroller", props, field, true);
        }
        return doUp;
    },
    checkForSubfileOption: function(field)
    {
        if (!this.rules["SubOptGUI"]) return false;
        
        if (this.regExps.subfileOption == null) return false;
        
        if (!this._fieldIsInRuleArea(field, "SubOpt")) return false;
        
        var testText = field.getValue();
        
        var re = this.regExps.subfileOption, match = re.exec(testText), p=0, prevIndex = 0;
        if (!match) return false;
            
        var sfoClickHandler = [
            "var fld = ENV.GETFIELDATCURSOR();",
            "if (fld && fld.isInputField()) {",
            "    fld.setValue('%1');",
            "    AXES.Terminal.doPost('Enter');",
            "}"].join("\n");
            
        var notFirst = false;
        var exAdded = false;
        while (match) 
        {
            if (!notFirst) 
            {
                var textBefore = testText.substring(0, match.index);
                if (textBefore && textBefore.slice(-1) != " ") 
                {
                    // Match must be the beginning of the string or preceded by a space
                    match = re.exec(testText);
                    continue;
                }
            }
            var props = {
                textBeforeLink: {
                    v: notFirst ? "" : textBefore
                },
                caption: {
                    v: (match[1] + match[2] + match[3]).trim()
                },
                style: {v: { "text-decoration":"none", "white-space":"pre"}},
                onClick: {v: AXES.Utils.strargs(sfoClickHandler,[match[1]]), s: true}
            };
            // subfile menu options
            var menuOptions = {
                sendkey: false,
                value  : match[1],
                text   : (match[1] + match[2] + match[3]).trim()
            };
            this._subfileOptionMenu.push(menuOptions);
            if (notFirst > 0) 
            {
                p += AXES.Utils.getDBCSLength(testText.substring(prevIndex,match.index));
                prevIndex = match.index;
                props.style = {
                    v: {
                        "text-decoration":"none",
                        "white-space":"pre",
                        position: "absolute",
                        left: (AXES.Layout.characterSize.width * p) + "px"
                    }
                };
            }
            notFirst = true;
            AXES.Extensions.createNewExtension("axHyperLink", props, field, true);
            exAdded = true;
            
            if (match[4]) 
            {
                var props = {
                    caption: {
                        v: "..."
                    },
                    style: {
                        v: {
                            "text-decoration": "none",
                            "white-space":"pre",
                            position: "absolute",
                            left: (AXES.Layout.characterSize.width * (testText.length - 3)) + "px"
                        }
                    },
                    onClick: {
                        v: "ENV.SENDKEY('F23');",
                        s: true
                    }
                };
                AXES.Extensions.createNewExtension("axHyperLink", props, field, true);
                exAdded = true;
                // subfile menu options
                var menuOptions = {
                    sendkey: true,
                    value  : "F23",
                    text   : "..."
                };
                this._subfileOptionMenu.push(menuOptions);
            }

            match = re.exec(testText);
        }
        return exAdded;
    },    
    checkForPromptButton: function(field, prevField)
    {
        if (!this.rules["PromptButtonGUI"]) return false;
        if (!prevField) return false;
        if (field && field.getValue() == "+" && !prevField.byp) 
        {
            if (prevField.type == AXES.Terminal.Field.Types.INPUT) 
            {
                var props = {
                    onClick: {
                        v: "var F = FIELDS('" + prevField.id + "');\nF.set5250Cursor();\nSENDKEY('F4');",
                        s: true
                    },
                    caption: {
                        v: "+"
                    },
                    tooltip: {
                        v: "More..."
                    },
                    sizeToField: {
                        v: false
                    }
                };
                AXES.Extensions.createNewExtension("axButton", props, field, true);
            }
        }
        // This rule doesn't affect "field" so always return false;
        return false;
    },
    checkForFKeyPanel: function()
    {
        if (!this.rules["FKeyGUIPanel"]) return false;
        // todo check if axFkeyPanel exists
        var fld, i;
        for (i = 0; fld = AXES.Layout.userFields[i]; i++) {
            for (var j = 0, ex; ex = fld.extensions[j]; j++) {
                if (ex.name == "axFkeyPanel") return false;
            }
        }
        // create an axFkeyPanel extension on the Fly...
        fld =AXES.Layout.addUserField();
        var props = {
            position : { v: this.rules["FKeyGUIPanelPosition"] },
            horizontalWidth: { v: this.rules["FKeyGUIPanelHorizontalWidth"]    },
            verticalHeight: { v: this.rules["FKeyGUIPanelVerticalHeight"] },
            resizePane: { v: this.rules["FKeyGUIPanelResizePane"] },
            buttonWidth: { v: this.rules["FKeyGUIPanelButtonWidth"] },
            buttonHeight: { v: this.rules["FKeyGUIPanelButtonHeight"] },
            buttonPadding: { v: this.rules["FKeyGUIPanelButtonGap"] },
            buttonStyle: { v: this.rules["FKeyGUIPanelButtonStyle"] },
            label: { v: this.rules["FKeyGUIPanelButtonLabel"] }
        };
        AXES.Extensions.createNewExtension("axFkeyPanel", props, fld, true);
    }, 
    checkForSubfileOptionMenu: function()
    {
        if (!this.rules["SubOptGUI"]) return false;    
        
        var props = {};
        for (var j=0, ex; ex = AXES.Layout.extensions[j]; j++) 
        {
            if (ex.name == "axSubfileContextMenu") 
            {
                return false;
            }
        }
        AXES.Extensions.createNewExtension("axSubfileContextMenu", props, AXES.Layout, true);
    }
};

/**
 * Handles the user interface (outside of the terminal area)
 * @namespace
 */
AXES.UI = {
    _dialogs: []
};

AXES.UI.show = function()
{
    if ( $( "#contentWrapper:hidden" ).length > 0 )
    {
        $( "#loadingIndicator" ).hide();

        $( ".ui-layout-west" ).tabs(
            {
                beforeLoad: function( event, ui )
                {
                    if ( ui.tab.data( "loaded" ) )
                    {
                        event.preventDefault();
                        return;
                    }

                    ui.jqXHR.success(
                        function()
                        {
                            ui.tab.data( "loaded", true );
                        }
                    );
                    ui.panel.text( TEXT.Core( "Loading..." ) );
                }
            }
        ).on( 
            "keydown keypress keyup",
            function( e )
            {
                e.stopPropagation();
            }
        ).find( ".ui-tabs-nav" ).sortable(
            {
                axis  : 'x',
                zIndex: 2
            }
        );

        AXES.UI.MenuBar.init();

        // Zoom controls
        $("#ax-terminal-zoomin").hover(
        function() {
            $(this).addClass("ui-state-hover");
        },
        function() {
            $(this).removeClass("ui-state-hover");
        }
    ).click(function() {
            AXES.Layout.setZoom("+");
    });

    $("#ax-terminal-zoomout").hover(
            function() {
                $(this).addClass("ui-state-hover");
            },
            function() {
                $(this).removeClass("ui-state-hover");
            }
        ).click(function() {
            AXES.Layout.setZoom("-");
        });

        $("#ax-terminal-setzoom").change(function()
        {
            var al = AXES.Layout, val = this.options[this.selectedIndex].value;
            if ((val == 0) || (val == Math.abs(al.getZoom() / 100)) || (val == al.getZoomSetting()))
            {
                return;
            }
            al.setZoom(val);
        });
    
        AXES.Layout.addListener("ZoomChanged", function() {
            var axesLayout = AXES.Layout, $sel = $("#ax-terminal-setzoom"), sel = $sel[0],
            zoom = axesLayout.getZoom() / 100,
            zoomSetting = axesLayout.getZoomSetting(),
            newSI=-1;
            
            for (var i=0,opt; opt = sel.options[i]; i++) 
            {
                if ((opt.value == zoom) || (opt.value == zoomSetting))
                {
                    newSI = i;
                    break;
                }
            }
            opt = sel.options[0];
            if (newSI == -1) 
            {
                if (opt.value != 0) 
                {
                    opt = document.createElement("option");
                    opt.value = 0;
                    sel.insertBefore(opt, sel.options[0]);
                }
                opt.text = (zoomSetting != "") ? zoomSetting : Math.round(Math.abs(zoom * 100)) + "%";
                newSI = 0;
            } else 
            {
                if (opt.value == 0) 
                {
                    opt.parentNode.removeChild(opt);
                    newSI--;
                }
            }
            if (sel.selectedIndex != newSI) sel.selectedIndex = newSI;

            if ((zoom < 0) || axesLayout.getZoomLock()) {
                $("#ax-terminal-zoomcontrols .ui-icon").addClass("ui-state-disabled");
                sel.disabled = true;
            } else {
                $("#ax-terminal-zoomcontrols .ui-icon").removeClass("ui-state-disabled");
                sel.disabled = false;
            }
        });

        AXES.UI.KeysTab.init();

        if ( AXES.devMode && AXES.LUAAvailable )
        {
            $(".ui-layout-west>ul>li:eq(1)").show();
            $(".ui-layout-west").tabs("option", "active", 1);
            
            $(".ui-layout-west").bind("tabsload", function(e, ui) {
                if (ui.panel.id == "ax-developerTab") {
                    var tabPane = $(ui.panel), dev = AXES.Dev, activeTab;
                    tabPane.tabs({
                        beforeLoad: function( event, ui ) {
                            if ( ui.tab.data( "loaded" ) ) {
                                event.preventDefault();
                                return;
                            }
                     
                            ui.jqXHR.success(function() {
                                ui.tab.data( "loaded", true );
                            });
                            ui.panel.text(TEXT.Core("Loading..."));
                        },
                        activate: dev.onTabSelected
                    }).find(".ui-tabs-nav").sortable({
                        axis: 'x',
                        zIndex: 2
                    });
                    
                    tabPane.resize(function(e)
                    {
                        var $t = $(this);
                        $t.children(".ui-tabs-panel").height($t.innerHeight() - $t.children("ul").outerHeight());
                    });
                    
                    tabPane.resize();
                    dev.ScreensTab.init();
                    dev.ExtensionsTab.init();
                    dev.AutoGUITab.init();
                    dev.ScreensTab.resize();
                    
                    dev.prepTerminalForDev();
                    activeTab = tabPane.tabs("option", "active");
                    dev._activeTab = (activeTab == 0) ? dev.ScreensTab : ((activeTab == 1) ? dev.ExtensionsTab : dev.AutoGUITab);
                    dev._activeTab._selectedFields = [];
                }
            });
        }

        var layout = $( "#contentWrapper" ).removeClass( "ui-helper-hidden" ).layout(
            {
                name: "outerLayout",
                defaults:
                {
                    size                 : "auto",
                    minSize              : 50,
                    contentIgnoreSelector: ".ui-layout-ignore,span",
                    togglerTip_open      : TEXT.Core( "Close This Pane" ),
                    togglerTip_closed    : TEXT.Core( "Open This Pane" ),
                    resizerTip           : TEXT.Core( "Resize This Pane" ),
                    fxName               : "slide",
                    fxSpeed              : "slow"
                },
                north:
                {
                    minSize     : 0,
                    spacing_open: 0,
                    resizable   : false,
                    slidable    : false,
                    fxName      : "none"
                },
                west:
                {
                    size      : 300,
                    minSize   : 290,
                    initClosed: !AXES.devMode,
                    resizable : true,
                    slidable  : false,

                    onresize: function ( paneName, pane )
                    {
                        if ( AXES.Dev )
                        {
                            if ( AXES.Dev.ScreensTab )
                                AXES.Dev.ScreensTab.resize();
                            if ( AXES.Dev.ExtensionsTab )
                                AXES.Dev.ExtensionsTab.resize();
                            if ( AXES.Dev.AutoGUITab )
                                AXES.Dev.AutoGUITab.resize();
                        }
                    }
                },
                center:
                {
                    onresize: function()
                    {
                        AXES.Layout.onResize();
                    }
                }
            }
        );

        if ( $( "#contentWrapper .ui-layout-west" ).length )
            layout.addPinBtn( "#btnToggleWest", "west" );
        this._layout = layout;

        $( "#terminalInner" ).append( $( "#loadingIndicator" ).clone().show() );

        if ( AXES.urlParameters.framework )
        {
            var urlParam = AXES.urlParameters;
            $.ajax(
                {
                    url     : "http://" + urlParam.host + "/agi/lua5/fw_save.lua",
                    type    : "POST",
                    dataType: "json",
                    data    : {
                                folder  : urlParam.folder.replace( / /g, "/" ) + "/",
                                filename: "user" + urlParam.userid + "-" + urlParam.index,
                                format  : "json",
                                content : '{ "loadSuccess": ' + ( $( "#terminalOuter" ).height() == urlParam.frameheight ) + ' }'
                              }
                }
            );
        }
    }
}
/**
 * Shows the header (menu bar) and/or footer (status area) of the terminal display area.
 * @param {String} which Specifies which items to show. Allowed values are "header", "footer" or "both".
 */
AXES.UI.showHeaderAndFooter = function(which)
{
    var selector;
    switch (which.toLowerCase())
    {
        case "header":
            selector = ".header";
            break;
        case "footer":
            selector = ".footer";
            break;
        case "both":
            selector = ".header,.footer";
            break;
        default:
            return;    }
    $("#terminalOuter").children(selector).removeClass("ui-layout-ignore").show();
    this._layout.sizeContent("center");
    AXES.Layout.applyZoom();
}
/**
 * Issues a number of commands to the sidebar. Possible commands are "open", "close", "toggle", "hide", "show".
 * Note that "open" will show the sidebar if necessary but "hide" and "show" will not changed the open/closed state of the sidebar.
 * @param {String} command
 */
AXES.UI.sideBar = function(command) 
{
    var l = this._layout;
    switch (command.toLowerCase()) 
    {
        case "open":
            l.open("west");
            break;
        case "close":
            l.close("west");
            break;
        case "toggle":
            l.toggle("west");
            break;
        case "hide":
            this._sideBarIsOpen = !l.state.west.isClosed
            this._layout.hide("west");
            break;
        case "show":
            l.show("west", this._sideBarIsOpen);
            break;
    }
}
/**
 * Hides the header (menu bar) and/or footer (status area) of the terminal display area.
 * @param {String} which Specifies which items to hide. Allowed values are "header", "footer" or "both".
 */
AXES.UI.hideHeaderAndFooter = function(which)
{
    var selector;
    switch (which.toLowerCase())
    {
        case "header":
            selector = ".header";
            break;
        case "footer":
            selector = ".footer";
            break;
        case "both":
            selector = ".header,.footer";
            break;
        default:
            return;    }
    $("#terminalOuter").children(selector).addClass("ui-layout-ignore").hide();
    this._layout.sizeContent("center");
    AXES.Layout.applyZoom();
}

AXES.UI.setTheme = function(themeName) 
{
    themeName = themeName.toLowerCase();
    if (themeName.slice(-4) == ".css") themeName = themeName.slice(0,-4);
    if (themeName.substring(0,6) != "theme_") themeName = "theme_" + themeName;
    $("#axThemeCSS").attr("href", AXESBUILD.updateURL("css/" + themeName + ".css"));
    AXES.Layout.AutoGUI.prepareBackgroundImages();
    
    var mc = this.MenuBar.menuCollection
    var m = (mc) ? mc.menusById["display"] : null;
    var themesMenu = (m) ? m.getItem("themes") : null;
    var subM = (themesMenu) ? themesMenu.subMenu : null;
    
    if (subM && subM.menuItems) 
    {
        for (var i = 0, menuItem; menuItem = subM.menuItems[i]; i++) 
        {
            menuItem.setChecked(menuItem.id == themeName);
        }
    }
}
AXES.UI.disableThemes = function() 
{
    this._savedTheme = this.getTheme();
    this.setTheme("default");
    if (this._allowThemesMenu) 
    {
        var mc = this.MenuBar.menuCollection.menusById["display"];
        var themesMenu = (mc) ? mc.getItem("themes") : null;
        if (themesMenu) themesMenu.disable();
    }
}
AXES.UI.enableThemes = function() 
{
    if (this._savedTheme) 
    {
        this.setTheme(this._savedTheme);
        delete this._savedTheme;
    }
    if (this._allowThemesMenu) 
    {
        var mc = this.MenuBar.menuCollection.menusById["display"];
        var themesMenu = (mc) ? mc.getItem("themes") : null;
        if (themesMenu) themesMenu.enable();
    }
}
AXES.UI.getTheme = function() 
{
    var href = $("#axThemeCSS").attr("href");
    
    var match = href.match(/css\/theme_(.*?)\.css/i);
    return (match && (match.length > 1)) ? match[1] : "";
}
AXES.UI.clearDialogs = function() {
    $("body>div.ui-dialog").dialog('close').dialog('destroy');
}
AXES.UI.doCopyTextDialog = function() {
    var terminalText = AXES.Terminal.getAsText();
    var dragBox = $("<div></div>").attr("class", "ax-dragbox");
    var dragBoxAt = null;
    var mouseIsDown = false;
    
    AXES.UI.clearDialogs();
    var dialogOptions = {
        autoOpen: true,
        dialogClass: "ax-copytextdialog",
        closeOnEscape: true,
        bgiframe: true,
        modal: true,
        position: "top",
        buttons: {}
    };

    dialogOptions.buttons[TEXT.Core("Copy As Text")] = function()
    {
        var charGrid = getSelectedCharacterGrid();
        var text = AXES.Terminal.getAsText(charGrid);
        AXES.Utils.copyTextToClipboard(text.trim());
        $(this).dialog("close").dialog("destroy");
    };
    dialogOptions.buttons[TEXT.Core("Copy As Table")] = function()
    {
        var charGrid = getSelectedCharacterGrid();
        var html = AXES.Terminal.getAsTable(charGrid);
        AXES.Utils.copyTextToClipboard(html);
        $(this).dialog("close").dialog("destroy");
    };
    
    var getSelectedCharacterGrid = function() {
        var charGrid = {
            top: 0,
            left: 0,
            right: AXES.Terminal.cols,
            bottom: AXES.Terminal.rows
        };
        
        if (dragBoxAt) 
        {
            var xUnit = Math.floor(d[0].scrollWidth / AXES.Terminal.cols);
            var yUnit = Math.floor(d[0].scrollHeight / AXES.Terminal.rows);
            
            dragBoxAt.height = (dragBoxAt.height < yUnit)? yUnit:dragBoxAt.height;
            dragBoxAt.width = (dragBoxAt.width < xUnit)? xUnit:dragBoxAt.width;
            charGrid.left = Math.floor(Math.min(dragBoxAt.left, dragBoxAt.left + dragBoxAt.width)/xUnit);
            charGrid.top = Math.ceil(Math.min(dragBoxAt.top, dragBoxAt.top + dragBoxAt.height)/yUnit);
            charGrid.right = Math.floor(Math.abs(dragBoxAt.left + dragBoxAt.width)/xUnit);
            charGrid.bottom = Math.floor(Math.abs(dragBoxAt.top + dragBoxAt.height)/yUnit);
        }
        return charGrid;
    }
    
    var d = $("<div></div>")
        .attr("title", TEXT.Core("Copy To Clipboard"))
        .text(AXES.Terminal.getAsText()).dialog(dialogOptions);
    

    d.dialog('option', 'width', d[0].scrollWidth).dialog('option', 'position', 'top');
    
    d.append(dragBox);
    
    d.bind("dragstart", function(e) {e.preventDefault();});
    d.mousedown(function(e) {
        var offset = $(this).offset();
        dragBoxAt = {
            left: e.pageX - offset.left,
            top: e.pageY - offset.top,
            width: 0, height: 0
        };
        dragBox.css({
            left: dragBoxAt.left + "px",
            top: dragBoxAt.top + "px",
            width: "0px",
            height: "0px"
        }).show();
        mouseIsDown = true;
        e.preventDefault();
    });
    d.mousemove(function(e) {
        if (mouseIsDown) 
        {
            var offset = $(this).offset();
            dragBoxAt.width = (e.pageX - offset.left) - dragBoxAt.left;
            dragBoxAt.height = (e.pageY - offset.top) - dragBoxAt.top;
            dragBox.css({
                left: Math.min(dragBoxAt.left, dragBoxAt.left + dragBoxAt.width) + "px",
                top: Math.min(dragBoxAt.top, dragBoxAt.top + dragBoxAt.height) + "px",
                width: Math.abs(dragBoxAt.width) + "px",
                height: Math.abs(dragBoxAt.height) + "px"
            });
        }
        e.preventDefault();
    });
    d.mouseup(function(e) {
        mouseIsDown = false;
        e.preventDefault();
    });
}
/**
 * Manages the Keys tab
 * @namespace
 */
AXES.UI.KeysTab = {
    init: function()
    {
        var keySelect = document.getElementById("ax-keymap-select");
        if (keySelect) 
        {
            for (var i = 0, keymap; keymap = AXES.Keyboard._keymaps[i]; i++) 
            {
                var displayName = TEXT.core(keymap);
                if (displayName == keymap) displayName = TEXT.cust(keymap);
                var opt = document.createElement("option");
                opt.setAttribute("value", keymap);
                if (AXES.Keyboard.currentKeymap == keymap) opt.setAttribute("selected", "selected");
                opt.appendChild(document.createTextNode(displayName));
                keySelect.appendChild(opt);
            }
            $("#ax-keymap-select").change(function()
            {
                AXES.Keyboard._loadMappingFile($(this).val());
            });
            
            $("#ax-keysTab-fkeys button.ax-fkeybutton").click(function()
            {
                var key = $(this).text().replace("F", "Function");
                AXES.Terminal.doPost(AXES.Keyboard.aidKeys[key]);
            });
            this.addUserDefKeys();
        }
    },
    addUserDefKeys: function() {
        var self = this;
        
        $("#ax-keysTab-udkeys .ax-udkeybutton").each(function()
        {
            var $t = $(this);
            var termKey = $t.attr("ax-key-command");
            var pcKey = AXES.Keyboard.reverseKeyMappings[termKey];
            pcKey = pcKey || "";
            $t.data("termKey",termKey).click(self.udKeyBtnPressed).next().text(pcKey);
        });
        return;
    },
    udKeyBtnPressed: function() {
        var termKey = $(this).data("termKey");
        AXES.UI.KeysTab.simulateTermKey(termKey);
    },
    simulateTermKey: function(termKey) {
        if (AXES.Layout._lastFocusElement) 
        {
            AXES.Layout._lastFocusElement.focus();
            
            var e = jQuery.Event("keydown");
            var pcKey = AXES.Keyboard.reverseKeyMappings[termKey];
            var aidKey = (termKey) ? AXES.Keyboard.aidKeys[termKey] : null;
            e.which = AXES.Keyboard.keysByName[pcKey];
            e.axKeyInfo = {
                evtOrigin: "button",
                name: pcKey,
                mappedTo: termKey,
                aidKey: (aidKey) ? aidKey : null
            }
            $(AXES.Layout._lastFocusElement).trigger(e);
        } else {
            AXES.Keyboard.doTermKeypress(termKey);
        }
    }
}

/**
 * A menu item
 * 
 * @param {Object} id An optional ID for the menuitem. Required if you need to access the menu item in the future.
 * @param {Object} options
 */
/*
 * Menu code based on clickMenu by Roman Weich. http://p.sohei.org/jquery-plugins/clickmenu/
 */
AXES.UI.MenuItem = function(id, options){
    this.id = id;
    this.options = options;
    
    this.separator = (options && options.content) ? false : true;
    this.enabled = (options && options.enabled === false) ? false : true;
    this.active = false;
    this.buildItem();
}

AXES.UI.MenuItem.prototype.buildItem = function()
{
    var innerDiv = document.createElement("div"), classes = [], icon, opts, content, isStr;
    innerDiv.className = "ax-menu-item";
    
    this.theLI = document.createElement("li");
    if (this.separator) 
        classes.push("ax-menu-separator");
    if (!this.enabled) classes.push("disabled");

    this.theLI.className = classes.join(" ");
    this.theLI.appendChild(innerDiv);
    this.$theLI = $(this.theLI);
    
    if (!this.separator) 
    {
        icon = document.createElement("span");
        icon.className = "ax-menu-item-icon";
        innerDiv.appendChild(icon);
    
        opts = this.options;
        isStr = typeof opts.content == 'string';
        if (isStr && opts.url) 
            content = $('<a href="' + opts.url + '"' + (opts.urlTarget ? 'target="' + opts.urlTarget + '"' : '') + '>' + opts.content + '</a>');
        else if (isStr || !opts.content.length) 
            content = [opts.content];
        
        //go through the passed DOM-Elements (or jquery objects or text nodes.) and append them to the menus list item
        //this.$theLI.append(this.src) is really slow when having a lot(!!!) of items
        for (var i=0, item; item = content[i]; i++) 
        {
            if (typeof item == 'string') 
            {
                //we cant use createTextNode, as html entities won't be displayed correctly (eg. &copy;)
                elem = document.createElement('span');
                elem.innerHTML = item;
                innerDiv.appendChild(elem);
            }
            else innerDiv.appendChild(item);
        }
        
        var self = this;
        this.$theLI.click(function(e)
        {
            self.onClick(e, this);
        });
        
        this.bindHover();
    }
}

AXES.UI.MenuItem.prototype.setIconClass = function(className) 
{
     this.$theLI.find("span.ax-menu-item-icon")[0].className = ("ax-menu-item-icon " + className).trim();
}

AXES.UI.MenuItem.prototype.setChecked = function(checked)
{
    if (checked) 
    {
        this.setIconClass("ui-icon ui-icon-check");
    } else 
    {
        this.setIconClass("");
    }
}

AXES.UI.MenuItem.prototype.onClick = function(e, scope)
{
    if (this.enabled && this.options.onClick) 
        this.options.onClick.call(scope, e, this);
    AXES.UI.Menu.closeAll();
}
AXES.UI.MenuItem.prototype.bindHover = function()
{
    var self = this;
    this.$theLI.hover(function()
    {
        self.hoverIn();
    }, function()
    {
        self.hoverOut();
    });
}
AXES.UI.MenuItem.prototype.hoverIn = function(noSubMenu)
{
    this.removeTimer();
    
    var i, pms = this.parentMenu.subMenus, pmi = this.parentMenu.menuItems, self = this;
    
    //remove the timer from the parent item, when there is one (e.g. to close the menu)
    if (this.parentMenu.timer) this.parentMenu.removeTimer();
    
    if (!this.enabled) return;
    
    //deactivate all menuItems on the same level
    for (i = 0; i < pmi.length; i++) 
    {
        if (pmi[i].active) pmi[i].setInactive();
    }
    
    this.setActive();
    AXES.UI.Menu.activeMenu = this.parentMenu;

    //are there open submenus on the same level? close them!
    for (i = 0; i < pms.length; i++) 
    {
        if (pms[i].visible && pms[i] != this.subMenu && !pms[i].timer) //close if there is no closetimer running already
            pms[i].addTimer(function()
            {
                this.hide();
            }, pms[i].options.hideDelay);
    }
    
    if (this.subMenu && !noSubMenu) 
    {
        //set timeout to show menu
        this.subMenu.addTimer(function() {this.show();}, this.subMenu.options.showDelay);
    }
}
            
AXES.UI.MenuItem.prototype.hoverOut = function()
{
    this.removeTimer();
    
    if (!this.enabled) 
    {
        return;
    }

    if (!this.subMenu || !this.subMenu.visible) 
        this.setInactive();
}
            
AXES.UI.MenuItem.prototype.removeTimer = function()
{
    if (this.subMenu) 
    {
        this.subMenu.removeTimer();
    }
}
AXES.UI.MenuItem.prototype.setActive = function()
{
    this.active = true;
    this.$theLI.addClass('active');
    
    //set the parent menu item active too if necessary
    var pmi = this.parentMenu.parentMenuItem;
    if (pmi && !pmi.active) pmi.setActive();
    AXES.UI.Menu.activeItem = this;
}
AXES.UI.MenuItem.prototype.setInactive = function()
{
    this.active = false;
    this.$theLI.removeClass('active');
    if (this == AXES.UI.Menu.activeItem) AXES.UI.Menu.activeItem = null;
}
AXES.UI.MenuItem.prototype.enable = function()
{
    this.$theLI.removeClass('disabled');
    this.enabled = true;
}
AXES.UI.MenuItem.prototype.disable= function()
{
    this.$theLI.addClass('disabled');
    this.enabled = false;
}
AXES.UI.MenuItem.prototype.destroy = function()
{
    this.removeTimer();
    
    this.$theLI.remove();
    
    //unbind events
    this.$theLI.unbind('mouseover').unbind('mouseout').unbind('click');
    //delete submenu
    if (this.subMenu) 
    {
        this.subMenu.destroy();
        delete this.subMenu;
    }
    this.parentMenu.removeItem(this);
}
AXES.UI.MenuItem.prototype.setContent = function(newContent) {
    this.$theLI.find("span:not(.ax-menu-item-icon)").text(newContent);
}
AXES.UI.MenuItem.prototype.addSubMenu = function(menu)
{
    if (this.subMenu) return;
    this.subMenu = menu;
    if (this.parentMenu && jQuery.inArray(menu, this.parentMenu.subMenus) == -1) 
        this.parentMenu.subMenus.push(menu);
    
    var a = AXES.UI.Menu.arrowElement.cloneNode(0);
    this.theLI.firstChild.appendChild(a);
}



AXES.UI.Menu = function(id, target, options)
{
    this.id = id;
    this.menuItems = []; //all direct child AXES.UI.MenuItem objects
    this.menuItemsById = {};
    this.subMenus = []; //all subMenus from this.menuItems
    this.visible = false;
    this.active = false;
    this.parentMenuItem = null;
    this.options = jQuery.extend({}, AXES.UI.Menu.defaults, options);
    this.target = target;
    this.$eUL = null;
    this.timer = null;
    this.menuCollection = null;
    this.openTimer = null;
        
    if (!target) 
    {
        return;
    }
    else if (target instanceof AXES.UI.MenuItem) 
    {
        this.parentMenuItem = this.target;
        this.target.addSubMenu(this);
        this.target = this.target.$theLI;
    }
    
    this.buildMenu();
};

// Add static methods and variables
jQuery.extend(AXES.UI.Menu, {
    activeMenu: null,
    activeItem: null,
    menuULElement: $('<ul class="ax-menu-ul ui-widget"></ul>')[0],
    arrowElement: $('<span class="ax-menu-item-arrow ui-icon ui-icon-triangle-1-e">&nbsp;</span>')[0],
    defaults: {
        // Menu options
        showDelay : 200,
        hideDelay : 200,
        hoverOpenDelay: 0,
        offsetTop : 0,
        offsetLeft : 0,
        minWidth: 0,
        onOpen: null,
        onClose: null,
        appendTo: "#terminalContent"
    },
    visibleMenus: function() {
        // To avoid risks of memory leaks and problems with menu elements
        // being removed without proper cleanup, we can't keep a global
        // list of visible menus.
        var visMenus = $("body").find(".ax-menu-ul:visible");
        return visMenus.map(function() {
            return $(this).data("menu");
        });
    },
    checkMouse: function(e)
    {
        var t = e.target;
        
        //the user clicked on the target of the currenty open menu
        var visibleMenus = AXES.UI.Menu.visibleMenus();
        if (visibleMenus.length && t == visibleMenus[0].target) 
        {
            return;
        }

        var $t = $(t).closest(".ax-menu-ul");
        if (($t.length == 0) || $t.is(":hidden")) AXES.UI.Menu.closeAll(visibleMenus);
    },
    checkKey : function(e)
    {
        var visibleMenus = AXES.UI.Menu.visibleMenus();
        switch ( e.keyCode )
        {
            case 13: //return
                if ( AXES.UI.Menu.activeItem )
                    AXES.UI.Menu.activeItem.$theLI.click(e, AXES.UI.Menu.activeItem.theLI);
                break;
            case 27: //ESC
                AXES.UI.Menu.closeAll(visibleMenus);
                break;
            case 37: //left
                if ( !AXES.UI.Menu.activeMenu )
                    AXES.UI.Menu.activeMenu = visibleMenus[0];
                var a = AXES.UI.Menu.activeMenu;
                if ( a && a.parentMenuItem ) //select the parent menu and close the submenu
                {
                    //unbind the events temporary, as we dont want the hoverout event to fire
                    var pmi = a.parentMenuItem;
                    pmi.$theLI.unbind('mouseout').unbind('mouseover');
                    a.hide();
                    pmi.hoverIn(true);
                    setTimeout(function(){ //bind again..but delay it
                        pmi.bindHover();
                    });
                }
                else if ( a && a.menuCollection ) //select the previous menu in the collection
                {
                    var pos,
                        mcm = a.menuCollection.menus;
                    if ( (pos = jQuery.inArray(a, mcm)) > -1 )
                    {
                        if ( --pos < 0 )
                            pos = mcm.length - 1;
                        AXES.UI.Menu.closeAll(visibleMenus);
                        mcm[pos].show();
                        mcm[pos].setActive();
                        if ( mcm[pos].menuItems.length ) //select the first item
                            mcm[pos].menuItems[0].hoverIn(true);
                    }
                }
                break;
            case 38: //up
                if ( AXES.UI.Menu.activeMenu )
                    AXES.UI.Menu.activeMenu.selectNextItem(-1);
                break;
            case 39: //right
                if ( !AXES.UI.Menu.activeMenu )
                    AXES.UI.Menu.activeMenu = visibleMenus[0];
                var m, a = AXES.UI.Menu.activeMenu,
                    asm = AXES.UI.Menu.activeItem ? AXES.UI.Menu.activeItem.subMenu : null;
                if ( a )
                {
                    if ( asm && asm.menuItems.length ) //select the submenu
                    {
                        asm.show();
                        asm.menuItems[0].hoverIn();
                    }
                    else if ( (a = a.inMenuCollection()) ) //select the next menu in the collection
                    {
                        var pos,
                            mcm = a.menuCollection.menus;
                        if ( (pos = jQuery.inArray(a, mcm)) > -1 )
                        {
                            if ( ++pos >= mcm.length )
                                pos = 0;
                            AXES.UI.Menu.closeAll(visibleMenus);
                            mcm[pos].show();
                            mcm[pos].setActive();
                            if ( mcm[pos].menuItems.length ) //select the first item
                                mcm[pos].menuItems[0].hoverIn(true);
                        }
                    }
                }
                break;
            case 40: //down
                if ( !AXES.UI.Menu.activeMenu )
                {
                    if ( visibleMenus.length && visibleMenus[0].menuItems.length )
                        visibleMenus[0].menuItems[0].hoverIn();
                }
                else
                    AXES.UI.Menu.activeMenu.selectNextItem();
                break;
        }
        if ( e.keyCode > 36 && e.keyCode < 41 )
            return false; //this will prevent scrolling
    },
    closeAll: function(visibleMenus)
    {
        var vm = (visibleMenus) ? visibleMenus : AXES.UI.Menu.visibleMenus();
        for (var i=vm.length-1; i>=0; i--) {
            vm[i].hide();
        }
    }
});

AXES.UI.Menu.prototype.buildMenu = function()
{
    var self = this;
    
    this.$eUL = $(AXES.UI.Menu.menuULElement.cloneNode(true)).data("menu", this);
    
    $("body").append(this.$eUL);
    
    //bind events
    if (!this.parentMenuItem) 
    {
        $(this.target).click(function(e)
        {
            self.onClick(e);
        }).hover(function(e)
        {
            self.setActive();
            
            if (self.options.hoverOpenDelay) 
            {
                self.openTimer = setTimeout(function()
                {
                    if (!self.visible) 
                        self.onClick(e);
                }, self.options.hoverOpenDelay);
            }
        }, function()
        {
            if (!self.visible) 
                $(this).removeClass('activetarget');
            
            if (self.openTimer) 
                clearTimeout(self.openTimer);
        });
    }
    else 
    {
        this.$eUL.hover(function()
        {
            self.setActive();
        }, function(){});
    }
}
AXES.UI.Menu.prototype.setActive = function()
{
    if (!this.parentMenuItem) 
        $(this.target).addClass('activetarget');
    else this.active = true;
}
AXES.UI.Menu.prototype.addItem = function(id, item)
{
    if (item instanceof AXES.UI.MenuItem) 
    {
        if (jQuery.inArray(item, this.menuItems) == -1) 
        {
            this.$eUL.append(item.$theLI);
            this.menuItems.push(item);
            if (typeof item.id == "string" && item.id.length > 0) {
                this.menuItemsById[item.id] = item;
            }
            item.parentMenu = this;
            if (item.subMenu) 
                this.subMenus.push(item.subMenu);
        }
    }
    else 
    {
        this.addItem(new AXES.UI.MenuItem(id, item));
    }
}
AXES.UI.Menu.prototype.getItem = function(item) {
    if (typeof item == "string" && item.length > 0) 
    { return this.menuItemsById[item]; }
    else if (typeof item == "number" && item < this.menuItems.length) 
    { return this.menuItems[item]; }
    else 
    { return undefined; }
}
AXES.UI.Menu.prototype.addItems = function(id, items)
{
    for (var i = 0; i < items.length; i++) 
    {
        this.addItem(id + i, items[i]);
    }
}
AXES.UI.Menu.prototype.removeItem = function(item)
{
    var pos = jQuery.inArray(item, this.menuItems);
    if (pos > -1) 
    {
        this.menuItems.splice(pos, 1);
    }
    if (typeof item.id == "string" && item.id.length > 0) 
    {
        delete this.menuItemsById[item.id];
    }
    item.parentMenu = null;
}
AXES.UI.Menu.prototype.hide = function()
{
    if (!this.visible) return;
    
    this.$eUL.hide();
    
    this.visible = this.active = false;
    
    $(this.target).removeClass("activetarget");
    
    //hide all submenus
    for (var i = 0, sub; sub = this.subMenus[i]; i++) 
    {
        sub.hide();
    }
    
    //set all items inactive (e.g. remove hover class..)
    for (var i = 0, item; item = this.menuItems[i]; i++) 
    {
        if (item.active) 
            item.setInactive();
    }
    
    if (AXES.UI.Menu.visibleMenus().length == 0) //unbind events when the last menu was closed
        $(document).unbind('mousedown', AXES.UI.Menu.checkMouse).unbind('keydown', AXES.UI.Menu.checkKey);
    
    if (AXES.UI.Menu.activeMenu == this) 
        AXES.UI.Menu.activeMenu = null;
    
    if (this.options.onClose) 
        this.options.onClose.call(this);
}
AXES.UI.Menu.prototype.show = function(e)
{
    if (this.visible) return;
    
    var zi, pmi = this.parentMenuItem;
    
    if (this.menuItems.length) //show only when it has items
    {
        var vmCount = AXES.UI.Menu.visibleMenus().length;
        if (pmi) //set z-index
        {
            zi = parseInt(pmi.parentMenu.$eUL.css('z-index'));
            this.$eUL.css('z-index', (isNaN(zi) ? 1 : zi + 1));
        }
        this.$eUL.css({
            visibility: 'hidden',
            display: 'block'
        });
        
        //set min-width
        if (this.options.minWidth) 
        {
            if (this.$eUL.width() < this.options.minWidth) 
                this.$eUL.css('width', this.options.minWidth);
        }
        
        this.setPosition();
        this.$eUL.css({
            display: 'none',
            visibility: ''
        }).show();
        
        
        if (this.options.onOpen) 
            this.options.onOpen.call(this);
    }
    if (vmCount == 0) 
        $(document).bind('mousedown', AXES.UI.Menu.checkMouse).bind('keydown', AXES.UI.Menu.checkKey);
    
    this.visible = true;
}
AXES.UI.Menu.prototype.setPosition = function()
{
    if ($("body").hasClass("mobile")) return;
    
    var $t, o, posX, posY, pmo, //parent menu offset
     wst, //window scroll top
     wsl, //window scroll left
     ww = $(window).width(), wh = $(window).height(), pmi = this.parentMenuItem, height = this.$eUL[0].clientHeight, width = this.$eUL[0].clientWidth, pheight; //parent height
    if (pmi) 
    {
        //position on the right side of the parent menu item
        o = pmi.$theLI.offset();
        posX = o.left + pmi.$theLI.width();
        posY = o.top;
    }
    else 
    {
        //position right below the target
        $t = $(this.target);
        o = $t.offset();
        posX = o.left + this.options.offsetLeft;
        posY = o.top + $t.height() + this.options.offsetTop;
    }
    
    //y-pos
    wst = $(window).scrollTop();
    if (wh < height) //menu is bigger than the window
    {
        //position the menu at the top of the visible area
        posY = wst;
    }
    else if (wh + wst < posY + height) //outside on the bottom?
    {
        if (pmi) 
        {
            pmo = pmi.parentMenu.$eUL.offset();
            pheight = pmi.parentMenu.$eUL[0].clientHeight;
            if (height <= pheight) 
            {
                //bottom position = parentmenu-bottom position
                posY = pmo.top + pheight - height;
            }
            else 
            {
                //top position = parentmenu-top position
                posY = pmo.top;
            }
            //still outside on the bottom?
            if (wh + wst < posY + height) 
            {
                //shift the menu upwards till the bottom is visible
                posY -= posY + height - (wh + wst);
            }
        }
        else 
        {
            //shift the menu upwards till the bottom is visible
            posY -= posY + height - (wh + wst);
        }
    }

    //x-pos
    wsl = $(window).scrollLeft();
    if (ww + wsl < posX + width) 
    {
        if (pmi) 
        {
            //display the menu not on the right side but on the left side
            posX -= pmi.$theLI.width() + width;
            //outside on the left now?
            if (posX < wsl) 
                posX = wsl;
        }
        else 
        {
            //shift the menu to the left until it fits
            posX -= posX + width - (ww + wsl);
        }
    }
    
    //set position
    this.$eUL.css({
        left: posX,
        top: posY
    });
}
AXES.UI.Menu.prototype.onClick = function(e)
{
    if (this.visible) 
    {
        this.hide();
        this.setActive(); //the class is removed in the hide() method..add it again
    }
    else 
    {
        //close all open menus
        AXES.UI.Menu.closeAll();
        this.show(e);
    }
}
AXES.UI.Menu.prototype.addTimer = function(callback, delay)
{
    var self = this;
    this.timer = setTimeout(function()
    {
        callback.call(self);
        self.timer = null;
    }, delay);
}
AXES.UI.Menu.prototype.removeTimer = function()
{
    if (this.timer) 
    {
        clearTimeout(this.timer);
        this.timer = null;
    }
}
AXES.UI.Menu.prototype.selectNextItem = function(offset)
{
    var i, pos = 0, mil = this.menuItems.length, o = offset || 1;
    
    //get current pos
    for (i = 0; i < mil; i++) 
    {
        if (this.menuItems[i].active) 
        {
            pos = i;
            break;
        }
    }
    this.menuItems[pos].hoverOut();
    
    do //jump over the separators
     {
        pos += o;
        if (pos >= mil) 
            pos = 0;
        else if (pos < 0) 
            pos = mil - 1;
    } while (this.menuItems[pos].separator);
    this.menuItems[pos].hoverIn(true);
}
AXES.UI.Menu.prototype.inMenuCollection = function()
{
    var m = this;
    while (m.parentMenuItem) 
        m = m.parentMenuItem.parentMenu;
    return m.menuCollection ? m : null;
}
AXES.UI.Menu.prototype.destroy = function() //delete menu
{
    var pos, item;
    
    this.hide();
    
    //unbind events
    if (!this.parentMenuItem) 
        $(this.target).unbind('click').unbind('mouseover').unbind('mouseout');
    else this.$eUL.unbind('mouseover').unbind('mouseout');
    
    //destroy all items
    while (this.menuItems.length) 
    {
        item = this.menuItems[0];
        item.destroy();
        delete item;
    }
    
    if (this.menuCollection) 
    {
        this.menuCollection.removeMenu(this);
    }
    
    this.$eUL.remove();
}


AXES.UI.MenuCollection = AXClass.extend(
/**
* @lends AXES.UI.MenuCollection.prototype
*/
{
    menus: [],
    menusById: {},
    /** 
     * Represents a collection of related menus (often in a menu bar).
     * @constructs
     * @augments AXClass
     * @param {Object[]} [items] an array of AXES.UI.Menu objects.
     */
    init: function(items)
    {
        if (items && items.length) 
        {
            for (var i = 0; i < items.length; i++) 
            {
                this.addMenu(items[i]);
                items[i].menuCollection = this;
            }
        }
    },
    addMenu: function(menu)
    {
        if (menu instanceof AXES.UI.Menu) 
        {
            this.menus.push(menu);
            
            menu.menuCollection = this;
            if (typeof menu.id == "string" && menu.id.length > 0) this.menusById[menu.id] = menu;
            
            var self = this;
            $(menu.target).hover(function()
            {
                if (menu.visible) 
                {
                    return;
                }

                //when there is an open menu in this collection, hide it and show the new one
                for (var i = 0; i < self.menus.length; i++) 
                {
                    if (self.menus[i].visible) 
                    {
                        self.menus[i].hide();
                        menu.show();
                        return;
                    }
                }
            }, function()
            {
            });
        }
    },
    removeMenu: function(menu)
    {
        if (typeof menu == "string" && menu.length > 0) 
            menu = this.menusById[menu];
        
        if (menu instanceof AXES.UI.Menu) 
        {
            var pos = jQuery.inArray(menu, this.menus);
            if (pos > -1) 
                this.menus.splice(pos, 1);
            
            if (typeof menu.id == "string" && menu.id.length > 0) 
                delete this.menusById[menu.id];
        }
    }
});


/**
 * @namespace
 */
AXES.UI.MenuBar = 
{
    _flags: {},
    init: function()
    {
        var m, item, subM;
        
        this.menuOptions = {
            minWidth: 120,
            appendTo: "body"
        };
        
        this.barUL = $('#ax-mainmenu')[0];
        this.menuCollection = new AXES.UI.MenuCollection(null);
        
        this._sessionMenu = m = this.addMenu("session", TEXT.Core("Session"));
        this.addItem(m, "new", {
            content: TEXT.Core("New"),
            onClick: this._sessionNew
        });
        this.addItem(m, "suspend", {
            content: TEXT.Core("Suspend"),
            enabled: false,
            onClick: this._sessionSuspend
        });
        this.addItem(m, "break", {
            content: TEXT.Core("Send Break"),
            enabled: false,
            onClick: this._sessionBreak
        });
        this.addItem(m, "close", {
            content: TEXT.Core("Close"),
            enabled: false,
            onClick: this._sessionClose
        });
        
        this._displayMenu = m = this.addMenu("display", TEXT.Core("Display"));
        this.addItem(m, "refresh", {
            content: TEXT.Core("Refresh"),
            onClick: this._displayRefresh
        });
        
        item = this.addItem(m, "themes", {
            content: TEXT.Core("Select Theme")
        });
        subM = new AXES.UI.Menu("", item, this.menuOptions);
        this.buildThemesMenu(subM);
        
        this.addItem(m, "", {
            content: ""
        }, null);
        
        this.addItem(m, "copy", {
            content: TEXT.Core("Copy..."),
            enabled: false,
            onClick: AXES.UI.doCopyTextDialog
        });
        this.addItem(m, "", {
            content: ""
        }, null);
        item = this.addItem(m, "autogui", {
            content: TEXT.Core("AutoGUI"),
            onClick: this._displayAutogui,
            enabled: AXES.Layout.AutoGUI.enabled
        });
        item.setChecked(AXES.Layout.AutoGUI.autoGUIOn);
        
        this._toolsMenu = m = this.addMenu("tools", TEXT.Core("Tools"));
        this.addItem(m, "spoolFiles", {
            content: TEXT.Core("Spool Files"),
            onClick: this._toolsSpool,
            enabled: false
        });
        this.addItem(m, "dataExplorer", {
            content: TEXT.Core("Data Explorer"),
            onClick: this._toolsDataExplorer,
            enabled: false
        });
        if (window.JSMINFO) JSMINFO.init();
        
        this._helpMenu = m = this.addMenu("help", TEXT.Core("Help"));
        item = this.addItem(m, "links", {
            content: TEXT.Core("Links")
        });
        subM = new AXES.UI.Menu("", item, this.menuOptions);
        this.addItem(subM, "lansa", {
            content: TEXT.Core("LANSA"),
            url: "http://www.lansa.com/",
            urlTarget: "_blank"
        });
        
        item = this.addItem(m, "diagnostics", {
            content: TEXT.Core("Diagnostics")
        });
        subM = new AXES.UI.Menu("", item, this.menuOptions);
        this.addItem(subM, "log", {
            content: TEXT.Core("Log..."),
            onClick: function()
            {
                AXES.Logging.showConsole();
            }
        });
        
        this.addItem(subM, "", {
            content: "",
            enabled: false
        });
        this.addItem(subM, "sethosttrace", {
            content: TEXT.Core("Host Trace On"),
            enabled: false,
            onClick: function(e, item)
            {
                if (item.hostTraceOn) 
                {
                    if (confirm("Disable Host Trace?")) 
                    {
                        AXES.Terminal.doSysPost("TRACE_OFF");
                        item.setContent(TEXT.Core("Host Trace On"));
                        item.hostTraceOn = false;
                    }
                }
                else 
                {
                    if (confirm(TEXT.Core("Enable Host Trace (SEVERE PERFORMANCE IMPACT)?"))) 
                    {
                        AXES.Terminal.doSysPost("TRACE_ON");
                        item.setContent(TEXT.Core("Host Trace Off"));
                        item.hostTraceOn = true;
                    }
                }
            }
        });
        this.addItem(subM, "clearhosttrace", {
            content: TEXT.Core("axTerm.T007"),
            enabled: false,
            onClick: function(e, item)
            {
                /* Translators: Confirming that the user wants to clear the host trace */
                if (confirm(TEXT.Core("Clear Host Trace?"))) 
                {
                    AXES.Terminal.doSysPost("TRACE_CLEAR");
                }
            }
        });
        
        item = this.addItem(m, "about", {
            content: TEXT.Core("About"),
            onClick: function()
            {
                AXES.UI.clearDialogs();
                var dialogOptions = {
                    autoOpen: true,
                    closeOnEscape: true,
                    bgiframe: true,
                    modal: true,
                    position: "top",
                    resizable: false,
                    title: TEXT.Core("About aXes"),
                    width: 348,
                    height: 250,
                    buttons: {},
                    close: function()
                    {
                        $(this).dialog("destroy");
                    }
                }
                dialogOptions.buttons[TEXT.Core("OK")] = function()
                {
                    $(this).dialog("close");
                };
                $("#ax-aboutDialog-version").text(AXESBUILD.getDisplayString());
                $("#ax-aboutDialog").dialog(dialogOptions);
            }
        });
        
        AXES.Terminal.addListener("ScreenArrived", this.refreshMenus, this);
        this.refreshMenus();
    },
    addMenu: function(id, text, options)
    {
        var li = document.createElement("li");
        li.className = "ax-menumain";
        li.innerHTML = text;
        this.barUL.appendChild(li);
        var m = new AXES.UI.Menu(id, li, jQuery.extend({}, this.menuOptions, options));
        this.menuCollection.addMenu(m);
        return m;
    },
    addItem: function(menu, id, options)
    {
        if (typeof menu == "string") menu = this.menuCollection.menusById[menu];
        if (menu) 
        {
            var newItem = new AXES.UI.MenuItem(id, options);
            menu.addItem(id, newItem);
        }
        return newItem;
    },
    refreshMenus: function()
    {
        // Update menu for new screen state.
        var m, i;
        var isConnected = AXES.Terminal.isConnected;
        var autoGUIOn = AXES.Layout.AutoGUI.autoGUIOn;
        var f = this._flags;
        
        if (f.isConnected != isConnected) 
        {
            m = this._sessionMenu;
            if (i = m.getItem("suspend")) (isConnected) ? i.enable() : i.disable();
            if (i = m.getItem("break")) (isConnected) ? i.enable() : i.disable();
            if (i = m.getItem("close")) (isConnected) ? i.enable() : i.disable();
            
            m = this._displayMenu;
            if (i = m.getItem("refresh")) (isConnected && !AXES.EXTHOST_TSInterface) ? i.enable() : i.disable();
            if (i = m.getItem("copy")) (isConnected) ? i.enable() : i.disable();
            
            m = this._toolsMenu;
            if (m) {
                if (i = m.getItem("spoolFiles")) (isConnected && window.JSMINFO && !AXES.EXTHOST_TSInterface) ? i.enable() : i.disable();
                if (i = m.getItem("dataExplorer")) (isConnected &&  window.JSMINFO && !AXES.EXTHOST_TSInterface) ? i.enable() : i.disable();
            }
            
            m = this._helpMenu = this.menuCollection.menusById["help"];
            if (m) 
            {
                var sm = (i = m.getItem("diagnostics")) ? i.subMenu : null;
                if (sm) 
                {
                    if (i = sm.getItem("transaction")) (isConnected) ? i.enable() : i.disable();
                    if (i = sm.getItem("sethosttrace")) (isConnected) ? i.enable() : i.disable();
                    if (i = sm.getItem("clearhosttrace")) (isConnected) ? i.enable() : i.disable();
                    if (i = sm.getItem("gethosttrace")) (isConnected) ? i.enable() : i.disable();
                }
            }
        }
        if (this._flags.autoGUIOn != autoGUIOn) 
        {
            i = this._displayMenu.getItem("autogui");
            if (i) i.setChecked(autoGUIOn);
        }
        
        this._flags.isConnected = isConnected;
        this._flags.autoGUIOn = autoGUIOn;
    },
    _sessionNew: function()
    {
        window.open(window.location, "", "scrollbars=no,height=500,width=800,resizable=yes");
    },
    _sessionSuspend: function()
    {
        AXES.Terminal.reset();
        var fld = AXES.Terminal.getFieldById("loginReconnect");
        if (fld) fld.setValue("Y");
        AXES.Terminal.messages.push(TEXT.Core("Session suspended."));
        AXES.Terminal.screenLoaded();
    },
    _sessionBreak: function()
    {
        AXES.Terminal.doSysPost("SYSREQ");
    },
    _sessionClose: function()
    {
        AXES.closeSession(false);
    },
    _displayRefresh: function()
    {
        AXES.Terminal.doSysPost("REDRAW");
    },
    _displayAutogui: function(e, item)
    {
        AXES.Layout.AutoGUI.setAutoGUI(!AXES.Layout.AutoGUI.autoGUIOn);
    },
    _openSpoolDEWindow: function(serverInfo, windowName) 
    {
        var l, port, url;
        if (serverInfo.enabled) 
        {
            var l = window.location,
            port = l.protocol == "https:" ? serverInfo.sslport : serverInfo.port,
            url = [l.protocol, "//", l.hostname, ((port) ? ":" + port : ""), 
            serverInfo.uri, "?lang=", TEXT.currentLanguage];
            
            window.open(url.join(""),windowName, "");
        }
    },
    _toolsSpool: function(e, item) 
    {
        JSMINFO.startWS2()
    },
    _toolsDataExplorer: function(e, item) {
        JSMINFO.startDE2();
    },
    _setTheme: function(e, item)
    {
        AXES.UI.setTheme(item.id);
        if (('localStorage' in window) && window['localStorage'] !== null) localStorage['axTheme'] = item.id;
    },
    buildThemesMenu: function(subM)
    {
        var self = this;
        
        if (AXES.UI._allowThemesMenu) 
        {
            $.ajax({
                url: "/agi/lua5/extservice_3.lua",
                type: "get",
                success: function(data, textStatus)
                {
                    if (data.filelist && data.filelist instanceof Array) 
                    {
                        var currentTheme = $("#axThemeCSS").attr("href").replace(/(.*)\?.*/, "$1");
                        for (var i = 0, fileInfo; fileInfo = data.filelist[i]; i++) 
                        {
                            fileInfo.name = fileInfo.name.toLowerCase();
                            var href = "css/" + fileInfo.name;
                            var menuName = fileInfo.name.split(".")[0];
                            var theme = menuName.split("_")[1];
                            theme = theme.charAt(0).toUpperCase() + theme.substring(1);
                            
                            var item = self.addItem(subM, menuName, {
                                content: TEXT.Core(theme),
                                onClick: self._setTheme
                            });
                            item.setChecked(currentTheme == href);
                        }
                    }
                    else 
                    {
                        AXES.Debug.error(TEXT.core("Error retrieving themes list."));
                        subM.parentMenuItem.disable();
                    }
                },
                error: function(XMLHttpRequest, textStatus, errorThrown)
                {
                    AXES.Debug.error(TEXT.core("Error retrieving themes list."));
                    subM.parentMenuItem.disable();
                },
                data: {
                    path: "ts/ts2/css/",
                    prefix: "theme_",
                    suffix: ".css"
                },
                dataType: "json"
            });
        }
        else 
        {
            subM.parentMenuItem.disable();
        }
    }
};

