/*********************************************************************
 *                                                                   *
 * Copyright (C) LANSA Group. All rights reserved                    *
 *                                                                   *
 *********************************************************************/

/**
 * @fileOverview contains code that is only loaded and used in developer mode.
 */

/**
 * Manages the Developer UI
 * @namespace
 * <strong>Note:</strong> This namespace is only available in developer mode.
 */
AXES.Dev =
{
    /**
     * Indicates if the screen is currently locked for editing
     * @type Boolean
     */
    screenLocked: false,
    /**
     * The namespace of the currently active developer tab (AXES.Dev.ScreensTab, AXES.Dev.ExtensionsTab or AutoGUITab)
     * @type Object
     */
    _activeTab: null,
    /**
     * Initialises The AXES Dev object and sets up handling of clicks in the screen when locked.
     * @returns {undefined}
     */
    init: function()
    {
        /**
         * Indicates that design mode has been turned on or off.
         * @name AXES.Dev#DesignModeChanged
         * @event
         */
        jQuery.extend( true, this, AXES.EventsBase, { _eventNames: { DesignModeChanged: true }, _listeners : {} } );
        $( "#terminalContent" ).on(
            "click",
            ".ax-terminal-field",
            function( e )
            {
                if ( !AXES.Dev.screenLocked )
                {
                    return;
                }
                AXES.Dev.selectField( $( this )[ 0 ].id, e );
                e.preventDefault();
                e.stopPropagation();
                return false;
            }
        );
        $(document).bind('keydown', AXES.Dev._handleKeyDown);
    },

    onTabSelected: function( e, ui )
    {
        var dev = AXES.Dev;
        var selectedFields = $( "#axDesignModeHelper" ).data( "attachedTo" ) || [];
        if ( dev._activeTab )
            dev._activeTab._selectedFields = selectedFields;
        switch ( ui.newPanel.attr( "id" ) )
        {
        case "ax-tab-screens":
            dev._activeTab = AXES.Dev.ScreensTab;
            break;
        case "ax-tab-extensions":
            dev._activeTab = AXES.Dev.ExtensionsTab;
            break;
        case "ax-tab-autogui":
            dev._activeTab = AXES.Dev.AutoGUITab;
            break;
        }
        dev._activeTab.onShow();
        // Each tab has its own set of selected fields (undo management across multiple tabs gets too complex otherwise)
        // so clear the currents selection and restore the set saved for this tab.
        for ( var i = 0, id; id = selectedFields[ i ]; i++ ) {
            AXES.Dev.removeFldHighlight(id);
        }
        
        $( "#axDesignModeHelper" ).data( "attachedTo", [] );
        if ( dev._activeTab._selectedFields )
            dev._changeSelectedFields( dev._activeTab._selectedFields );
    },
    /**
    * handles keyboard arrow key to move selected element in design mode
    **/
    _handleKeyDown: function(e)
    {    
        if (AXES.Dev.screenLocked) {
            var helperDiv = $( "#axDesignModeHelper" ), z = AXES.Layout._zoom;
            var attachedTo = helperDiv.data( "attachedTo" ); 
            if (attachedTo) {
                var numItems = attachedTo.length;
                for (var i = 0; i < numItems; i++) {
                    var el = $("#" + attachedTo[i]);
                    var propVal = {
                        left : el.css("left"),
                        top : el.css("top")
                    };
                    var p = el.position();
                    var charCode = (e.which) ? e.which : e.keyCode;
                    switch(charCode) {
                    case 37: //left arrow key
                            p.left -= z; 
                        break;
                    case 38: //Up arrow key
                            p.top -= z;
                        break;
                    case 39: //Right arrow key
                            p.left += z;
                        break;
                    case 40: //Down arrow key
                            p.top += z;
                        break;
                    }
                    propVal.left = Math.ceil(p.left / z) + "px";
                    propVal.top = Math.ceil(p.top / z ) + "px";
                    AXES.Dev.updateFieldStyle(attachedTo[i], propVal);
                }
            }
            e.preventDefault();
            e.stopPropagation();
        }
    },
    /**
     * Updates the screen display to show the supplied fields as selected. If the first field in the array has changed, the current
     * tab is notified that the selection has changed.
     * @param {String[]} newSet Array of field IDs in selected order
     */
    _changeSelectedFields: function( newSet )
    {
        var oldSet = $( "#axDesignModeHelper" ).data( "attachedTo" ) || [];
        for ( var i = 0, id; id = oldSet[ i ]; i++ )
            $( "#" + id ).removeClass( "ui-state-highlight" );
        $( "#axDesignModeHelper" ).data( "attachedTo", newSet );
        this.applySelectedState();
        var oldMain = oldSet.length > 0 ? oldSet[ 0 ] : "";
        var newMain = newSet.length > 0 ? newSet[ 0 ] : "";
        if ( newMain != oldMain )
            if ( AXES.Dev._activeTab.selectFieldById )
                AXES.Dev._activeTab.selectFieldById( newMain );
        $( "#ax-dev-designMenuButton" ).button( newSet.length < 2 ? "disable" : "enable" );
    },
    /**
     * Selects a field or fields. If the event parameter indicates that the Ctrl-key is pressed, the field will be added to the current selection.
     * Otherwise it replaces the current selection.
     * @param {String} fieldId The id of the field selected
     * @param {Event} [event] The jQuery event object of the event that triggerred the select (if any).
     */
    selectField: function( fieldId, event )
    {
        fieldId          = fieldId || "";
        var helperDiv    = $( "#axDesignModeHelper" );
        var attachedTo   = [];

        if ( helperDiv.length )
        {
            helperDiv.zIndex(this.screenLocked ? 10001 : -99);
            attachedTo = helperDiv.data( "attachedTo" );
            if ( attachedTo && attachedTo.length && !( event && event.ctrlKey ) )
            {
                var id;
                while ( id = attachedTo.pop() )
                {
                    this.removeFldHighlight(id);
                }
            }
        }
        else
        {
            helperDiv = $( "<div>" )
                .attr(
                    {
                        id   : "axDesignModeHelper",
                        style: "display:none;"
                    }
                )
                .appendTo(
                    AXES.Terminal.windows.length > 1 ?
                    "#" + AXES.Terminal.windows[ AXES.Terminal.windows.length - 1 ].id :
                    "#terminalContent"
                );
            helperDiv.zIndex(this.screenLocked ? 10001 : -99);
            helperDiv.append( $( "<div>" ).attr( { id : "axDesignModeHelperBackground" } ) );
            helperDiv.draggable(
                {
                    containment     : helperDiv.parent().parent(),
                    cursor          : "move",
                    refreshPositions: true,
                    start           : AXES.Dev.helperDivDragStart,
                    drag            : AXES.Dev.helperDivDrag,
                    stop            : AXES.Dev.helperDivDragStop
                }
            );
            helperDiv.resizable(
                {
                    ghost  : false,
                    handles: "all",
                    start  : AXES.Dev.helperDivResizeStart,
                    resize : AXES.Dev.helperDivResize,
                    stop   : AXES.Dev.helperDivResizeStop
                }
            );
            helperDiv.data( "attachedTo", attachedTo );
        }

        var index = jQuery.inArray( fieldId, attachedTo );
        if ( 0 >  index ) //check if fieldId does not exist in attachedTo
            attachedTo.push( fieldId );
        else  // deselect the field and remove highlight
        {
            attachedTo.splice( index, 1 );
            this.removeFldHighlight(fieldId);
        }
        
        if (this._activeTab.selectFieldById)
            this._activeTab.selectFieldById( fieldId );

        this.applySelectedState();
        
        //if multiple elements is selected by pressing the CTRL key prevent them from 
        //changing any properties from field extension property. multi selected elements work only with 
        //alignment tools. subfile column is considered a single selected element.
        if (attachedTo.length < 2) {
            AXES.Dev.ExtensionsTab.FieldExtReadOnly = false;
        } else {
            AXES.Dev.ExtensionsTab.FieldExtReadOnly = true;
        }
        $( "#ax-dev-designMenuButton" ).button( attachedTo.length < 2 ? "disable" : "enable" );
    },
    getSelectedFieldId: function() {
        var attachedTo = $("#axDesignModeHelper").data("attachedTo");
        return (attachedTo && attachedTo.length) ? attachedTo[0] : null;
    },
    helperDivResizeStart: function( event, ui )
    {
        var windows    = AXES.Terminal.windows;
        var $container = 1 < windows.length ? $( "#" + windows[ windows.length - 1 ].id ) : $( "#terminalContent" );
        var attachedTo = ui.helper.data( "attachedTo" );
        var startSizes = [];
        var z          = AXES.Layout._zoom;
        var p          = ui.position;
        p.top          = Math.round( p.top / z );
        p.left         = Math.round( p.left / z );

        AXES.Dev.setFrame( ui.helper, p, "px" );

        // Save initial positions of each selected element
        for ( var i = 0, id, fieldDiv, frame; id = attachedTo[ i ]; i++ )
        {
            fieldDiv     = $container.find( "#" + id );
            frame        = fieldDiv.position();
            frame.top    = Math.round( frame.top / z );
            frame.left   = Math.round( frame.left / z );
            frame.width  = fieldDiv.innerWidth();
            frame.height = fieldDiv.innerHeight();

            startSizes.push( frame );
        }

        ui.helper
            .data( "resizeStartSizes", startSizes )
            .data( "flags", { n: false, w: false } );
    },
    helperDivResize: function( event, ui )
    {
        // jQuery doesn't know about zoom. This results in a size and position calculation that is the 
        // original size and position in unzoomed co-ordinates plus the mouse delta in zoomed co-ordinates.
        // So we need to fix it.

        // originalSize is correct but position is in zoomed co-ordinates.
        var z              = AXES.Layout._zoom;
        var originalFrame  = jQuery.extend( {}, ui.originalPosition, ui.originalSize );
        originalFrame.top  = Math.round( originalFrame.top / z );
        originalFrame.left = Math.round( originalFrame.left / z );

        // If resizing is in the south or east directions, the position is in original co-ordinates
        // If resizing is in north or west directions the position has been calculated by adding the mouse offset
        // in zoomed co-ordinates to the original position in un-zoomed co-ordinates. Fix it.
        var frame  = jQuery.extend( {}, ui.position, ui.size );
        var helper = ui.helper;
        var flags  = helper.data( "flags" );

        if ( flags.n || frame.top != originalFrame.top )
        {
            frame.top = originalFrame.top + Math.round( ( frame.top - ui.originalPosition.top ) / z );
            flags.n   = true;
        }
        if ( flags.w || frame.left != originalFrame.left )
        {
            frame.left = originalFrame.left + Math.round( ( frame.left - ui.originalPosition.left ) / z );
            flags.w    = true;
        }

        var os       = ui.originalSize;
        frame.width  = Math.round( os.width + Math.round( ( frame.width - os.width ) / z ) );
        frame.height = Math.round( os.height + Math.round( ( frame.height - os.height ) / z ) );

        var dev = AXES.Dev;
        dev.setFrame( ui.helper, frame, "px" );

        // Calculate the way the frame has changed and apply to all the selected elements.
        var offsets =
        {
            height: frame.height - originalFrame.height,
            left  : frame.left   - originalFrame.left,
            top   : frame.top    - originalFrame.top,
            width : frame.width  - originalFrame.width
        };
        var attachedTo = helper.data( "attachedTo" );
        var startSizes = helper.data( "resizeStartSizes" );
        var windows    = AXES.Terminal.windows;
        var $container = 1 < windows.length ? $( "#" + windows[ windows.length - 1 ].id ) : $( "#terminalContent" );

        for ( var i = 0, id, startSize; id = attachedTo[ i ], startSize = startSizes[ i ]; i++ )
            dev.setFrame(
                $container.find( "#" + id ),
                {
                    left  : startSize.left   + offsets.left,
                    top   : startSize.top    + offsets.top,
                    width : startSize.width  + offsets.width,
                    height: startSize.height + offsets.height
                },
                "px"
            );
    },
    helperDivResizeStop: function( event, ui )
    {
        var attachedTo = ui.helper.data( "attachedTo" );
        var windows    = AXES.Terminal.windows;
        var $container = 1 < windows.length ? $( "#" + windows[ windows.length - 1 ].id ) : $( "#terminalContent" );
        var dev        = AXES.Dev;
        for ( var i = 0, id, $el; id = attachedTo[ i ]; i++ )
        {
            $el = $container.find( "#" + id );
            dev.updateFieldStyle(
                id,
                {
                    left  : $el.css( "left" ),
                    top   : $el.css( "top" ),
                    width : $el.css( "width" ),
                    height: $el.css( "height" )
                }
            );
        }

        var numItems = attachedTo.length;
        var tab      = dev._activeTab;
        var undo     = tab._undo;

        if ( undo )
            undo.addUndo(
                {
                    context         : tab,
                    redoArgs        : [],
                    undoArgs        : [],
                    updateDirtyState: true,

                    redo: function()
                    {
                        for ( var i = 0; i < numItems; i++ )
                            undo.doRedo();
                    },
                    undo: function()
                    {
                        for ( var i = 0; i < numItems; i++ )
                            undo.doUndo();
                    }
                }
            );
    },
    helperDivDragStart: function( event, ui )
    {
        var z              = AXES.Layout._zoom;
        var attachedTo     = ui.helper.data( "attachedTo" );
        var startPositions = [];
        var windows        = AXES.Terminal.windows;
        var $container     = 1 < windows.length ? $( "#" + windows[ windows.length - 1 ].id ) : $( "#terminalContent" );

        // Save initial positions of each selected element
        for ( var i = 0, id, fieldDiv, pos; id = attachedTo[ i ]; i++ )
        {
            fieldDiv = $container.find( "#" + id );
            pos      = fieldDiv.position();
            pos.top  = Math.round( pos.top / z );
            pos.left = Math.round( pos.left / z );

            startPositions.push( pos );
        }

        ui.helper
            .data( "dragStartPositions", startPositions )
            .data(
                "helperDragStart",
                {
                    top : ui.position.top,
                    left: ui.position.left
                }
            );
    },
    helperDivDrag: function( event, ui )
    {
        var p  = ui.position;
        var op = ui.originalPosition;
        var z  = AXES.Layout._zoom;
        var offsets =
        {
            left: Math.round( ( p.left - op.left ) / z ),
            top : Math.round( ( p.top - op.top ) / z )
        };

        var windows        = AXES.Terminal.windows;
        var $container     = 1 < windows.length ? $( "#" + windows[ windows.length - 1 ].id ) : $( "#terminalContent" );
        var attachedTo     = ui.helper.data( "attachedTo" );
        var startPositions = ui.helper.data( "dragStartPositions" );

        for ( var i = 0, id, pos, dev = AXES.Dev; id = attachedTo[ i ], pos = startPositions[ i ]; i++ )
            dev.setFrame(
                $container.find( "#" + id ),
                {
                    left: pos.left + offsets.left,
                    top : pos.top + offsets.top
                },
                "px"
            );

        p.left = Math.round( op.left / z ) + offsets.left;
        p.top = Math.round( op.top / z ) + offsets.top;
    },
    helperDivDragStop: function( event, ui )
    {
        var dev        = AXES.Dev;
        var windows    = AXES.Terminal.windows;
        var $container = 1 < windows.length ? $( "#" + windows[ windows.length - 1 ].id ) : $( "#terminalContent" );
        var attachedTo = ui.helper.data( "attachedTo" );
        var numItems   = attachedTo.length;

        for ( var i = 0, el; i < numItems; i++ )
        {
            el = $container.find( "#" + attachedTo[ i ] );
            dev.updateFieldStyle(
                attachedTo[ i ],
                {
                    left: el.css( "left" ),
                    top : el.css( "top" )
                }
            );
        }

        var tab = dev._activeTab;
        var undo = tab._undo;
        if ( undo )
            undo.addUndo(
                {
                    context         : tab,
                    redoArgs        : [],
                    undoArgs        : [],
                    updateDirtyState: true,

                    redo: function()
                    {
                        for ( var i = 0; i < numItems; i++ )
                            undo.doRedo();
                    },
                    undo: function()
                    {
                        for ( var i = 0; i < numItems; i++ )
                            undo.doUndo();
                    }
                }
            );
    },
    setFrame : function(element, frame, units) {
        var css = {};
        if (frame.left)
            css.left = frame.left + units;
        if (frame.top)
            css.top = frame.top + units;
        if (frame.right)
            css.right = frame.right + units;
        if (frame.bottom)
            css.bottom = frame.bottom + units;
        if (frame.width)
            css.width = frame.width + units;
        if (frame.height)
            css.height = frame.height + units;
        $(element).css(css);
    },
    /**
     * Reposition the helper and reapply the highlight state to the field it is attached to.
     */
    applySelectedState: function()
    {
        var helperDiv = $( "#axDesignModeHelper" );

        if ( !helperDiv.length )
        {
            return;
        }

        var attachedTo = helperDiv.data( "attachedTo" );
        if ( !attachedTo || !attachedTo.length )
        {
            helperDiv.hide();
            return;
        }

        var windows = AXES.Terminal.windows;
        var $container = 1 < windows.length ? $( "#" + windows[ windows.length - 1 ].id ) : $( "#terminalContent" );
        if ( this._activeTab === this.ExtensionsTab )
        {
            var z = AXES.Layout._zoom;
            var fieldDiv = $container.find( "#" + attachedTo[ 0 ] );
            var p = fieldDiv.position();
            p.top = ( p.top / z ) + "px";
            p.left = ( p.left / z ) + "px";
            p.width = fieldDiv[ 0 ].offsetWidth + "px";
            p.height = fieldDiv[ 0 ].offsetHeight + "px";
            helperDiv.css( p ).show();
        }
        else
            helperDiv.hide();

        for (var i = 0, id; id = attachedTo[i]; i++) {
            this.addFldHighlight(id);
        }
    },
    
    addFldHighlight : function(id) {
        if ( $( "#" + id ).hasClass( "ui-state-selected" ) )
            $( "#" + id ).removeClass( "ui-state-selected" );
        $("#" + id).addClass("ui-state-highlight");
        var fld = AXES.Dev.getFieldById( id );
        if ( fld ) fld.setHighlight( true );
        // check if field added to a selection belong to a subfile columns
        if (fld && fld.recordFormat && fld.recordFormat.cols ) {
            var col = fld.recordFormat.getSubfileColumn(fld);
            var sbColumn = col[Object.keys(col)[0]];
            var attachedTo = $("#axDesignModeHelper").data( "attachedTo" ) || [];
            // highlight selected subfile columns
            for (var cfld in sbColumn.items) {
                if (sbColumn.items[cfld]) {
                    var itemId = sbColumn.items[cfld].id;
                    if ( (id !== itemId) && (0 > jQuery.inArray( itemId, attachedTo)) &&
                         !($( "#" + itemId ).hasClass( "ui-state-selected" )) )
                        $( "#" + itemId ).addClass("ui-state-selected");
                }
            }
        }
    },
    
    removeFldHighlight: function(id) {
        $("#" + id).removeClass("ui-state-highlight");
        var fld = AXES.Dev.getFieldById( id );
        if ( fld ) fld.setHighlight( false );
        // check if field added to a selection belong to a subfile columns
        if (fld && fld.recordFormat && fld.recordFormat.cols ) {
            var col = fld.recordFormat.getSubfileColumn(fld);
            var sbColumn = col[Object.keys(col)[0]];
            // highlight selected subfile columns
            for (var cfld in sbColumn.items) {
                if ( sbColumn.items[cfld] )
                    $( "#" + sbColumn.items[cfld].id ).removeClass("ui-state-selected");
            }
        }
    },
    
    getFieldById : function(id) {
        var fld = AXES.Terminal.getFieldById(id);
        if (!fld) 
        {
            for (var i = 0, fld; fld = AXES.Layout.userFields[i]; i++) 
            {
                if (fld.id == id)
                    break;
            }
        }
        return fld;
    },
    onFieldRowHoverStart : function() {
        var t = $(this);
        var id = this.id.split(".")[1];
        var fld = AXES.Dev.getFieldById(id);
        t.addClass("ui-state-hover");
        if (fld)
            $(fld.terminalDiv).addClass("ui-state-hover");
    },
    onFieldRowHoverEnd : function() {
        var t = $(this);
        var id = this.id.split(".")[1];
        var fld = AXES.Dev.getFieldById(id);
        t.removeClass("ui-state-hover");
        if (fld)
            $(fld.terminalDiv).removeClass("ui-state-hover");
    },
    openHelpDocument: function(fileName, fileExtension, docTarget)
    {
        var language = TEXT.currentLanguage;
        if (language == null)
            language = "";
        language = language.toLowerCase().trim();
        var winOptions = "width=800, height=500, menubar=no, resizable=yes, scrollbars=yes, status=no, toolbar=no";
        var helpUrl = this._helpDocUrl(fileName + "." + fileExtension);
        if (docTarget)
            helpUrl += "#nameddest=" + docTarget + "&pagemode=none";

        if (language != "en")
        {
            var helpUrl2 = this._helpDocUrl(fileName + "_" + language + "." + fileExtension);
            if (docTarget)
                helpUrl2 += "#nameddest=" + docTarget + "&pagemode=none";
            $.ajax(
            {
                url : helpUrl2,
                type : "HEAD",
                success: function() 
                {
                    helpUrl = helpUrl2;
                },
                complete: function() 
                {
                    window.open(helpUrl, "axhelp", winOptions);
                }
            });
        }
        window.open(helpUrl, "axhelp", winOptions);
    },
    _helpDocUrl: function(fileName)
    {
        return "/docs/" + fileName;
    },
    /**
     * Upload a file to the current project directory
     * @param {String} fileName The file name, excluding the extension
     * @param {String} ext The file extension
     * @param {String} content The file content.
     */
    uploadFileToServer: function(fileName, ext, content)
    {
        var data =
        {
            definitionSet : AXES.definitionSet,
            file : fileName,
            suffix : "." + ext,
            content : content
        };
        var dataList = this.encodeDataForPost(data, "content", 4000, "append=y");
        var returnData = null;
        for (var i = 0, chunk; chunk = dataList[i]; i++) 
        {
            $.ajax({
                url : "/agi/lua5/extservice_1.lua",
                type : "POST",
                async : false,
                dataType : "json",
                data : chunk,
                success: function(data, textStatus)
                {
                    returnData = data;
                },
                error: function(XMLHttpRequest, textStatus, errorThrown)
                {
                    var msg = "";
                    if (errorThrown)
                        msg = TEXT.Core("Error occurred uploading %s1.%s2 to the server: %s3", fileName, ext, errorThrown);
                    else if (textStatus == "parsererror") 
                        msg = TEXT.Core("Unspecified error occurred uploading %s1.%s2 to the server.", fileName, ext);
                    else 
                        msg = TEXT.Core("HTTP error %s1 uploading %s2.%s3 to the server: %s4", XMLHttpRequest.status, fileName, ext, XMLHttpRequest.statusText);

                    if ( i ) 
                        msg += "\n\nData transfer to the server has been interrupted. Some data may have been saved on the server - manually check the integrity of the data"
                    AXES.Trace.error(msg);
                    returnData = {
                        error : true,
                        errorDesc : msg
                    };
                }
            });
            if (returnData.error)
                break;
        }
        return (returnData.error) ? returnData.errorDesc : "";

    },
    /**
     * Deletes a file from the current project directory
     * @param {String} fileName The file name, excluding the extension
     * @param {String} ext The file extension
     * @return {String} Empty string if successful, an error message otherwise.
     */
    deleteFileFromServer: function(fileName, ext)
    {
        var msg = "";
        $.ajax({
            url : "/agi/lua5/extservice_1.lua",
            type : "POST",
            async : false,
            dataType : "json",
            data : {
                definitionSet : AXES.definitionSet,
                file : fileName,
                suffix : "." + ext,
                del : "1"
            },
            success: function()
            {
            },
            error: function(XMLHttpRequest, textStatus, errorThrown)
            {
                if (errorThrown) 
                    msg = TEXT.Core("Error occurred deleting %s1.%s2 from the server: #s3", fileName, ext, errorThrown);
                else if (textStatus == "parsererror") 
                    msg = TEXT.Core("Unspecified error occurred deleting %s1.%s2 from the server.", fileName, ext);
                else 
                    msg = TEXT.Core("HTTP error %s1 deleting %s2.%s3 from the server: %s4", XMLHttpRequest.status, fileName, ext, XMLHttpRequest.statusText);
                AXES.Trace.error(msg);
            }
        });
        return msg;
    },
    /**
     * Encodes a supplied collection of name/value pairs ready for a POST.
     *
     * Large amounts of data can be split into chunks by specifying a content
     * property name, chunk size and optional append indicator. In this scenario
     * the property indicated by contentPropertyName will be split into chunks
     * such that the size of the data sent will be approximately chunkSize. Each
     * chunk will contain all of the properties (other than the content) in full
     * plus a slice of the content. The appendIndicator will be included in all
     * but the first chunk.
     * @param {String} data
     * @param {String} contentPropertyName
     * @param {Integer} chunkSize
     * @param {String} appendIndicator
     * @returns [{String}] An array of chunks
     */
    encodeDataForPost: function(data, contentPropertyName, chunkSize, appendIndicator)
    {
        var result = [], paramsToRepeat="", delimiter="", content="", startIndex=0, prefix,
        contentLen, contentChunkSize, chunkData;

        if (data != null)
        {
            for (var p in data)
            {
                if (p != contentPropertyName) 
                {
                    paramsToRepeat += delimiter + p + "=" + encodeURIComponent(data[p]);
                    delimiter = "&";
                }
                else 
                    content = data[p];
            }
        }

        if (contentPropertyName != null)
        {
            startIndex = 0;
            prefix = paramsToRepeat + delimiter + contentPropertyName + "=";
            contentChunkSize = chunkSize - prefix.length;

            contentLen = content.length;
            while (startIndex < contentLen)
            {
                chunkData = content.substr(startIndex, contentChunkSize);
                result.push(prefix + encodeURIComponent(chunkData));
                if (startIndex == 0 && appendIndicator != null) 
                {
                    prefix = paramsToRepeat + delimiter + appendIndicator + "&" + contentPropertyName + "=";
                    contentChunkSize = chunkSize - prefix.length;
                }
                startIndex += chunkData.length;
            }
        }
        else 
            result.push(paramsToRepeat);

        return result;
    },
    /**
     * Updates the terminal with developer mode controls and sets up event listeners necessary to support developer mode.
     */
    prepTerminalForDev : function() {
        var $b1, $b2, $b3, $b4, $s;
        $b1 = $("<input>").attr({
            type : "radio",
            id : "ax-dev-radioLockScreen1",
            name : "ax-dev-radioLockScreen",
            value : "lock"
        });
        $b2 = $("<input>").attr({
            type : "radio",
            id : "ax-dev-radioLockScreen2",
            name : "ax-dev-radioLockScreen",
            value : "unlock",
            checked : "checked"
        });
        $b3 = $("<button>").attr({
            type : "button",
            id : "ax-dev-lockScreenMenu",
            title : TEXT.Dev("GridSettings...")
        });

        $("#devtab-screens").text(TEXT.dev("Screens"));
        $("#devtab-extensions").text(TEXT.dev("Extensions"));
        $("#devtab-autogui").text(TEXT.dev("AutoGUI"));

        $s = $("<div id='ax-dev-lockScreenButtonset'></div>").append($b1).append($("<label>").attr("for", "ax-dev-radioLockScreen1").text(TEXT.Dev("Lock Screen"))).append($b2).append($("<label>").attr("for", "ax-dev-radioLockScreen2").text(TEXT.Dev("Unlock Screen")));
        $s.buttonset();
        $b2.button({
            text : false,
            icons : {
                primary : "ui-icon-unlocked"
            }
        }).click(function(e) {
            AXES.Dev.setScreenLocked(false);
        });
        $b1.button({
            text : false,
            icons : {
                primary : "ui-icon-locked"
            }
        }).click(function(e) {
            AXES.Dev.setScreenLocked(true);
        });
        $b3.button({
            text : false,
            icons : {
                primary : "ui-icon-grip-dotted-horizontal"
            }
        }).click(this.onScreenLockSettingsClicked);

        $("#terminalOuter .header").append($b3).append($s);

        $b4 = $("<button>").attr({
            type : "button",
            id : "ax-dev-designMenuButton",
            title : TEXT.Dev("Alignment")
        });
        $b4.button({
            text : false,
            icons : {
                primary : "ui-icon-gear"
            }
        }).click(this.onDesignMenuButtonClicked);

        $("#terminalOuter .header").append($b4);

        $("#ax-dev-designMenu").menu({
            select : this.onDesignMenuItemSelected
        });

        AXES.Terminal.addListener("ScreenArrivedPreRender", this.onScreenArrivedPreRender, this);
        AXES.Terminal.addListener("ScreenWillPost", this.onScreenWillPost, this);

        this._gridSettings = {
            front : false,
            size : 10
        };
        if (('localStorage' in window) && window['localStorage'] !== null) 
        {
            var savedGridSettings = localStorage['axDevGridSettings'];
            if (savedGridSettings)
                this._gridSettings = JSON.parse(savedGridSettings);
        }

        // prevent keydown on lock and unlock button
        $("#ax-dev-radioLockScreen1, #ax-dev-radioLockScreen2").keydown(function(e)
        {
            var arrowKeys = [37, 38, 39, 40];
            if (arrowKeys.indexOf(e.which) !== -1)
            {
                $(this).blur();
                return false;
            }
        });
    },
    onScreenArrivedPreRender : function(eventObject) {
        $("#ax-dev-radioLockScreen2").attr("checked", "checked").button("refresh");
        this.setScreenLocked(false);
        $("#ax-dev-lockScreenButtonset").buttonset(AXES.Terminal.isLogin ? "disable" : "enable");
        $("#ax-dev-designMenuButton").button("disable");
    },
    onScreenWillPost : function(eventObject) {
        if (this.screenLocked) {
            AXES.SysTrace.info(TEXT.dev("Cannot post data while screen is locked."));
            return false;
        }
    },
    onScreenZoomChanged : function() {
        if (this.screenLocked) {
            this._setDesignGrid(false);
            $("#axDesignGrid").remove();
            this._setDesignGrid(true);
        }
    },
    _setDesignGrid: function( lockState )
    {
        if ( !lockState )
        {
            $( "#axDesignGrid, #axDesignModeHelper" ).remove();
            $( "#ax-terminal-setzoom" ).prop( "disabled", false );
            $( "#ax-terminal-zoomcontrols .ui-icon" ).removeClass( "ui-state-disabled" );

            return;
        }

        var $span = $( "#axDesignGrid" );
        if ( $span.length )
        {
            $span.show();
            $( "#ax-terminal-setzoom" ).prop( "disabled", "disabled" );
            $( "#ax-terminal-zoomcontrols .ui-icon" ).addClass( "ui-state-disabled" );

            return;
        }

        var windows = AXES.Terminal.windows;
        var $designLayout = ( 1 < windows.length ) ?
                            $( "#" + windows[ windows.length - 1 ].id ).children().eq( 0 ) :
                            $( "#terminalContent" );

        var gs = this._gridSettings;
        var w  = $designLayout.outerWidth();
        var h  = $designLayout.outerHeight();
        $span = $(
                    '<span \
                        id="axDesignGrid" \
                        class="' + ( gs.front ? "axDesignGridFront" : "" ) +
                        '" style="width:' + ( w - 2 ) + 'px; height:' + ( h - 2 ) + 'px; ' + ( gs.front ? "" : "zIndex:-99;" ) + '"> \
                    </span>'
                );

        var size = gs.size;
        for ( var x = size - 2; x < w; x += size )
            $span.append( '<div class="axDesignGridLineVert" style="left:' + x + 'px"></div>' );
        for ( var x = size - 2; x < h; x += size )
            $span.append( '<div class="axDesignGridLineHoriz" style="top:' + x + 'px"></div>' );

        $designLayout.append( $span );

        $( "#ax-terminal-setzoom" ).prop( "disabled", "disabled" );
        $( "#ax-terminal-zoomcontrols .ui-icon" ).addClass( "ui-state-disabled" );
    },
    setScreenLocked: function( lockState )
    {
        if ( this.screenLocked == lockState )
        {
            return;
        }

        this.screenLocked = lockState;
        var layout = AXES.Layout;
        layout.designMode( lockState );
        this.fire( "DesignModeChanged", [ { designMode: lockState } ] );
        this.doLayout();
        var windows = AXES.Terminal.windows;
        var popupWindow = 1 < windows.length;

        if ( lockState )
        {
            layout.addListener( "ZoomChanged", this.onScreenZoomChanged, this );
            $( popupWindow ? ( "#" + windows[ 1 ].id ) : "#terminalWrapper" ).addClass( "ui-designer-pane" );
            layout.Cursor.hide();
            if ( this._activeTab.selectedFieldRow )
                this._activeTab.selectFieldAndRow( this._activeTab.selectedFieldRow[ 0 ].id );
        }
        else
        {
            layout.Cursor.show();
            $( popupWindow ? ( "#" + windows[ 1 ].id ) : "#terminalWrapper" ).removeClass( "ui-designer-pane" );
            layout.removeListener( "ZoomChanged", this.onScreenZoomChanged );
            if ( this._activeTab.selectedFieldRow && this._activeTab.getFieldRowFromRowId)
                this._activeTab.selectFieldRow( this._activeTab.getFieldRowFromRowId( this._activeTab.selectedFieldRow[ 0 ].id ) )
        }
        this._setDesignGrid( lockState );
    },
    onScreenLockSettingsClicked : function(event) {
        var self = AXES.Dev, gs = self._gridSettings;

        AXES.UI.clearDialogs();
        var dialogOptions = {
            autoOpen : true,
            closeOnEscape : true,
            bgiframe : true,
            modal : true,
            position : "top",
            resizable : false,
            title : TEXT.Dev("GridSettings"),
            buttons : {},
            close : function() {
                $(this).dialog("destroy");
            }
        }
        dialogOptions.buttons[TEXT.Core("OK")] = function()
        {
            var oldSize = gs.size;
            gs.size = parseInt($("#ax-devGridSettings-Size").val()) || gs.size;
            gs.front = $("#ax-devGridSettings-Front")[0].checked;
            $(this).dialog("close");
            if (oldSize != gs.size) {
                $("#axDesignGrid").remove();
                self.setScreenLocked(self.screenLocked);
            } else {
                $("#axDesignGrid").attr("class", (gs.front) ? "axDesignGridFront" : "");
            }
            if (('localStorage' in window) && window['localStorage'] !== null) {
                localStorage['axDevGridSettings'] = JSON.stringify(gs);
            }
        };
        dialogOptions.buttons[TEXT.Core("Cancel")] = function() {
            $(this).dialog("close");
        };
        $("#ax-devGridSettings-Size").val(gs.size);
        $("#ax-devGridSettings-Front").attr((gs.front ? "checked" : ""));
        $("#ax-devGridSettingsDialog").dialog(dialogOptions);
    },
    onDesignMenuButtonClicked : function(event) {
        var menu = $("#ax-dev-designMenu");
        if (menu.is(":visible")) {
            menu.hide();
        } else {
            menu.show().position({
                my : "left top",
                at : "left bottom",
                of : this
            });
        }
        $(document).one("click", function() {
            menu.hide();
        });
        return false;
    },
    onDesignMenuItemSelected : function(event, ui) {
        var helperDiv = $("#axDesignModeHelper"), action = ui.item.data("action"), z = AXES.Layout._zoom,
            attachedTo, a, getVal, setVal, masterVal, i, l, el,
            exTab = AXES.Dev.ExtensionsTab;

        if (action == "")
        {
            return;
        }
        if (helperDiv.length == 0)
        {
            return;
        }
        attachedTo = helperDiv.data("attachedTo");
        l = attachedTo.length;
        if (l < 2)
        {
            return;
        }
        a = action.split(".");
        switch (a[0]) {
            case "VA":
                switch (a[1]) {
                    case "1":
                        getVal = function(el) {
                            return Math.ceil(el.position().top / z);
                        };
                        setVal = function(el, val) {
                            return {
                                "top" : val + "px"
                            };
                        };
                        break;
                    case "2":
                        getVal = function(el) {
                            return Math.ceil(el.position().top / z) + Math.ceil(el.height() / 2);
                        };
                        setVal = function(el, val) {
                            return {
                                "top" : (val - Math.ceil(el.height() / 2)) + "px"
                            };
                        };
                        break;
                    case "3":
                        getVal = function(el) {
                            return Math.ceil(el.position().top / z) + el.height();
                        };
                        setVal = function(el, val) {
                            return {
                                "top" : (val - el.height()) + "px"
                            };
                        };
                        break;
                }
                break;
            case "HA":
                switch (a[1]) {
                    case "1":
                        getVal = function(el) {
                            return Math.ceil(el.position().left / z);
                        };
                        setVal = function(el, val) {
                            return {
                                "left" : val + "px"
                            };
                        };
                        break;
                    case "2":
                        getVal = function(el) {
                            return Math.ceil(el.position().left / z) + Math.ceil(el.width() / 2);
                        };
                        setVal = function(el, val) {
                            return {
                                "left" : (val - Math.ceil(el.width() / 2)) + "px"
                            };
                        };
                        break;
                    case "3":
                        getVal = function(el) {
                            return Math.ceil(el.position().left / z) + el.width();
                        };
                        setVal = function(el, val) {
                            return {
                                "left" : (val - el.width()) + "px"
                            };
                        };
                        break;
                }
                break;
            case "E":
                switch (a[1]) {
                    case "1":
                        getVal = function(el) {
                            return el.width();
                        };
                        setVal = function(el, val) {
                            return {
                                "width" : val + "px"
                            };
                        };
                        break;
                    case "2":
                        getVal = function(el) {
                            return el.height();
                        };
                        setVal = function(el, val) {
                            return {
                                "height" : val + "px"
                            };
                        };
                        break;
                }
                break;
            case "D":
                switch (a[1]) {
                    case "1":
                        getVal = function(el) {
                            return Math.ceil(el.position().top / z);
                        };
                        setVal = function(el, val) {
                            return {
                                "top" : val + "px"
                            };
                        };
                        break;
                    case "2":
                        getVal = function(el) {
                            return Math.ceil(el.position().left / z);
                        };
                        setVal = function(el, val) {
                            return {
                                "left" : val + "px"
                            };
                        };
                        break;
                }
                break;
        }

        if (a[0] == "D") {
            var vals = [], curr, step, el;
            for ( i = 0; i < l; i++) {
                el = $("#" + attachedTo[i]);
                vals.push({
                    el : el,
                    val : getVal(el)
                });
            }
            vals.sort(function(a, b) {
                return a.val - b.val;
            })
            curr = vals[0].val;
            step = (vals[l - 1].val - curr) / (l - 1);

            for ( i = 0; i < l; i++) {
                AXES.Dev.updateFieldStyle(attachedTo[i], setVal(vals[i].el, Math.ceil(curr)));
                curr += step;
            }
            AXES.Dev.applySelectedState();
        } else {
            masterVal = getVal($("#" + attachedTo[0]));
            for ( i = 1; i < l; i++) {
                AXES.Dev.updateFieldStyle(attachedTo[i], setVal($("#" + attachedTo[i]), masterVal));
            }
        }
    },
    /**
     * A utility method to merge new styles into the existing style property of a field's basic properties eXtension
     * @param {String} fieldId The ID of the field to update
     * @param {Object} style An object containing the properties and values to merge.
     */
    updateFieldStyle : function(fieldId, style) {
        var exTab = AXES.Dev.ExtensionsTab, fld = AXES.Dev.getFieldById(fieldId), property = fld.extensions[0].getPropertyByName("style"),
            propVal = jQuery.extend(true, {}, property.getRawValue());

        if (propVal.s)
            propVal = exTab.oldModeVals[property._displayRowId] || {
                v : {}
            };
        jQuery.extend(true, propVal.v, style);
        exTab.updatePropertyValue(property, propVal, true);
        var ex = property._owner;
        if (ex && ex.isAutoGenerated) ex.isAutoGenerated = false;
        for (var i = 0; i < fld.extensions.length; i++) {
            exTab.refreshExtension(fld.extensions[i]);
        }
    },
    /*  refresh the layout 
     *  a method to redraw/render all extension to apply recent change on properties.
     */    
    refreshLayout: function() {
        var exTab = AXES.Dev.ExtensionsTab;
        var fldList = exTab.tabContent.find( "#ax-extensions-fieldListUL li" );

        fldList.each(
            function()
            {
                var extendableObject = $( this ).data( "ExtendableObject" );
                if (extendableObject) { 
                    for ( var j = 0, ex; ex = extendableObject.extensions[ j ]; j++ )
                    {
                        exTab.refreshExtension(ex);
                    }
                }
            }
        );
    }    
}
/**
 * Calls AXES.Layout.doLayout passing in the current working definition, if necessary.
 */
AXES.Dev.doLayout = function() {
    var exTab = this.ExtensionsTab, def = exTab.currentDefinition;
    if (exTab.isSaveNeeded()) {
        def = exTab.buildScreenDefinitionForSave();
    }
    AXES.Layout.doLayout(def);
    AXES.Dev.ExtensionsTab.redrawExTab();
}
/**
 * Manages and undo/redo stack.
 * The owner of this queue may have an updateDirtyState method. If it does, this method will
 * be called when an undo is added as well as when an undo/redo is performed.  The updateDirtyState method will take a single parameter that is the
 * amount to change the dirty state by. This will usually be +/- 1. A value of -10 (< -9) should reset the dirty state.
 * @constructor
 * @param {Object} owner The object (a developer tools tab) of this undo/redo queue.
 */
AXES.Dev.Undo = function(owner) {
    this._owner = owner;
    this._undoLog = [];
    this._redoLog = [];
    this._ignoreAdd = false;
    this._undoUIItems = [];
    this._redoUIItems = [];
    this.addUndoUIItem = function(item) {
        this._undoUIItems.push(item);
    };
    this.addRedoUIItem = function(item) {
        this._redoUIItems.push(item);
    };
    this.reset = function() {
        this._undoLog.length = 0;
        this._redoLog.length = 0;
        this.setUIItems();
    };
    this.addUndo = function(undoInfo) {
        if (this._ignoreAdd)
        {
            return;
        }
        var ul = this._undoLog;
        if (!undoInfo.context)
            undoInfo.context = window;
        ul.push(undoInfo);
        this._redoLog.length = 0;
        if (ul.length > 50) {
            ul.splice(50, ul.length - 50);
        }
        if (undoInfo.updateDirtyState && this._owner.updateDirtyState)
            this._owner.updateDirtyState(1);
        this.setUIItems();
    };
    this.doUndo = function() {
        if (this._undoLog.length > 0) 
        {
            var undoInfo = this._undoLog.pop();
            this._ignoreAdd = true;
            try 
            {
                undoInfo.undo.apply(undoInfo.context, undoInfo.undoArgs);
            } catch (e) {
                AXES.Debug.error(TEXT.Core("JavaScript error performing undo: %s1.", e.message));
            }
            this._ignoreAdd = false;
            this._redoLog.push(undoInfo);
            if (undoInfo.updateDirtyState && this._owner.updateDirtyState)
                this._owner.updateDirtyState(-1);
        }
        this.setUIItems();
    };
    this.doRedo = function() {
        if (this._redoLog.length > 0) 
        {
            var undoInfo = this._redoLog.pop();
            this._ignoreAdd = true;
            try 
            {
                undoInfo.redo.apply(undoInfo.context, undoInfo.redoArgs);
            } catch (e) {
                AXES.Debug.error(TEXT.Core("JavaScript error performing redo: %s1.", e.message));
            }
            this._ignoreAdd = false;
            this._undoLog.push(undoInfo);
            if (undoInfo.updateDirtyState && this._owner.updateDirtyState)
                this._owner.updateDirtyState(1);
        }
        this.setUIItems();
    };
    this.setUIItems = function() 
    {
        for (var i = 0, item; item = this._undoUIItems[i]; i++) {
            item.attr("disabled", (this._undoLog.length == 0));
        }
        for (var i = 0, item; item = this._redoUIItems[i]; i++) {
            item.attr("disabled", (this._redoLog.length == 0));
        }
    }
}

/**
 * Manages the Screens tab.
 * <strong>Note:</strong> This namespace is only available in developer mode.
 * @namespace
 */
AXES.Dev.ScreensTab = {
    initDone : false,
    init : function() {
        var self = this;
        var screensTab = this.tabContent = $("#ax-tab-screens");

        var jqNull = $("null");
        this.selectedSubfile = jqNull;
        this.selectedElement = jqNull;
        this.selectedDef = jqNull;

        TEXT.apply(window, this.tabContent[0]);

        this.btnSave = $("#btnDefSave", screensTab).attr("disabled", "disabled").click(this.onSaveClicked);
        this.btnCancel = $("#btnDefCancel", screensTab).attr("disabled", "disabled").click(this.onCancelEditClicked);
        this.btnDelete = $("#btnDefDelete", screensTab).attr("disabled", "disabled").click(this.onDeleteDefClicked);
        this.btnSuggest = $("#btnDefSuggest", screensTab).attr("disabled", "disabled").click(this.onSuggestClicked);
        this.fldName = $("#screenName", screensTab);
        this.fldVariantName = $("#screenVariantName", screensTab);
        this.fldDescription = $("#screenDescription", screensTab);
        this.txtScreenSIG = $("#screenSIG", screensTab);
        this.txtScreenId = $("#screenID", screensTab);

        this.possMatchList = $("#matches .listBody", screensTab);
        this.subfileList = $("#subfiles .listBody", screensTab)[0];
        this.fieldList = $("#fields .listBody", screensTab)[0];

        $("#ax-tab-screens #matches .listRow").live("click", function() {
            var selectedRow = this;
            var callbackToContinue = function() {
                self.selectDefinition(selectedRow);
            };
            if (self.checkIfSaveNeededAndSave(callbackToContinue) == false)
            {
                return false;
            }
            self.selectDefinition(selectedRow);
        });

        $("#ax-tab-screens #subfiles .listRow").live("click", function() {
            self.selectedElement.removeClass("ui-state-highlight");
            self.selectedElement = jqNull;
            if (self.selectedSubfile != $(this)) {
                self.selectedSubfile.removeClass("ui-state-highlight");
                self.selectedSubfile = $(this).addClass("ui-state-highlight");
            }
        });

        $("#ax-tab-screens #fields .listRow").live("click", function() {
            if (self.selectedElement != $(this)) {
                var id = this.id.split(".")[1];
                AXES.Dev.ScreensTab.selectFieldById(id);
            }
        });

        $("#ax-tab-screens #screenDefSection").keyup(function() {
            if ($("#screenName", screensTab).val() != "") {
                self.btnSave.attr("disabled", false);
                self.btnCancel.attr("disabled", false);
            } else {
                self.btnSave.attr("disabled", "disabled");
                self.btnCancel.attr("disabled", "disabled");
            }
        });

        $("#ax-tab-screens .cbxSelectElement").live("click", this.onFieldInDefinitionClicked);

        this.fldName.data("CharactersToRemove", "[^a-zA-Z0-9_-]");
        //Used by guessScreenName
        this.fldName.constrain({
            allow : {
                regex : "[a-zA-Z0-9_-]"
            }
        });
        this.fldVariantName.constrain({
            allow : {
                regex : "[a-zA-Z0-9_-]"
            }
        });
        $(".txtElementName", screensTab).constrain({
            allow : {
                regex : "[a-zA-Z0-9_-]"
            }
        });
        this.fldName.change(this.onScreenNameChanged);
        this.fldVariantName.change(this.onScreenVariantNameChanged);
        this.fldDescription.change(this.onScreenDescriptionChanged);

        AXES.Terminal.addListener("ScreenArrived", this.onScreenArrived, this);
        AXES.Terminal.addListener("ScreenWillPost", this.onScreenWillPost, this);

        this.onScreenArrived();
        this.initDone = true;

        if (!AXES.Screens._dataNeedsRepair) {
            // Early versions of TS1 saved screens with no DSPF with the key "A.undefined"
            var kh = AXES.Screens.keysHash, tmp;
            if (kh["A.undefined"]) {
                tmp = [];
                for (k in kh["A.undefined"]) {
                    tmp.push(kh["A.undefined"][k]);
                }
                AXES.Screens._dataNeedsRepair = JSON.stringify(tmp);
            }
        }
        if (AXES.Screens._dataNeedsRepair) 
            window.setTimeout(this._fixScreensData, 1000);
    },
    /**
     * Fix any known errors in old versions of the screens data.
     */
    _fixScreensData: function()
    {
        var screens = eval(AXES.Screens._dataNeedsRepair), l = screens.length, toFix = [], alreadyPushed = false, screen;
        for (var i=0; i < l; i++) 
        {
            screen = screens[i];
            alreadyPushed = false;
            if (screen.key == undefined) 
                toFix.push(screen);
            else if (screen.recordFormats) 
                for (rfmtName in screen.recordFormats) 
                {
                    var rfmt = screen.recordFormats[rfmtName];
                    if (isNaN(rfmt.rowsInheader)) 
                    {
                        rfmt.rowsInHeader = null;
                        if (!alreadyPushed) 
                        {
                            alreadyPushed = true;
                            toFix.push(screen);
                        }
                    }
                }
            if (screen.key == "A.undefined") {
                // Early versions of TS1 saved screens with no DSPF with the key "A.undefined"
                screen.key = "A.";
                if (!alreadyPushed) {
                    alreadyPushed = true;
                    toFix.push(screen);
                }
            }
        }

        var fixNext = function()
        {
            var currentDef = toFix.pop();
            if (currentDef) 
            {
                var data = {
                    action : (currentDef.key == undefined) ? "delete" : "update",
                    id : currentDef.id,
                    group : "",
                    oldTimestamp : currentDef.timestamp,
                    newTimestamp : currentDef.timestamp,
                    definitionSet : AXES.definitionSet,
                    jsdata : jQuery.toJSON(currentDef)
                };

                $.ajax({
                    url : "/agi/lua5/screens.lua",
                    type : "post",
                    async : true,
                    success : fixNext,
                    error : fixNext,
                    data : data,
                    dataType : "json"
                });
            }
            else 
                delete AXES.Screens._dataNeedsRepair;
        };
        fixNext();
    },
    onShow : function() {
        if (!this.initDone)
        {
            return;
        }
        this.resize();
    },
    resize : function() {
        var tc = this.tabContent, $flb;
        if (this.initDone && tc.is(":visible")) {
            tc.height($("#ax-developerTab").innerHeight() - $("#ax-developerTab>ul").outerHeight());
            $flb = $("#fields .listBody");
            $flb.height(tc.innerHeight() - $flb[0].offsetTop + 8);
        }
    },

    onScreenArrived : function( eventObject )
    {
        var term = ( eventObject && eventObject.terminal ) ? eventObject.terminal : AXES.Terminal;

        var curentMatchText = TEXT.Dev( "Current Winner: " );
        if ( term.matchedDefinition )
        {
            curentMatchText += term.symbolicName ? term.symbolicName : TEXT.Dev( "unnamed" );
            curentMatchText += " (" + term.matchedDefinition.id + ")";
        }
        $( "#currentMatch", this.tabContent ).text( curentMatchText );

        var selectedRow = null;
        var newRowHTML = '<span class="matchID">%1</span><span class="matchName">%2</span> <span class="matchScore">%3</span>';
        this.possMatchList.empty();
        var possibleMatches = AXES.Screens.possibleMatches;
        if ( possibleMatches && possibleMatches.length )
        {
            possibleMatches.sort(
                function( a, b )
                {
                    return b.score - a.score;
                }
            );
            for ( var i = 0, pm; pm = possibleMatches[ i ]; i++ ) 
            {
                var definition = pm.definition;

                var newRow = document.createElement( "div" );
                newRow.setAttribute( "id", "possibleMatch_" + definition.id );
                newRow.className = "listRow ui-widget-content";
                newRow.innerHTML = AXES.Utils.strargs( newRowHTML, [ definition.id, definition.name, pm.score ] );
                $( newRow ).data( "MatchInfo", pm );

                this.possMatchList.append( newRow );
                if ( pm.definition.id == term.matchedDefinition.id )
                    selectedRow = newRow;
            }
        }
        var newRow = document.createElement( "div" );
        newRow.setAttribute( "id", "possibleMatch_new" );
        newRow.className = "listRow ui-widget-content";
        newRow.innerHTML = AXES.Utils.strargs( newRowHTML, [ "", ( term.isLogin ? term.symbolicName : "&lt;" + TEXT.Dev( "new definition" ) + "&gt;" ), "" ] );
        $( newRow ).data( "MatchInfo", null );

        this.possMatchList.append( newRow );

        this.btnSave.attr( "disabled", "disabled" );
        this.btnCancel.attr( "disabled", "disabled" );
        this.btnDelete.attr( "disabled", "disabled" );
        this.btnSuggest.attr( "disabled", "disabled" );

        if ( null == selectedRow )
            selectedRow = newRow;
        $( selectedRow ).click();
    },

    onScreenWillPost : function(eventObject) {
        var holdKey, callbackToContinue = function() {
            AXES.Terminal.clearHoldOnPost(holdKey);
        };
        if (!AXES.Dev.ScreensTab.checkIfSaveNeededAndSave(callbackToContinue)) {
            var holdKey = AXES.Terminal.placeHoldOnPost();
        }
    },
    selectDefinition : function(defRow) {
        defRow = $(defRow);
        this.selectedDef.removeClass("ui-state-highlight");
        this.selectedDef = defRow.addClass("ui-state-highlight");

        var matchInfo = defRow.data("MatchInfo");
        var term = AXES.Terminal;

        if (matchInfo) {
            // Work with a copy of the definition
            this.currentDef = jQuery.extend(true, {}, matchInfo.definition);
            // In case of old version without a variant name
            if (!this.currentDef.variantName)
                this.currentDef.variantName = "";
            this.currentDefOriginal = matchInfo.definition;
        } else {
            this.currentDef = {
                id : "",
                key : AXES.Terminal.getKey("A"),
                name : "",
                description : "",
                timestamp : "",
                fields : {}
            };
            this.currentDefOriginal = jQuery.extend({}, this.currentDef);
        }

        var currDef = this.currentDef;

        (currDef.id == "") ? this.txtScreenId.parent().hide() : this.txtScreenId.parent().show();
        this.txtScreenId.text(currDef.id);
        this.fldName.val((term.isLogin) ? term.symbolicName : currDef.name);
        this.fldVariantName.val(currDef.variantName);
        this.fldDescription.val(currDef.description);

        this.txtScreenSIG.text(currDef.key).ellipsis().attr("title", currDef.key);

        $("#screenName.error", this.tabContent).removeClass("error").prev(".errorIndicator").hide();
        $("#screenVariantName.error", this.tabContent).removeClass("error").prev(".errorIndicator").hide();

        // Update the field list
        $(this.fieldList).empty();
        var newRowHTML = '<div class="listRowSubline"><span class="fieldSelect"><input type="checkbox" class="cbxSelectElement" %6 title="Select to make this field part of the definition." /></span><span class="errorIndicator">!</span> <span class="fieldName"><input type="text" class="txtElementName" maxlength="32" value="%7" /></span><span class="fieldRow">%1</span><span class="fieldColumn">%2</span><span class="fieldIsSubfile">%3</span></div><div class="listRowSubline" style="%4"><span class="fieldIndent"></span><span class="fieldText" title="%5">%5</span></div>';
        for (var i = 0, fld; fld = term.sortedFields[i]; i++) {
            // Skip fields with a byp flag.
            if (fld.byp) continue;

            var newRow = document.createElement("div");
            newRow.setAttribute("id", "ax-fldRow-def." + fld.id);
            newRow.className = "listRow ui-widget-content";

            var fldDef = currDef.fields[fld.adjustedId()];

            var args = [fld.row, fld.col,
            /* TRANSLATORS: This is a short yes indicator. */
            ((fld.recordFormat) && (fld.recordFormat.type == AXES.Terminal.RecordFormat.Types.SUBFILE)) ? TEXT.Dev("Y") : "",
            (fld.type == AXES.Terminal.Field.Types.OUTPUT) ? "" : "display:none",
            (fld.type == AXES.Terminal.Field.Types.OUTPUT) ? fld.getValue() : "",
            (fldDef && fldDef.formId) ? 'checked="checked"' : "",
            (fldDef) ? fldDef.name : ((term.isLogin) ? fld.id : "")
            ];

            newRow.innerHTML = AXES.Utils.strargs(newRowHTML, args);
            this.fieldList.appendChild(newRow);
            $(newRow).hover(AXES.Dev.onFieldRowHoverStart, AXES.Dev.onFieldRowHoverEnd)
                .find("input.txtElementName").change(this.onFieldNameChanged);
        }

        // Update the Subfiles list
        $(this.subfileList).empty();
        newRowHTML = '<span class="subfileStart">%1</span><span class="subfileEnd">%2</span><span class="subfileName">%3</span><span class="subfileLPI"><input type="text" size="2" maxlength="2" class="txtSubfileLinesPerItem" value="%4" /></span><span class="subfileLPH"><input type="text" size="2" maxlength="2" class="txtSubfileLinesInHdr" value="%5" /></span>';
        for (var i = 0, rfmt; rfmt = term.recordFormats[i]; i++) {
            if (rfmt.type == "sfl") {
                var newRow = document.createElement("div");
                newRow.setAttribute("id", "ax-sflRow-def." + i);
                newRow.className = "listRow ui-widget-content";

                var args = [rfmt.startLine, rfmt.endLine, rfmt.name,
                (rfmt.rowsPerItem) ? rfmt.rowsPerItem : "1",
                (rfmt.rowsInHeader != null) ? rfmt.rowsInHeader : ""
                ];
                
                $(newRow).data("RecordFormat", rfmt);
                newRow.innerHTML = AXES.Utils.strargs(newRowHTML, args);
                this.subfileList.appendChild(newRow);
                
                if (!currDef.recordFormats) {
                    currDef.recordFormats = {};
                }
                this.updateDefinitionSubfiles($(newRow));
                
                $(newRow).find("input").numeric();
                $(newRow).find("input").change(this.onSubfilesChanged);
            }
        }

        if (term.isLogin) 
            $("#screenDefSection :input").attr("disabled", "disabled");
        else 
        {
            $("#screenDefSection :input:not(:button)").attr("disabled", false);
            this.fldName.attr("disabled", false);
            this.fldVariantName.attr("disabled", false);
            this.fldDescription.attr("disabled", false);
            this.btnSave.attr("disabled", "disabled");
            this.btnCancel.attr("disabled", "disabled");
            this.btnDelete.attr("disabled", (currDef.id == "") ? "disabled" : false);
            this.btnSuggest.attr("disabled", (currDef.id == "") ? false : "disabled");
        }
    },
    onScreenNameChanged : function(e) {
        // DOM event handler so "this" does not refer to AXES.Dev.ScreensTab
        var st = AXES.Dev.ScreensTab;
        st.currentDef.name = this.value = jQuery.trim(this.value);
        var errMsg = "";
        if (st.customValidateScreenName) {
            errMsg = st.customValidateScreenName(this.value);
        } else {
            errMsg = st.sendFieldValidateEvent(this.value, "ScreenName")
        }
        st.showError(this, errMsg);
    },
    onScreenVariantNameChanged : function(e) {
        // DOM event handler so "this" does not refer to AXES.Dev.ScreensTab
        var st = AXES.Dev.ScreensTab;
        st.currentDef.variantName = this.value = jQuery.trim(this.value);
        var errMsg = "";
        if (st.customValidateScreenVariantName) {
            errMsg = st.customValidateScreenVariantName(this.value);
        } else {
            errMsg = st.sendFieldValidateEvent(this.value, "ScreenVariantName")
        }
        st.showError(this, errMsg);
    },
    onScreenDescriptionChanged : function(e) {
        // DOM event handler so "this" does not refer to AXES.Dev.ScreensTab
        var st = AXES.Dev.ScreensTab;
        st.currentDef.description = this.value = jQuery.trim(this.value);
        var errMsg = "";
        if (st.customValidateScreenDescription) {
            errMsg = st.customValidateScreenDescription(this.value);
        } else {
            errMsg = st.sendFieldValidateEvent(this.value, "ScreenDescription")
        }
        st.showError(this, errMsg);
    },
    onFieldNameChanged : function(e) {
        // DOM event handler so "this" does not refer to AXES.Dev.ScreensTab
        var st = AXES.Dev.ScreensTab;
        this.value = jQuery.trim(this.value);
        st.updateDefinitionField($(this).parents(".listRow:first"));
        var errMsg = "";
        if (st.customValidateScreenDescription) {
            errMsg = st.customValidateScreenDescription(this.value);
        } else {
            errMsg = st.sendFieldValidateEvent(this.value, "FieldName")
        }
        st.showError(this, errMsg);
    },
    onSubfilesChanged: function(e) {
        var st = AXES.Dev.ScreensTab;
        var rf = $(this).parents(".listRow:first").data("RecordFormat");
        if ($(this).hasClass("txtSubfileLinesPerItem")) {
            var val = parseInt($(this).val());
            if (!val)
                $(this).val( val = 1);
            rf.rowsPerItem = val;
        } else if ($(this).hasClass("txtSubfileLinesInHdr")) {
            var val = parseInt($(this).val());
            if (isNaN(val)) {
                $(this).val("");
                val = null;
            }
            rf.rowsInHeader = val;
        }
        var row = $(this).parents(".listRow:first");
        st.updateDefinitionSubfiles(row);
    },
    onFieldInDefinitionClicked : function(e) {
        var st = AXES.Dev.ScreensTab;
        var cd = st.currentDef;
        if (cd.name == "") {
            cd.name = st.guessScreenName($(".fieldText", $(this).parents(".listRow")).text());
            $("#screenName", st.tabContent).val(cd.name).change();

        }
        st.updateDefinitionField($(this).parents(".listRow:first"));
        st.btnSave.attr("disabled", false);
        st.btnCancel.attr("disabled", false);
    },
    showError : function(theField, errMsg) {
        if (errMsg == "") {
            $(theField).removeClass("ui-state-error").parent().find(".errorIndicator").hide();
        } else {
            $(theField).addClass("ui-state-error").parent().find(".errorIndicator").show().attr("title", errMsg);
        }
    },
    sendFieldValidateEvent : function(val, theFieldName) {
        var errObj = {
            fieldName : theFieldName,
            fieldValue : val,
            errorText : ""
        };
        AXES.fire("DefinitionFieldNeedsValidate", [errObj]);
        return errObj.errorText;
    },
    saveDefinition : function() {
        var currentDef = this.currentDef;
        var ae = $(document.activeElement);
        if (ae.parents("#ax-tab-screens").length > 0) {
            ae.blur();
        }

        var saveDialog = $(document.createElement("div")).attr("title", TEXT.Dev("Save Definition"));
        var saveDialogOptions = {
            bgiframe : true,
            modal : true,
            position : "top",
            buttons : {}
        };
        saveDialogOptions.buttons["OK"] = function() {
            $(this).dialog('close').dialog('destroy');
        };

        if ($(".errorIndicator:visible", this.tabContent).length > 0) {
            var msg = TEXT.Dev("This definition contains errors. Please correct and try again.");
            saveDialog.text(msg).dialog(saveDialogOptions);
            return false;
        }

        this.btnSave.attr("disabled", "disabled");
        this.btnCancel.attr("disabled", "disabled");

        // Do some cleaning up
        for (var fld in currentDef.fields) {
            if ((fld.value == "") && (!fld.formId))
                delete currentDef.fields[fld.adjustedId()];
        }

        currentDef.timestamp = (new Date()).getTime();
        // Make sure it is the new key format
        currentDef.key = AXES.Terminal.getKey("A");
        if (currentDef.id == "")
            currentDef.id = "%%id%%";

        var data = {
            action : (currentDef.id == "%%id%%") ? "create" : "update",
            id : currentDef.id,
            group : "",
            oldTimestamp : this.currentDefOriginal.timestamp,
            newTimestamp : currentDef.timestamp,
            definitionSet : AXES.definitionSet,
            jsdata : jQuery.toJSON(currentDef)
        };

        var saveOK = false;
        var options = {
            url : "/agi/lua5/screens.lua",
            type : "post",
            async : false,
            success : function(data, textStatus)
            {
                if (data.code == 0) {
                    if (data.id)
                        currentDef.id = data.id;
                    AXES.Screens.saveScreenDefinition(currentDef);
                    AXES.Screens.removeFromScreenNameIndex(currentDef.name, currentDef.variantName);
                    AXES.Screens.addToScreenNameIndex(currentDef.name, currentDef.variantName);

                    saveOK = true;
                } else {
                    saveDialog.text(TEXT.Dev("Unable to save changes: %s1",TEXT.Dev(data.message)))
                        .dialog(saveDialogOptions);
                }
            },
            error: function(XMLHttpRequest, textStatus, errorThrown)
            {
                saveDialog.text(TEXT.Dev("Unable to save changes: %s1", textStatus))
                        .dialog(saveDialogOptions);
            },
            data : data,
            dataType : "json"
        };

        $.ajax(options);
        return saveOK;
    },
    onSaveClicked : function(e) {
        var st = AXES.Dev.ScreensTab, newMatch;
        if (st.saveDefinition()) {
            newMatch = AXES.Screens.findMatch(AXES.Terminal);
            if (st.currentDefOriginal.id != newMatch.id) {
                AXES.Terminal.screenLoaded();
            } else {
                AXES.Terminal.matchedDefinition = newMatch;
                AXES.Terminal.updateWithMatch(newMatch);
                AXES.Layout.doLayout();
                st.onScreenArrived();
            }
        } else {
            st.btnSave.attr("disabled", false);
            st.btnCancel.attr("disabled", false);
        }
    },
    onCancelEditClicked : function(e) {
        AXES.Dev.ScreensTab.onScreenArrived({
            terminal : AXES.Terminal
        });
    },
    onDeleteDefClicked : function(e) {
        var st = AXES.Dev.ScreensTab;

        if (st.currentDef.id != "") {
            var delDialog = $(document.createElement("div")).attr("title", TEXT.Dev("Delete Definition"));
            var delDialogOptions = {
                bgiframe : true,
                modal : true,
                position : "top",
                buttons : {}
            };
            delDialogOptions.buttons[TEXT.Dev("OK")] = function() {
                $(this).dialog('close').dialog('destroy');
            };

            var data = {
                action : "delete",
                id : st.currentDef.id,
                definitionSet : AXES.definitionSet
            };
            var options = {
                url : "/agi/lua5/screens.lua",
                type : "post",
                async : false,
                success : function(data, textStatus) {
                    if (data.code == 0) {
                        var def = st.currentDef;
                        AXES.Screens.deleteScreenDefinition(def);
                        AXES.Screens.removeFromScreenNameIndex(def.name, def.variantName);
                        AXES.fire("ScreenDefinitionDeleted", [{
                            axesID : def.id,
                            symbolicName : def.symbolicName,
                            variantName : def.variantName
                        }]);

                        if (st.currentDefOriginal.id == def.id) {
                            //Deleted the current match so need to re-match
                            AXES.Terminal.screenLoaded();
                        } else {
                            st.onScreenArrived();
                        }
                    } else {
                        var msg = TEXT.Dev("Unable to delete the definition: %s1.", TEXT.Dev(data.message));
                        delDialog.text(msg).dialog(delDialogOptions);
                        st.btnCancel.click();
                    }
                },
                error : function(XMLHttpRequest, textStatus, errorThrown) {
                    var msg = TEXT.Dev("Unable to delete the definition: %s1.", textStatus);
                    delDialog.text(msg).dialog(delDialogOptions);
                },
                data : data,
                dataType : "json"
            };
            $.ajax(options);
        }
        return false;
    },
    updateDefinitionField : function(row) {
        var cd = this.currentDef;
        var id = row[0].id.split(".")[1];
        var fld = AXES.Terminal.getFieldById(id);
        var fldName = row.find("input.txtElementName").val();
        var formId = (row.find("input.cbxSelectElement:checked").length > 0);
        if (formId || fldName != "") {
            var fldDef = cd.fields[fld.adjustedId()];
            if (!fldDef)
                fldDef = cd.fields[fld.adjustedId()] = {};
            fldDef.name = fldName;
            fldDef.formId = formId;
            if (formId && fld.type == AXES.Terminal.Field.Types.OUTPUT)
                fldDef.text = fld.getValue();
        } else {
            delete cd.fields[fld.adjustedId()];
        }
        // TODO update subfile column name if required
        //if (fld.isIndexable() && (fld.name != "")) {
        //    var rfDef = st.currentDef.recordFormats[fld.recordFormat.name];
        //    if (rfDef && (rfDef.rowsPerItem > 0)) {
        //        if (!rfDef.colNames[fld.subfileCol]) rfDef.colNames[fld.subfileCol] = fld.name;
        //    }
        //}

    },
    updateDefinitionSubfiles: function(row) {
        var cd = this.currentDef;
        var rf = row.data("RecordFormat");
        var rowData = {
                "rowsInHeader": rf.rowsInHeader,
                "rowsPerItem": rf.rowsPerItem,
                "colNames": {}
                };                
        cd.recordFormats[rf.name] = rowData;               
    },
    guessScreenName : function(title) {
        var screenName = title.replace(/\ba\b|\ban\b|\band\b|\bthe\b|\bor\b|\bif\b]/gi, ""),
                    invalidCharacters = $("#screenName", this.contentTab).data("CharactersToRemove"),
                    re = new RegExp(invalidCharacters, "g");
        screenName = screenName.replace(re, "");
        screenName = screenName.replace(/\s/gi, "");
        if (screenName.length > 20)
            screenName = screenName.replace(/[aeiou]/g, "");
        if (screenName.length > 20)
            screenName = screenName.replace(/[AEIOU]/g, "");
        if (screenName.length > 20)
            screenName = screenName.substring(0, 20);
        return screenName;
    },
    onSuggestClicked : function(e) {
        // We are looking for the first 2 fields - usually an id and a title.  Sometimes
        // Title may be on second line so look for first two items, excluding anything on
        // the right of the screen.
        var st = AXES.Dev.ScreensTab;
        var fields = AXES.Terminal.sortedFields;
        var fieldsToUse = [];
        var secondRow = -1;
        for (var i = 0, fld; fld = fields[i]; i++) {
            if ((fld.type == AXES.Terminal.Field.Types.OUTPUT) && (fld.col <= 60)) {
                if (secondRow < 0)
                    secondRow = fld.row + 1;
                if (fld.row > secondRow)
                    break;
                //do not use field with empty value for TS1 compatibity.
                if (fld.data !=" ") fieldsToUse.push(fld);
                //we need only 1 field for TS1 compatibity
                if (fieldsToUse.length == 1)
                    break;
            }
        }
        var screenName = (fieldsToUse.length > 0) ? fieldsToUse[0].getValue() : "";

        // We needed to determine the "fields to use" because they are
        // used to determine the screen name but they only need to be part
        // of the definition if there are other possible matches.
        if (AXES.Screens.possibleMatches && (AXES.Screens.possibleMatches.length > 0)) {
            var select = "";
            for (var i = 0, fld; fld = fieldsToUse[i]; i++) {
                var listRow = $(st.fieldList).find("#ax-fldRow-def\\." + fld.id);
                listRow.find(".cbxSelectElement").attr("checked", "checked");
                st.updateDefinitionField(listRow);
            }
        }
        screenName = st.guessScreenName(screenName)
        $("#screenName", st.tabContent).val(screenName).change();
        $("#screenDescription", st.tabContent).val((fieldsToUse.length > 1) ? fieldsToUse[1].getValue() : "").change();

        AXES.Screens.indexScreenNames();
        var variantNum = 1;
        var variantStr = "";
        while (AXES.Screens.isScreenNameInIndex(screenName, variantStr)) {
            variantStr = (variantNum++).toString();
        }
        $("#screenVariantName", st.tabContent).val(variantStr).change();

        st.btnSave.attr("disabled", false);
        st.btnCancel.attr("disabled", false);
    },
    checkIfSaveNeededAndSave : function(callBackToContinue) {
        var st = this;
        if ($("#btnDefSave:enabled", this.tabContent).length > 0) {

            var dialogOptions = {
                closeOnEscape : false,
                bgiframe : true,
                modal : true,
                position : "top",
                buttons : {}
            }
            dialogOptions.buttons[TEXT.Dev("Save")] = function() {
                $(this).dialog('close').dialog('destroy');
                if (st.saveDefinition())
                    callBackToContinue();
            };
            dialogOptions.buttons[TEXT.Dev("Discard")] = function() {
                $(this).dialog('close').dialog('destroy');
                st.btnSave.attr("disabled", "disabled");
                st.btnCancel.attr("disabled", "disabled");
                st.btnDelete.attr("disabled", "disabled");
                st.btnSuggest.attr("disabled", "disabled");
                st.onCancelEditClicked();
                callBackToContinue();
            };
            dialogOptions.buttons[TEXT.Dev("Cancel")] = function() {
                $(this).dialog('close').dialog('destroy');
            };

            $(document.createElement("div")).attr("title", TEXT.Dev("Definition Changed"))
            .text(TEXT.Dev("Do you want to save your definition changes first?"))
            .dialog(dialogOptions);
            return false;
        } else {
            return true;
        }
    },
    selectFieldById : function(fieldId) {
        if (this.selectedElement != $(this)) {
            this.selectedElement.removeClass("ui-state-highlight");
            this.selectedElement = $("#ax-fldRow-def\\." + fieldId).addClass("ui-state-highlight");
        }
    }
}

/**
 * Manages the Extensions tab.
 * <strong>Note:</strong> This namespace is only available in developer mode.
 * @namespace
 */
AXES.Dev.ExtensionsTab =
{
    initDone        : false,
    AppsDefNeedSave : false,
    ScreenDefNeedSave  : false,
    FieldExtReadOnly: false,
    selectedFieldRow: null,
    oldModeVals     : {},

    init: function()
    {
        this.tabContent        = $( "#ax-tab-extensions" );
        this.fieldItemTemplate = $( "#ax-extensions-hiddenresources .fieldItem" );
        this.btnSave           = $( "#btnExSave" ).attr( "disabled", "disabled" ).click( this.onSaveClicked );
        this.btnCancel         = $( "#btnExCancel" ).attr( "disabled", "disabled" ).click( this.onCancelClicked );
        this.btnReset          = $( "#btnExReset" ).attr( "disabled", "disabled" ).click( this.onResetClicked );
        this._undo             = new AXES.Dev.Undo( this );

        TEXT.apply(window, this.tabContent[0]);

        var self = this;
        this._undo.addUndoUIItem(
            $( "#btnExUndo" ).click(
                function( e )
                {
                    self._undo.doUndo();
                }
            )
        );
        this._undo.addRedoUIItem(
            $( "#btnExRedo" ).click(
                function( e )
                {
                    self._undo.doRedo();
                }
            )
        );
        this._undo.setUIItems();

        $( "#ax-extensions-fieldListUL li" ).live( "click", this, this.onFieldSelected );
        $( "#ax-extensions-fieldDetail-exList" ).delegate( ".ax-extensions-exMenuButton", "click", this, this.onExMenuButtonClick );
        $( "#ax-extensions-fieldListMenuButton, #ax-extensions-extensionListMenuButton" )
            .button( { text: false, icons: { primary: "ui-icon-triangle-1-s" } } )
            .click(
                function( event )
                {
                    var menu = $( this ).next();
                    if ( menu.is( ":visible" ) )
                    {
                        menu.hide();
                        return false;
                    }
                    menu.menu( "blur" )
                        .show()
                        .css( { top: 0, left: 0 } )
                        .position( { my: "right top", at: "right bottom", of: this } );
                    $( document ).one(
                        "click",
                        function()
                        {
                            menu.hide();
                        }
                    );
                    return false;
                }
            );

        $( "#ax-extensions-fieldListMenuButton" )
            .next()
            .menu( { select : this.onFieldMenuItemSelect } )
            .hide();
        $( "#ax-extensions-extensionListMenuButton" )
            .next()
            .menu({ select : this.onExListMenuSelect } )
            .hide();
        $( "#ax-extensions-extensionMenu" )
            .menu( { select : this.onExMenuItemSelect } )
            .hide();

        AXES.Dev.CustomPropertyEditors.init();

        AXES.Terminal.addListener( "ScreenArrived", this.onScreenArrived, this );
        AXES.Terminal.addListener( "ScreenWillPost", this.onScreenWillPost, this );
        this.onScreenArrived();
        this.initDone = true;
    },

    _footerButtonDirtyCount : 0,
    /**
     * Update the "dirty" state of the tab. Each undoable/redoable action represents a state change of +/- 1.
     * @param {Integer} dirtyCountIncr The amount to change the dirty state. This will usually be +/- 1. A value of -10 (< -9) should reset the dirty state.
     */
    updateDirtyState : function(dirtyCountIncr) {
        this._footerButtonDirtyCount += dirtyCountIncr;
        if ((this._footerButtonDirtyCount < 0) || (dirtyCountIncr < -9))
            this._footerButtonDirtyCount = 0;
        var val = (this._footerButtonDirtyCount == 0) ? "disabled" : false;
        this.btnSave.attr("disabled", val);
        this.btnCancel.attr("disabled", val);
        this.btnReset.attr("disabled", (this.currentDefinition) ? false : "disabled");
    },
    resize : function() {
        if (this.initDone && $("#ax-tab-extensions").is(":visible")) {
            this.tabContent.height($("#ax-developerTab").innerHeight() - $("#ax-developerTab>ul").outerHeight());
            this.layout.resizeAll();
            this.resizeExPropertyColumns();
        }
    },
    resizeExPropertyColumns : function() {
        var w = $("#ax-extensions-fieldDetail-exList")[0].clientWidth;
        $(".ax-extensions-exSection").each(function() {
            var w2 = w - $(this).find(".propertyName:first").outerWidth() - $(this).find(".propertyMode:first").outerWidth();
            w2 = w2 - ($(this).find(".propertyValue").outerWidth() - $(this).find(".propertyValue").width());
            $(this).find(".propertyValue").css("max-width", (w2 - 2) + "px");
        });
    },
    onExListMenuSelect: function(event, ui)
    {
        var $t = $(this), ex = $t.data("currentEx"), fld = $t.data("currentField"), exTab = AXES.Dev.ExtensionsTab;
        if ($t.hasClass("ui-state-disabled"))
        {
            return;
        }

        switch (ui.item[0].id)
        {
            case "ax-extensions-exListMenuItemAdd":
                exTab.onAddExtensionMenuSelected();
                break;
            case "ax-extensions-exListMenuItemCustomize":
                exTab.onCustomizeFieldMenuItemSelected();
                break;
            case "ax-extensions-exListMenuItemRevert":
                exTab.onResetFieldMenuItemSelected();
                break;
        }

        $t.hide();
    },
    onAddExtensionMenuSelected: function()
    {
        var exTab = this;
        var onSelClick = function()
        {
            $( "#ax-extensions-addExDialogAddButton" ).button( this.selectedIndex == -1 ? "disable" : "enable" );
        };
        var dialogOptions =
        {
            bgiframe: true,
            modal   : true,
            position: "top",
            buttons : [
                {
                    id   : "ax-extensions-addExDialogAddButton",
                    text : TEXT.Dev( "Add" ),
                    click: function()
                    {
                        var rowId = exTab.selectedFieldRow.attr( "id" );
                        var exObj = exTab.selectedFieldRow.data( "ExtendableObject" );
                        // check for subfile
                        if (exObj.recordFormat && exObj.recordFormat.cols ) {
                            var col = exObj.recordFormat.getSubfileColumn(exObj);
                            var colKey = Object.keys(col)[0];
                            var sbColumn = col[colKey];                        
                            // add extension to all the field in subfile columns
                            for (var cfld in sbColumn.items) {
                                if ( sbColumn.items[cfld] )
                                    exTab.addExtensionToExObj( sel.value, sbColumn.items[cfld]);
                            }
                            if (AXES.Dev.screenLocked) { // do a refresh
                                AXES.Dev.setScreenLocked(false);
                                AXES.Dev.setScreenLocked(true);
                            } else {
                                AXES.Dev.doLayout();
                            }
                            exTab.selectFieldAndRow(rowId);
                        } else {
                            exTab.addExtensionToExObj( sel.value, exObj );
                        }
                        $( this ).dialog( 'close' );
                    }
                },
                {
                    text : TEXT.Dev( "Cancel" ),
                    click: function()
                    {
                        $( this ).dialog( 'close' );
                    }
                }
            ],
            close: function()
            {
                $( sel ).unbind( "change", onSelClick );
                $( this ).dialog( 'destroy' );
            }
        };
        var $fd           = $( "#ax-extensions-fieldDetails" );
        var exType        = $fd.hasClass( "application" ) ? "application" : ( $fd.hasClass( "screen" ) ? "screen" : "element" );
        var allExtensions = AXES.Extensions.getAllExtensions();
        var len           = allExtensions.length;
        var div           = $( "#ax-extensions-addExtensionDialog" );
        var sel           = div.find( "select" ).empty()[ 0 ];
        var exArray       = [];
        var ex;

        for ( var i = 0; i < len; i++ )
        {
            ex = allExtensions[ i ].prototype;
            if ( ex.type == exType )
                exArray.push( ex );
        }

        exArray.sort(
            function( a, b )
            {
                return a.display < b.display ? -1 : ( b.display < a.display ? 1 : 0 );
            }
        );

        var opt;
        for ( var i = 0; i < exArray.length; i++ )
        {
            ex = exArray[ i ];
            opt = document.createElement( "option" );
            opt.setAttribute( "value", ex.name );
            opt.appendChild( document.createTextNode( ex.display ) );
            sel.appendChild( opt );
        }

        div.attr( 
            "title",
            TEXT.Dev( "Add eXtension" )
        ).dialog( dialogOptions );
        $( "#ax-extensions-addExDialogAddButton" ).button( "disable" );
        $( sel ).bind( "change", onSelClick );
    },
    /**
     * Customize or uncustomize a field. A field is customized by setting the isAutoGenerated flag on each eXtension to false.
     * If the useFlags parameter is supplied, it should contain an array of values to apply to the isAutoGenerated flag on
     * each eXtension. This can be used to undo a customize action
     * @param {String} fieldId The ID of the field to customize
     * @param {Array} useFlags An optional array of values to use for isAutoGenerated on each eXtension
     */
    _customizeExObj : function(rowId, useFlags) {
        var row = this.getFieldRowFromRowId(rowId), exObj = row.data("ExtendableObject"), oldVals = new Array(exObj.extensions.length);
        
        if (exObj && exObj.recordFormat && exObj.recordFormat.cols) {
            var col = exObj.recordFormat.getSubfileColumn(exObj);
            var colKey = Object.keys(col)[0];
            var sbColumn = col[colKey];
            for (var cfld in sbColumn.items) {
                if ( sbColumn.items[cfld] ) {
                    for (var i = 0, ex; ex = sbColumn.items[cfld].extensions[i]; i++) {
                        oldVals[i] = ex.isAutoGenerated;
                        ex.isAutoGenerated = (useFlags) ? useFlags[i] : false;
                    }
                }
            }
        } else {
            for (var i = 0, ex; ex = exObj.extensions[i]; i++) {
                oldVals[i] = ex.isAutoGenerated;
                ex.isAutoGenerated = (useFlags) ? useFlags[i] : false;
            }
        }
        this.selectFieldAndRow(rowId);
    },
    onCustomizeFieldMenuItemSelected : function()
    {
        var fieldRow = this.selectedFieldRow;
        var rowId    = fieldRow.attr( "id" );

        // customize screens
        if ( "ax-fldRow-screen" == rowId )
        {
            var screenName = AXES.Terminal.symbolicName;
            if ( screenName == "" )
            {
                alert( TEXT.dev( "designer.warningUnidentifiedScreenEditing" ) );
                return;
            }

            var hasAXSX = false;
            var exObj = fieldRow.data( "ExtendableObject" );
            if ( exObj.extensions.length )
            {
                var extensions = exObj.extensions;
                var len = extensions.length;
                for ( var i = 0; i < len; i++ )
                    if ( extensions[ i ].name == "axsx" )
                    {
                        hasAXSX = true;
                        break;
                    }
            }
            if ( !hasAXSX )
                this.addExtensionToExObj( "axsx", exObj, screenName );
        }

        this._customizeExObj(rowId);
    },
    onResetFieldMenuItemSelected : function()
    {
        var fieldRowId = this.selectedFieldRow[0].id, exObj = this.selectedFieldRow.data("ExtendableObject"),
        a = fieldRowId.split("."), id = (a.length > 1) ? a[1] : "", exObj, undoCount = 1, prevAGValue;

        if (exObj.recordFormat && exObj.recordFormat.cols ) {
            var col = exObj.recordFormat.getSubfileColumn(exObj);
            var colKey = Object.keys(col)[0];
            var sbDef = exObj.recordFormat.name.toLowerCase() + "_" + colKey;
            var sfCols = this.currentDefinition.sfCols;
            if ( sfCols && (sfCols.hasOwnProperty(sbDef)) ) {
                delete sfCols[sbDef];
                var sbColumn = col[Object.keys(col)[0]];
                for (var cfld in sbColumn.items) {
                    if ( sbColumn.items[cfld] ) {                    
                        var fld = sbColumn.items[cfld];
                        this.setExtensionProperties(fld, 0);
                        for (var i = 1, ex; ex = fld.extensions[i]; i++) {
                            this.removeExtensionFromExObj(fld, i);
                            i--
                            undoCount++;
                        }
                    }
                }
            }
            else {
                this.removeFieldDefinition(exObj);
                this.setExtensionProperties(exObj, 0);
                prevAGValue = exObj.extensions[0].isAutoGenerated;
                exObj.extensions[0].isAutoGenerated = true;
                for (var i = 1, ex; ex = exObj.extensions[i]; i++) {
                    this.removeExtensionFromExObj(exObj, i);
                    i--
                    undoCount++;
                }
            }
        } else {
            this.removeFieldDefinition(exObj);
            this.setExtensionProperties(exObj, 0);
            prevAGValue = exObj.extensions[0].isAutoGenerated;
            exObj.extensions[0].isAutoGenerated = true;
            for (var i = 1, ex; ex = exObj.extensions[i]; i++) {
                this.removeExtensionFromExObj(exObj, i);
                i--
                undoCount++;
            }
        }
        if (AXES.Dev.screenLocked) { // do a refresh
            AXES.Dev.setScreenLocked(false);
            AXES.Dev.setScreenLocked(true);
        } else {
            AXES.Dev.doLayout();
        }

        this.selectFieldAndRow( fieldRowId );
        this.ScreenDefNeedSave = true;
        
        this._undo.addUndo({
                undo: function()
                {
                    for ( var i = 0; i < undoCount; i++ )
                        this._undo.doUndo();
                    AXES.Dev.doLayout();
                    this.selectFieldAndRow( fieldRowId );
                },
                redo: function()
                {
                    for ( var i = 0; i < undoCount; i++ )
                        this._undo.doRedo();
                    AXES.Dev.doLayout();
                    this.selectFieldAndRow( fieldRowId );
                },
                undoArgs        : [],
                redoArgs        : [],
                updateDirtyState: true,
                context         : this
        });
    },
    onExMenuButtonClick: function(event)
    {
        var exTab = AXES.Dev.ExtensionsTab
        var self = event.data, $t = $(this), tr = $t.closest("tr"), d = tr.data(), ex=d.extension,
            menu = $("#ax-extensions-extensionMenu"),
            isBaseEx = ("axappx,axsx,axex,axuex,axtrm,axAppStyles".indexOf(ex.name) >= 0),
            propHash = exTab.buildPropertyHashForSave( ex.properties ),
            revertNotAllowed = (propHash.length > 0) && !ex.isAutoGenerated && !AXES.Terminal.isLogin && !exTab.FieldExtReadOnly,
            deleteNotAllowed = (!isBaseEx && !AXES.Terminal.isLogin && !exTab.FieldExtReadOnly);
           
        if (($t).is(":disabled"))
        {
            return false;
        }
        if (menu.is(":visible") && menu.data("currentEx") === ex) 
        {
            menu.removeData("currentEx").hide();
            return false;
        }

        $("#ax-extensions-exMenuItemDelete")[( deleteNotAllowed ) ? "removeClass" : "addClass"]("ui-state-disabled");
        $("#ax-extensions-exMenuItemRevert")[( revertNotAllowed ) ? "removeClass" : "addClass"]("ui-state-disabled");

        menu.data({
            "currentEx" : ex,
            "currentField" : d.field,
            "currentTbody" : tr.parent()
        }).menu("blur").show().css({
            top : 0,
            left : 0
        }).position({
            my : "right top",
            at : "right bottom",
            of : this
        });
        $(document).one("click", function()
        {
            menu.removeData("currentEx").hide();
        });
        event.stopPropagation();
        return false;
    },
    onExMenuItemSelect: function( event, ui )
    {
        var $t = $(this), ex = $t.data("currentEx"), exObj = $t.data("currentField"), 
            exIndex = jQuery.inArray(ex, exObj.extensions),
            exTab = AXES.Dev.ExtensionsTab;
            
        if ($t.hasClass("ui-state-disabled"))
        {
            return;
        }

        switch (ui.item[0].id) {
            case "ax-extensions-exMenuItemHelp":
                AXES.Dev.openHelpDocument("Extension_" + (ex.helpID || ex.name), "pdf");
                break;
            case "ax-extensions-exMenuItemDelete":
                // check for subfile
                if (exObj.recordFormat && exObj.recordFormat.cols ) {
                    var col = exObj.recordFormat.getSubfileColumn(exObj);
                    var colKey = Object.keys(col)[0];
                    var sbColumn = col[colKey];                        
                    // remove extension to all the field in subfile columns
                    for (var cfld in sbColumn.items) {
                        if ( sbColumn.items[cfld] )
                            exTab.removeExtensionFromExObj(sbColumn.items[cfld], exIndex);
                    }
                } else {
                    exTab.removeExtensionFromExObj(exObj, exIndex);
                }                                                                
                break;
            case "ax-extensions-exMenuItemRevert":
                exTab.setExtensionProperties(exObj, exIndex);
                break;
        }
        $t.removeData("currentEx").hide();
        AXES.Dev.ExtensionsTab.selectFieldById(exObj.id);
    },
    /**
     * Sets an extension properties to the values in the supplied array (or default)
     * @param {Object} exObj The extendable object the eXtension belongs to
     * @param {Integer} exIndex the position of the eXtension in the extensions array
     * @param [{Object}] propValues A hash (indexed by property name) of property values.
     */
    setExtensionProperties : function(exObj, exIndex, propValues) {
        // exIndex is used instead of the actual ex object because eXtensions may get recreated which would break the undo
        var newVal, oldVals = {}, property, ex = exObj.extensions[exIndex], fieldRowId = this.selectedFieldRow.attr("id");

        for (propName in ex.properties) {
            property = ex.properties[propName];
            oldVals[propName] = property.getRawValue();
            newVal = (propValues && propValues[propName]) ? propValues[propName] : {};
            this._updatePropertyValue(exObj, exIndex, propName, newVal, true);
        }
        this.refreshExtension(ex);

        if ("application" == ex.type || "axappx" == ex.name) this.AppsDefNeedSave = true;            
        else this.ScreenDefNeedSave = true;  
   
        this._undo.addUndo(
        {
            undo : this.setExtensionProperties,
            undoArgs : [exObj, exIndex, oldVals],
            redo : this.setExtensionProperties,
            redoArgs : [exObj, exIndex, propValues],
            updateDirtyState : true,
            context : this
        });
    },
    onShow : function() {
        if (!this.initDone)
        {
            return;
        }

        if (this.layout) {
            this.resize();
        } else {
            this.layout = this.tabContent.children("div").layout({
                closable : false,
                slidable : false,
                north : {
                    closable : false,
                    minSize : 50,
                    size : 250,
                    contentSelector : "#ax-extensions-fieldListUL"
                },
                south : {
                    closable : false,
                    resizable : false
                },
                center : {
                    contentSelector : "#ax-extensions-fieldDetail-exList"
                }
            });
        }

        //Update any names that may have changed.
        var aNamehasChanged = false;
        this.tabContent.find("#ax-extensions-fieldListUL .listRow").each(function() {
            var t = $(this);
            var id = this.id.split(".")[1];
            if (id) 
            {
                var fld = AXES.Dev.getFieldById(id);
                var nameDiv = t.find(".fieldName");
                var name = fld.symbolicName;
                if (!name) 
                {
                    name = "Unnamed field at " + fld.row + ", " + fld.col + ".";
                }
                if (name != nameDiv.text())
                {
                    aNamehasChanged = true;
                    nameDiv.text(name);
                }
            }
        });
        if (aNamehasChanged) {
            var cbo = $("#ax-extensions-showfieldtypecombo");
            if (cbo.val() == "named")
                cbo.change();
        }
    },
    redrawExTab: function()
    {
        var newRow = document.createElement( "li" );
        newRow.setAttribute( "id", "ax-fldRow-application" );
        newRow.className = "fieldItem listRow ui-widget-content";
        $( newRow )
            .data( "ExtendableObject", AXES.Application )
            .data( "Definition", AXES.Application.definition );

        newRow.appendChild( document.createTextNode( TEXT.Dev( "Application" ) ) );
        var theList = this.tabContent.find( "#ax-extensions-fieldListUL" ).empty()[ 0 ];
        theList.appendChild( newRow );

        newRow = document.createElement( "li" );
        newRow.setAttribute( "id", "ax-fldRow-screen" );
        newRow.className = "fieldItem listRow ui-widget-content";
        $( newRow )
            .data( "ExtendableObject", AXES.Layout )
            .data( "Definition", this.currentDefinition );
        newRow.appendChild( document.createTextNode( TEXT.Dev( "Screen" ) ) );
        theList.appendChild( newRow );

        var hasPopupWindow = 1 < AXES.Terminal.windows.length;
        var fields = AXES.Terminal.sortedFields;
        for ( var i = 0, fld; fld = fields[ i ]; i++ )
        {
            if ( fld.byp == "1" || ( hasPopupWindow && !fld.winNum ) )
                continue;

            newRow = document.createElement( "li" );
            newRow.setAttribute( "id", "ax-fldRow-ex." + fld.id );
            newRow.className = "fieldItem listRow ui-widget-content";
            $( newRow )
                .data( "ExtendableObject", fld )
                .data( "Definition", this.getFieldDefinition( fld ) );

            newRow.appendChild(
                document.createTextNode( fld.symbolicName || TEXT.Dev( "Unnamed field at %s1, %s2.", fld.row, fld.col ) )
            );
            theList.appendChild( newRow );
        }

        fields = AXES.Layout.userFields;
        for ( var i = 0, fld; fld = fields[ i ]; i++ )
            this.addUserFieldToList( theList, fld );

        $( "#ax-extensions-fieldListUL li" ).hover( AXES.Dev.onFieldRowHoverStart, AXES.Dev.onFieldRowHoverEnd );

        $( "#ax-extensions-fieldDetail-name" ).text( "" );
        $( "#ax-extensions-fieldDetail-type" ).text( "" );
        $( "#ax-extensions-fieldDetail-row" ).text( "" );
        $( "#ax-extensions-fieldDetail-col" ).text( "" );
        $( "#ax-extensions-fieldDetail-exList" ).empty();
        $( "#ax-extensions-showfieldtypecombo" ).change();
    },

    onScreenArrived: function( eventObject )
    {
        this.currentDefinition = AXES.Extensions.getCurrentScreenDefinition();
        this.redrawExTab();
        this.oldModeVals = {};
        this.updateDirtyState( -10 );
        this._undo.reset();
        this.selectFieldRow( $( "#ax-fldRow-screen" ) );
    },

    addUserFieldToList : function(list, fld) {
        var newRow = document.createElement("li"), name = fld.symbolicName;
        newRow.setAttribute("id", "ax-fldRow-ex." + fld.id);
        newRow.className = "fieldItem listRow ui-widget-content";
        $(newRow).data("ExtendableObject", fld).data("Definition", this.getFieldDefinition(fld));

        if (!name)
            name = TEXT.Dev("Unnamed user field %s1.", fld.id);
        newRow.appendChild(document.createTextNode(name));

        list.appendChild(newRow);
    },
    onScreenWillPost : function(eventObject) {
        var holdKey, callbackToContinue = function() {
            AXES.Terminal.clearHoldOnPost(holdKey);
        };
        if (!AXES.Dev.ExtensionsTab.checkIfSaveNeededAndSave(callbackToContinue)) {
            var holdKey = AXES.Terminal.placeHoldOnPost();
        }
    },
    getFieldDefinition: function(fld)
    {
        var fldDef = null, def = this.currentDefinition, adjustedId = fld.adjustedId(),
            otherId = AXES.Terminal.swapInputOutputFieldId(adjustedId), possibleOther = null;
        if (def && def.elements) 
        {
            for (var i = 0; fldDef = def.elements[i]; i++) 
            {
                if (fldDef.fieldId == adjustedId)
                    break;
                if (fldDef.fieldId == otherId)
                    possibleOther = fldDef;
            }

            if (!fldDef && possibleOther) 
            {
                if (possibleOther.basePropExtension && possibleOther.basePropExtension["applyToInputOutput"] && possibleOther.basePropExtension["applyToInputOutput"].v) 
                {
                    fldDef = possibleOther;
                }
            }
        }
        return fldDef || null;
    },
    removeFieldDefinition: function(fld)
    {
        var def = this.currentDefinition;
        if ( !def )
        {
            return;
        }        
        for (var i = 0; fldDef = def.elements[i]; i++) 
        {
          if (fldDef.fieldId == fld.id) {
              def.elements.splice(i, 1);
              break;
          }
        }
    },
    onFieldSelected: function( event )
    {
        event.data.selectFieldAndRow( this.id );
    },

    getFieldRowFromRowId : function(rowId) {
        var selector = "#" + rowId.replace(".", "\\.");
        return $(selector);
    },
    selectFieldById : function(fieldId) {
        if (fieldId) {
            var row = $("#ax-fldRow-ex\\." + fieldId);
            this.selectFieldRow(row.length ? row : null);
        }
    },

    selectFieldAndRow : function( rowId )
    {
        var tmp = rowId.split( "." );
        if ( tmp.length > 1 )
            if (AXES.Dev.screenLocked)
                AXES.Dev.selectField( tmp[ 1 ] );
            else
                this.selectFieldById( tmp[ 1 ] );
        else
        {
            this.selectFieldRow( this.getFieldRowFromRowId( rowId ) );
            //remove all the highlight
            var attachedTo = $("#axDesignModeHelper").data( "attachedTo" );
            if ( !attachedTo || !attachedTo.length ) {
                return;
            }
            for (var i = 0, id; id = attachedTo[i]; i++) {
                AXES.Dev.removeFldHighlight(id);
            }
        }
    },

    selectFieldRow : function( row )
    {
        var rowId = "";
        var extendableObject = null;
        var elDef = null;

        if ( row && row.length > 0)
        {
            rowId = row[ 0 ].id;
            extendableObject = row.data( "ExtendableObject" );
            elDef = row.data( "Definition" );
        }
        $( "#ax-extensions-fieldDetails" )
            .data( "ExtendableObject", extendableObject )
            .data( "Definition", elDef )
            .removeClass( "application" )
            .removeClass( "screen" ).removeClass( "field" );

        switch ( rowId )
        {
        case "":
            break;
        case "ax-fldRow-application":
            $( "#ax-extensions-fieldDetails" ).addClass( "application" );
            break;
        case "ax-fldRow-screen":
            $( "#ax-extensions-fieldDetails" ).addClass( "screen" );
            break;
        default:
            $( "#ax-extensions-fieldDetails" ).addClass( "field" );
        }

        if ( extendableObject )
        {
            switch ( rowId )
            {
            case "ax-fldRow-application":
                $( "#ax-extensions-fieldDetail-name" ).text( AXES.definitionSet || TEXT.Dev( "[Unnamed]" ) );
                break;
            case "ax-fldRow-screen":
                var name = AXES.Terminal.symbolicName;
                if ( AXES.Terminal.variantName )
                    name += " (" + AXES.Terminal.variantName + ")";
                $( "#ax-extensions-fieldDetail-name" ).text( name || TEXT.Dev( "[Unnamed]" ) );
                break;
            default:
                $( "#ax-extensions-fieldDetail-name" ).text( row.text() );
                $( "#ax-extensions-fieldDetail-type" ).text( extendableObject.type );
                $( "#ax-extensions-fieldDetail-row" ).text( extendableObject.row ? extendableObject.row : "" );
                $( "#ax-extensions-fieldDetail-col" ).text( extendableObject.column ? extendableObject.column : "" );
                $( "#ax-extensions-fieldDetail-value" ).text( extendableObject.getValue() );
            }

            $( "#ax-extensions-fieldDetail-exList" ).empty().text( "Loading..." );

            window.setTimeout(
                function()
                {
                    AXES.Dev.ExtensionsTab.renderFieldExtensions( extendableObject, elDef );
                },
                1
            );

            if ( this.selectedFieldRow )
                this.selectedFieldRow.removeClass( "ui-state-highlight" );
            this.selectedFieldRow = row.addClass( "ui-state-highlight" );
        }
        else
        {
            $( "#ax-extensions-fieldDetail-name" ).text( "" );
            $( "#ax-extensions-fieldDetail-type" ).text( "" );
            $( "#ax-extensions-fieldDetail-row" ).text( "" );
            $( "#ax-extensions-fieldDetail-col" ).text( "" );
            $( "#ax-extensions-fieldDetail-exList" ).empty();
        }
        this.updateExListMenu();
        if ( this.layout )
            this.layout.sizeContent( "center" );
    },

    updateExListMenu: function() 
    {
        var exObj = $( "#ax-extensions-fieldDetails" ).data( "ExtendableObject" );
        if ( !exObj )
        {
            return;
        }

        var isCustomized = false;
        var revertNotAllowed = false;
        var deleteAllowed = false;
        var addItemNotAllowed = false;
        if ( exObj instanceof AXES.Layout.UserField )
        {
            isCustomized = true;
            deleteAllowed = true;
            revertNotAllowed = true;
        }
        else if ( exObj.extensions )
            for ( var i = 0, ex; ex = exObj.extensions[ i ]; i++ )
            {
                if ( ex.isAutoGenerated == false )
                {
                    isCustomized = true;
                    break;
                }
            }

        var custMenuText = "";
        if ( exObj == AXES.Application ) { 
            custMenuText = TEXT.dev( "ExtensionsTab.CustomizeApplication" );
            revertNotAllowed = true;
        }
        else if ( exObj == AXES.Layout )
            custMenuText = TEXT.dev( "ExtensionsTab.CustomizeScreen" );
        else
            custMenuText = TEXT.dev( "ExtensionsTab.CustomizeField" );
        $( "#ax-extensions-exListMenuItemCustomize a" ).text( custMenuText );
        
        // if multi-selected field disable all field menu item options
        var data = $( "#axDesignModeHelper" ).data( "attachedTo" );
        if ( data && 1 < data.length ) {
            isCustomized = true;
            revertNotAllowed = true;
            addItemNotAllowed = true;
        }
        
        $( "#ax-extensions-exListMenuItemAdd" )[ ( !AXES.Terminal.isLogin && isCustomized && !addItemNotAllowed) ? "removeClass" : "addClass" ]( "ui-state-disabled" );
        $( "#ax-extensions-exListMenuItemCustomize" )[ ( !AXES.Terminal.isLogin && !isCustomized ) ? "removeClass" : "addClass" ]( "ui-state-disabled" );
        $( "#ax-extensions-exListMenuItemRevert" )[ ( (!AXES.Terminal.isLogin && isCustomized ) && !revertNotAllowed ) ? "removeClass" : "addClass" ]( "ui-state-disabled" );
        $( "#ax-extensions-FieldListMenuItemDeleteField" )[ ((!AXES.Terminal.isLogin && isCustomized) && deleteAllowed )? "removeClass" : "addClass" ]( "ui-state-disabled" );
        $( "#ax-extensions-FieldListMenuItemAddField" )[ ( !AXES.Terminal.isLogin ) ? "removeClass" : "addClass" ]( "ui-state-disabled" );
    },
    onPropertyModeClicked: function(event) 
    {
        var et = AXES.Dev.ExtensionsTab, $propRow = $(event.target).closest("tr"), propRowId = $propRow[0].id,
        newVal = et.oldModeVals[propRowId] || ((this.value == "static") ? {v:""} : {v:"", s:true} );

        et.updatePropertyValue($propRow.data("property"), newVal, true);
    },

    renderFieldExtensions: function( fld, elDef )
    {
        var table = document.createElement( "table" );
        var extensions = fld.extensions;
        var readOnly = true;
        for ( var x = 0, ex; ex = extensions[ x ]; x++ )
            if (ex.isAutoGenerated == false) {
              readOnly = ex.isAutoGenerated;
              break;
            }
        for ( var i = 0, ex; ex = extensions[ i ]; i++ )            
            table.appendChild( this.renderFieldExtension( ex, fld, i, readOnly) );
        var $a = $( "#ax-extensions-fieldDetail-exList" ).empty();
        var a = $a[ 0 ];
        a.appendChild( table );
        this.resizeExPropertyColumns();
    },

    renderFieldExtension : function( ex, fld, exIndex, autoGen )
    {
        var tbody = document.createElement( "tbody" );
        tbody.className = "ax-extensions-exSection";
        tbody.id = [ ( ex._field ? ex._field.id : "" ), "ex", exIndex ].join( "_" );

        var tr = document.createElement( "tr" );
        $( tr ).data( { "extension": ex, "field": fld } );
        var th = document.createElement( "th" );
        th.setAttribute( "colspan", "3" );
        th.className = "ui-widget-header";
        th.appendChild( document.createTextNode( ex.display + ( autoGen ? TEXT.dev( " (automatically generated)" ) : "" ) ) );

        var b = document.createElement( "button" );
        b.className = "ax-extensions-exMenuButton";
        $( b ).text( "Options" ).button( { text: false, icons: { primary: "ui-icon-triangle-1-s" } } );
        th.appendChild( b );

        tr.appendChild( th );
        tbody.appendChild( tr );

        var exProp = ex.properties;
        for ( var propName in exProp )
            tbody.appendChild( this._renderExProperty( ex, exIndex, exProp[ propName ], autoGen ) );

        return tbody;
    },

    _renderExProperty: function( ex, exIndex, property, setDisabled )
    {
        if (AXES.Terminal.isLogin) setDisabled = AXES.Terminal.isLogin;
        if (!setDisabled && this.FieldExtReadOnly) setDisabled = this.FieldExtReadOnly;
        var tr = document.createElement( "tr" );
        var rowId = [ "exTab_fld", ( ex._field ? ex._field.id : "" ), "ex", exIndex, property._name ].join( "_" );
        tr.id = rowId;
        tr.title = property.description || "";

        var td = document.createElement( "td" );
        td.className = "propertyName";
        td.appendChild( document.createTextNode( property._name ) );
        tr.appendChild( td );
        td = document.createElement( "td" );
        td.className = "propertyValue";
        var propEditor = property.renderEditor( setDisabled );
        if ( propEditor )
            td.appendChild( propEditor );
        tr.appendChild( td );

        var radioName = rowId + "_modeSwitch";
        var $b1 = $( "<input>" ).attr(
            {
                type    : "radio",
                id      : radioName + "1",
                name    : radioName,
                value   : "static",
                tabindex: "-1"
            }
        );
        var $b2 = $( "<input>" ).attr(
            {
                type    : "radio",
                id      : radioName + "2",
                name    : radioName,
                value   : "dynamic",
                tabindex: "-1"
            }
        );

        if ( property.getRawValue().s )
        {
            $b2.attr( "checked", "checked" );
            tr.className = "ax-extension-propertyDynamicMode";
        }
        else 
        {
            $b1.attr( "checked", "checked" );
            tr.className = "ax-extension-propertyStaticMode";
        }
        if ( setDisabled || property.noDynamic || property.noStatic || "Event" == property.type )
        {
            $b1.attr( "disabled", "disabled" );
            $b2.attr( "disabled", "disabled" );
        }
        var $s = $( "<span>" ).append( $b1 )
            .append( $( "<label>" ).attr( "for", radioName + "1" ).text( "value" ))
            .append( $b2 )
            .append( $( "<label>" ).attr( "for", radioName + "2" ).text( "script" ) );

        $b1.button( { text: false, icons: { primary: "ui-icon-pencil" } } ).click( this.onPropertyModeClicked );
        $b2.button( { text: false, icons: { primary: "ui-icon-gear" } } ).click( this.onPropertyModeClicked );
        $s.buttonset();

        td = document.createElement( "td" );
        td.className = "propertyMode";
        td.appendChild( $s[ 0 ] );
        tr.appendChild( td );

        $( tr ).data( { "extension": ex, "property": property } );
        property._displayRowId = rowId;

        return tr;
    },

    refreshExtension : function( ex )
    {
        switch ( ex.type )
        {
        case "application":
            break;
        case "screen":
            ex.render( AXES.Layout._terminalContent.get( 0 ) );
            break;
        case "element":
            ex.render( ex._field, ex._fieldDiv );
            AXES.Dev.applySelectedState();
            break;
        case "internal":
            if ( "axsx" == ex.name )
            {
                AXES.Dev.doLayout();
                this.selectFieldRow( $( "#ax-fldRow-screen" ) );
            }
            else
            {
                ex.render( ex._field, ex._fieldDiv );
                AXES.Dev.applySelectedState();
            }
        }
    },

    _updatePropertyValue: function( exObj, exIndex, propName, newVal, refreshEditorDisplay )
    {
        // The indirect references to the property object are required to support undo/redo (because
        // the extension object may get recreated during some screen redraw operations).
        var ex       = exObj.extensions[ exIndex ];
        var property = ex.getPropertyByName( propName );
        var oldVal   = property.getRawValue();
        property.setValue( newVal );
        if ( refreshEditorDisplay )
        {
            if ( oldVal.s != newVal.s )
            {
                // mode has changed
                var $propRow = $( "#" + property._displayRowId )
                var newModeIsDynamic = ( newVal.s == true );
                if ( $propRow.length )
                {
                    $propRow.toggleClass( "ax-extension-propertyStaticMode", !newModeIsDynamic ).toggleClass( "ax-extension-propertyDynamicMode", newModeIsDynamic );
                    $propRow.find( ".propertyMode input:radio" ).val( [ ( newModeIsDynamic ? "dynamic" : "static") ] );
                    $propRow.find( ".propertyMode .ui-buttonset" ).buttonset( "refresh" );
                    $propRow.children( "td.propertyValue" ).empty().append( property.renderEditor( false ) );
                    this.oldModeVals[ property._displayRowId ] = oldVal;
                }
            }
            else
                property._refreshEditorDisplay();
        }
        this.refreshExtension( ex );
    },

    updatePropertyValue: function( property, newVal, refreshEditorDisplay )
    {
        var ex = property._owner;
        if ( !ex )
        {
            // If it has no _owner then it is a sub-property of another property. No need to worry about undo
            property.setValue( newVal );
            return;
        }

        var exObj = null;
        if ( "application" == ex.type || "axappx" == ex.name )
        { 
            exObj = AXES.Application;
            this.AppsDefNeedSave = true;
        }
        else
        {
            exObj = ( "screen" == ex.type || "axsx" == ex.name ) ? AXES.Layout : ex._field;
            this.ScreenDefNeedSave = true;
        }

        var exIndex  = jQuery.inArray( ex, exObj.extensions );
        var propName = property._name;
        this._undo.addUndo(
            {
                undo            : this._updatePropertyValue,
                undoArgs        : [ exObj, exIndex, propName, property.getRawValue(), true ],
                redo            : this._updatePropertyValue,
                redoArgs        : [ exObj, exIndex, propName, newVal, true ],
                updateDirtyState: true,
                context         : this
            }
        );
        this._updatePropertyValue( exObj, exIndex, propName, newVal, refreshEditorDisplay || false );
    },

    updateFieldList : function(which) {
        which = which || "all";
        $("#ax-extensions-fieldListUL .listRow").each(function() {
            var id = this.id.split(".")[1];
            if (id) {
                var fld = AXES.Dev.getFieldById(id);
            if ((which == "all")
            || (which == "named" && fld.symbolicName)
            || (which == "input" && fld.type == AXES.Terminal.Field.Types.INPUT)
            || (which == "output" && fld.type == AXES.Terminal.Field.Types.OUTPUT)
            ) {
                    $(this).show();
                } else {
                    $(this).hide();
                }
            }
        });
    },
    onFieldMenuItemSelect: function(event, ui) 
    {
        switch (ui.item[0].id)
        {
            case "ax-extensions-FieldListMenuItemAddField":
                AXES.Dev.ExtensionsTab.addUserFieldToScreen();
                break;
            case "ax-extensions-FieldListMenuItemDeleteField":
                AXES.Dev.ExtensionsTab.removeUserFieldToScreen();
                break;
            case "ax-extensions-FieldListMenuItemShowAll":
                AXES.Dev.ExtensionsTab.updateFieldList("all");
                break;
            case "ax-extensions-FieldListMenuItemShowInput":
                AXES.Dev.ExtensionsTab.updateFieldList("input");
                break;
            case "ax-extensions-FieldListMenuItemShowOutput":
                AXES.Dev.ExtensionsTab.updateFieldList("output");
                break;
            case "ax-extensions-FieldListMenuItemShowNamed":
                AXES.Dev.ExtensionsTab.updateFieldList("named");
                break;
        }
        $(this).hide();
    },
    addUserFieldToScreen : function() {
        var prevFieldId = AXES.Dev.getSelectedFieldId(), fld = this._addUserField();
        AXES.Dev.selectField(fld.id);
        this.ScreenDefNeedSave = true;
        
        this._undo.addUndo(
        {
            undo: function()
            {
                this._removeUserField(fld);
                AXES.Dev.selectField(prevFieldId);
            },
            undoArgs : [],
            redo : function() {
                fld = this._addUserField();
                AXES.Dev.selectField(fld.id);
            },
            redoArgs : [],
            updateDirtyState : true,
            context : this
        });
    },
    _addUserField : function() {
        var windows = AXES.Terminal.windows, win = windows[windows.length - 1], fld, newEl, i, newId = 1;
        for ( i = 0; fld = AXES.Layout.userFields[i]; i++) {
            newId = Math.max(newId, parseInt(fld.id) + 1);
        }

        fld = new AXES.Layout.UserField(newId.toString());
        AXES.Layout.userFields.push(fld);
        fld.extensions.push(AXES.Extensions.createNewExtension("axuex", {
            "style" : {
                "v" : {
                    "left" : "20px",
                    "top" : "20px",
                    "width" : "100px",
                    "height" : "30px"
                }
            }
        }, null, false));

        newEl = document.createElement("div");
        newEl.setAttribute("id", fld.id);
        newEl.className = "ax-terminal-field";
        fld.terminalDiv = newEl;
        AXES.Layout.drawField(fld);

        $("#" + ((windows.length > 1) ? win.id : "terminalContent")).append(newEl);
        fld.addedToDOM();

        this.addUserFieldToList(this.tabContent.find("#ax-extensions-fieldListUL")[0], fld);
        return fld;
    },
    removeUserFieldToScreen: function()
    {
        var fieldId = AXES.Dev.getSelectedFieldId();
        var fld = AXES.Dev.getFieldById( fieldId );
        if ( !( fld instanceof AXES.Layout.UserField ) )
        {
            alert( TEXT.dev( "unable to delete not a user field.." ) );
            return;
        }

        this.removeFieldDefinition(fld);

        this._removeUserField( fld );
        this.ScreenDefNeedSave = true;
        this._undo.addUndo(
            {
                context         : this,
                redoArgs        : [],
                undoArgs        : [],
                updateDirtyState: true,

                undo: function()
                {
                    //fld = this._addUserField();
                    //AXES.Dev.selectField(fld.id);
                },
                redo: function()
                {
                    //this._removeUserField(fld);
                    //AXES.Dev.selectField(fieldId);
                }
            }
        );
    },
    _removeUserField : function(fld) {
        if (fld.terminalDiv) {
            for (var i = 0, ex; ex = fld.extensions[i]; i++) {
                if (ex.destroy)
                    ex.destroy(fld, fld.terminalDiv);
                delete ex._field;
            }
            $(fld.terminalDiv).remove();
        }
        for (var i = 0, f; f = AXES.Layout.userFields[i]; i++) {
            if (f === fld) {
                AXES.Layout.userFields.splice(i, 1);
                break;
            }
        }
        $("#ax-fldRow-ex\\." + fld.id).remove();
    },

    buildPropertyHashForSave: function( properties )
    {
        var propHash =
        {
            length: 0,
            values: 
	    {
            	clsid : ""
            }
        };
        for ( var p in properties )
        {
            var property = properties[ p ];
            var val = property.getJSONObject( true );
            if ( val )
            {
                propHash.values[ p ] = val;
                propHash.length++;
            }
        }
        return propHash;
    },

    onSaveClicked: function()
    {
        var exTab = AXES.Dev.ExtensionsTab 
        if (exTab.ScreenDefNeedSave)
            exTab.saveScreen();
        if (exTab.AppsDefNeedSave)
            exTab.saveAppsDefinition();
    },

    buildAppsDefinitionForSave: function()
    {
        var applicationDef =
        {
            clsid            : "axapp",
            extensions       : [],
            basePropExtension: null
        };
        var exTab            = AXES.Dev.ExtensionsTab
        var extendableObject = AXES.Application
        var basePropExt      = extendableObject.basePropExtension

        // extensions (terminal settings and styles)
        for ( var j = 0, ex; ex = extendableObject.extensions[ j ]; j++ )
        {
            if ( 0 == j )
            {
                // basePropExtensions
                propHash = exTab.buildPropertyHashForSave( ex.properties );
                propHash.values.clsid = ex.name;
                applicationDef.basePropExtension = propHash.values;
            }
            else
            {
                propHash = exTab.buildPropertyHashForSave( ex.properties );
                propHash.values.clsid = ex.name;
                applicationDef.extensions.push( propHash.values );
            }
        }

        return applicationDef;
    },    
    cleanElementsDuplicity: function(elements)
    {
        var arr = {};
        for (var k in elements) {
            if (elements.hasOwnProperty(k)) {
                arr[elements[k].fieldId] = elements[k];
            }
        }
        
        var newElements = new Array();
        for ( var key in arr )
            newElements.push(arr[key]);
        
        return newElements;
    },
    buildScreenDefinitionForSave: function()
    {
        var screenDef =
        {
            clsid            : "axs",
            version          : 1,
            extensions       : [],
            elements         : this.currentDefinition ? this.cleanElementsDuplicity(this.currentDefinition.elements) : [],
            basePropExtension:
            {
                clsid: "axsx",
                name : { v: AXES.Terminal.symbolicName }
            }
        };
        var exTab = AXES.Dev.ExtensionsTab;
        var fldList = exTab.tabContent.find( "#ax-extensions-fieldListUL li" );

        fldList.each(
            function()
            {
                var addToNode        = null;
                var addToElements    = null;
                var extendableObject = $( this ).data( "ExtendableObject" );

                switch ( this.id )
                {
                case "ax-fldRow-application":
                    return;
                case "ax-fldRow-screen":
                    addToNode = screenDef;
                    break;
                default:
                    extendableObject = AXES.Dev.getFieldById( this.id.split( "." )[ 1 ] );
                    addToNode =
                    {
                        clsid     : "axe",
                        fieldId   : extendableObject.adjustedId()
                    };
                    if ( extendableObject instanceof AXES.Layout.UserField )
                        addToNode.ud = extendableObject.id;
                    addToElements = screenDef.elements;
                }

                var propHash;
                var addToDef = false;
                
                //check for subfile columns
                if ( extendableObject.recordFormat && 
                    extendableObject.recordFormat.cols ) {                    
                    var rf  = extendableObject.recordFormat;                         
                    var cols = rf.cols;
                    var col = rf.getSubfileColumn(extendableObject);
                    var colKey = Object.keys(col)[0];
                    var sbColumn = col[colKey];                        
                    var key = rf.name.toLowerCase() + "_" + colKey;
                    addToSfColsNode =
                    {   
                        clsid     : "axscol",
                        extensions: []
                    };
                    // we are only interested for the first row field of the subfilcolumn                            
                    var exObj = sbColumn.items[0];
                    if (exObj && extendableObject.id == exObj.id) {
                        //if it is first row field then check for extension customization
                        //and apply it to sfcols
                        for (var j = 0, ex; ex = extendableObject.extensions[j]; j++) {
                            if ((j == 0) || !ex.isAutoGenerated) {
                                propHash = exTab.buildPropertyHashForSave(ex.properties);
                                propHash.values.clsid = ex.name;                               
                                if ( (j == 0) && (addToNode.clsid != "axs") ) {
                                    // will get the basic extension property and add to extensions 
                                    // this is how it work in TS
                                    if (propHash.length > 0) addToDef = true;
                                    addToNode.basePropExtension = propHash.values
                                    if (addToDef && addToElements) {                                    
                                        var found = false;
                                        for (var k in addToElements)
                                            if (addToElements.hasOwnProperty(k) && addToElements[k].fieldId == addToNode.fieldId) {
                                                addToElements[k] = addToNode;
                                                found = true;
                                                break;
                                            }
                                        if ( found ) found = false;
                                        else addToElements.push(addToNode);
                                    }
                                } else {
                                    //otherwise add to sfCols
                                    addToSfColsNode.extensions.push(propHash.values);
                                if (!screenDef.sfCols) screenDef.sfCols = {};
                                    screenDef.sfCols[key] = addToSfColsNode; 
                                }
                            }
                        } // end for
                    } else {
                        // will get the basic extension property and add to extensions 
                        // this is how it work in TS
                        for (var j = 0, ex; ex = extendableObject.extensions[j]; j++) {
                            if (j == 0 && !ex.isAutoGenerated) {
                                propHash = exTab.buildPropertyHashForSave(ex.properties);
                                propHash.values.clsid = ex.name;
                                if ( j == 0 && (addToNode.clsid != "axs") ) {// for extension
                                    if (propHash.length > 0) addToDef = true;
                                    addToNode.basePropExtension = propHash.values;
                                    if (addToDef && addToElements) {                                    
                                        var found = false;
                                        for (var k in addToElements)
                                            if (addToElements.hasOwnProperty(k) && addToElements[k].fieldId == addToNode.fieldId) {
                                                addToElements[k] = addToNode;
                                                found = true;
                                                break;
                                            }
                                        if ( found ) found = false;
                                        else addToElements.push(addToNode);
                                    }
                                } 
                            }
                        } // end for                            
                    } // endif (extendableObject.id == exObj.id)
                } else { 
                    // for non-subfile columns and screen properties
                    for (var j = 0, ex; ex = extendableObject.extensions[j]; j++) {
                        if ((j == 0) || !ex.isAutoGenerated) {
                            propHash = exTab.buildPropertyHashForSave(ex.properties);
                            propHash.values.clsid = ex.name;
                            if ( ((ex.name == "axsx") && (addToNode.clsid == "axs")) 
                                 || (j == 0 && (addToNode.clsid != "axs")) ) {// for screen basic properties
                                if (propHash.length > 0) addToDef = true;
                                addToNode.basePropExtension = propHash.values;
                            } else {
                                if (!addToNode.extensions) addToNode.extensions = [];
                                addToNode.extensions.push(propHash.values);
                                addToDef = true;
                            }
                        }
                    }
                    if (addToDef && addToElements) {
                        var found = false;
                        for (var k in addToElements)
                            if (addToElements.hasOwnProperty(k) && addToElements[k].fieldId == addToNode.fieldId) {
                                addToElements[k] = addToNode;
                                found = true;
                                break;
                            }
                        if ( found ) found = false;
                        else addToElements.push(addToNode);                           
                    }
                }
            }
        )
                     
        return screenDef;
    },

    saveAppsDefinition: function()
    {
        var exTab      = AXES.Dev.ExtensionsTab;
        var appsDef    = exTab.buildAppsDefinitionForSave();
        var saveResult = AXES.Dev.uploadFileToServer( "application_definition", "js", JSON.stringify( appsDef ) );

        if ( "" == saveResult )
        {
            // get appstyles
            var exObject = AXES.Application;
            var ex     = exObject.getExtensions("axAppStyles");
            var ex_styles = ex[0].properties.styles;
            var propVal   = AXES.Extensions._deserialisePropertyValue( ex_styles._owner, ex_styles._name, ex_styles.getJSONObject( false ) );
            var styles = propVal.v;
            if (styles.length > 0) 
            {
                var stylesheets = exTab.generateAppsStyleSheets(styles);
                //generate stylesheets if necessary
                var fileRegister = {};
                var result;
                for (var name in stylesheets)
                {
                    var stylesheet = stylesheets[name];
                    if (stylesheet.length == 0) stylesheet = " ";

                    //
                    var file = AXES.Application._file;
                    if (name.length > 0) file += "_" + name;
                    result = AXES.Dev.uploadFileToServer( file, "css", stylesheet);
                    if (!(result =="")) break;

                    //
                    var fileAsId = file.trim().toLowerCase();
                    fileRegister[fileAsId] = true;
                }

                // delete left over stylesheets
                if (result == "")
                {
                    var ds = AXES.Screens.definitionSet.trim();
                    var params =
                    {
                        path: "ts/screens/",
                        prefix: "application_definition",
                        suffix: ".css"
                    };
                    if (ds.length > 0) params.path += ds + "/";
                    var requestUrl = "/agi/lua5/extservice_3.lua";
                    $.ajax({ 
                        url: requestUrl,
                        type: "get",
                        data: params,
                        dataType: "json",
                        success: function(data, textStatus)
                        { 
                            var deleteList = [];
                            for (var i in data.filelist)
                            {
                                var fileName = data.filelist[i].name;
                                fileName = fileName.trim().toLowerCase();
                                fileName = fileName.substring(0, fileName.length - 4);
                                if (fileRegister[fileName] == null) deleteList.push(fileName);
                            }

                            // delete file from deleteList
                            for (var i in deleteList)
                            {
                                var fileName = deleteList[i];
                                AXES.Dev.deleteFileFromServer(fileName, "css");
                            }
                        },
                        error: function(XMLHttpRequest, textStatus, errorThrown)
                        {
                            AXES.Debug.error(TEXT.core("Error retrieving css list."));
                        }
                    });
                }
                // refesh application styling..                
                var extendableObject = AXES.Application;
                for ( var j = 0, ex; ex = extendableObject.extensions[ j ]; j++ )
                {
                    extendableObject.refreshExtensions(ex.name);
                }                                    
            }
            this.updateDirtyState(-10);
            exTab.AppsDefNeedSave = false;
            // all done..
            return true;
        }

        var saveDialogOptions =
        {
            bgiframe: true,
            modal   : true,
            position: "top",
            buttons : {},

            close: function()
            {
                $( this ).dialog( 'destroy' );
            }
        };

        saveDialogOptions.buttons[ TEXT.Dev( "OK" ) ] = function()
        {
            $( this ).dialog( 'close' );
        };

        $( document.createElement( "div" ) )
            .attr( "title", TEXT.Dev( "Save Definition") )
            .text( TEXT.Dev( "Unable to save application definition: %s1.", saveResult ) )
            .dialog( saveDialogOptions );
        return false;
    },

    generateAppsStyleSheets: function(styles) {
        var styleContainer = {};
        var result = {};
        //
        for (var i in styles)
        {
            var style = styles[i];

            var name = style.getPropertyValue("name").trim().toLowerCase();
            var theme = style.getPropertyValue("theme").trim().toLowerCase();
            //
            if (result[theme] == null) result[theme] = {};
            //
            var id = name;
            if (id.length > 0)
            {
                if (theme.length > 0) id += "$" + theme;
                styleContainer[id] = style;
            }
        }

        for (var i in styles)
        {
            var style = styles[i];
            var theme = style.getPropertyValue("theme").trim().toLowerCase();
            var styleName = style.getPropertyValue("name").trim().toLowerCase();
            // get html tag
            var styleFor = style.getPropertyValue("styleFor").trim().toLowerCase();
            var htmlTag = "";
            switch (styleFor)
            {
                case "all": htmlTag = "*"; break;
                case "body": htmlTag = "body,#terminalInner"; break;
                case "html": htmlTag = style.getPropertyValue("htmlTag").trim(); break;
            }

            // get style names
            var id = styleName;
            if (htmlTag.length > 0) id += "$" + htmlTag;

            if (styleName.length > 0) styleName = ".ax_" + AXES.Utils.textToCSSName(styleName);
            if (htmlTag.length > 0)
            {
                if (styleName.length > 0) styleName += ",";
                styleName += htmlTag;
            }


            // get for 5250 attributes
             var for5250Attributes = style.properties.for5250Attributes.getRawValue().v;

            for (var j in for5250Attributes)
            {
                var str5250Attribute = for5250Attributes[j];
                if (str5250Attribute.length > 0)
                {
                    if (styleName.length > 0) styleName += ",";
                    styleName += " .ax_x" + str5250Attribute;
                }
            }

            var cssText = styleName + "{";

            //
            var traversed = {};
            var traverse = style;
            var effStyle = {};
            do
            {
                // get the style
                var s = traverse.getPropertyValue("style");
                for (var attr in s)
                {
                    if (effStyle[attr] == null) effStyle[attr] = s[attr];
                }

                //
                var tid = traverse.getPropertyValue("name").trim().toLowerCase();
                if (tid.length > 0) traversed[tid] = true;

                // get base style 
                var baseStyleId = "";
                if (traverse.properties.style != null) baseStyleId = traverse.properties.style.getRawValue().v['_base'];

                traverse = null;

                if (baseStyleId != null)
                {
                    //baseStyleId = baseStyleId.getRawValue().v.trim().toLowerCase();
                    if (baseStyleId.length > 0)
                    {
                        if (theme.length > 0) traverse = styleContainer[baseStyleId + "$" + theme];
                        if (traverse == null) traverse = styleContainer[baseStyleId];

                        // make sure the style has not been visited to prevent infinite loop
                        if (traverse != null)
                        {
                            var tid = traverse.getPropertyValue("name").trim().toLowerCase();
                            if (traversed[tid] != null) traverse = null;
                        }
                    }
                }
                
            } while (traverse != null);

            // compose the CSS styles
            cssText += AXES.Utils.getCssText(effStyle);
            cssText += "}";

            // get theme
            if (theme.length == 0)
            {
                for (var theme in result)
                {
                    var list = result[theme];
                    if (list[id] == null) list[id] = cssText;
                }
            }
            else
            {
                var list = result[theme];
                list[id] = cssText;
            }
        }

        for (var theme in result)
        {
            var list = result[theme];
            result[theme] = AXES.Utils.joinStrings(list, "\r\n");
        }

        return result;
    },

    saveScreen: function()
    {
        var terminal     = AXES.Terminal;
        var dev          = AXES.Dev;
        var screenName   = terminal.symbolicName || "";
        var screenDef    = this.buildScreenDefinitionForSave();
        var extensionTab = this;

        function doSave()
        {
            if ( "" != dev.uploadFileToServer( "Screen_" + screenName, "js", JSON.stringify( screenDef ) ) )
            {
                return;
            }

            AXES.Extensions.replaceCurrentScreenDefinition( screenDef );
            extensionTab.currentDefinition = screenDef;
            extensionTab.updateDirtyState( -10 );
            extensionTab.ScreenDefNeedSave = false;
        }

        if ( "" != screenName )
        {
            doSave();
            return true;
        }

        var locked = dev.screenLocked;
        if ( locked )
            dev.screenLocked = false;

        $(
            '<div id="screenname-dialog" style="text-align: center">\
                <input type="text" name="screenname" size="30" value="">\
            </div>'
        ).dialog(
            {
                hide : 'clip',
                modal: true,
                show : 'clip',
                title: TEXT.Dev( "Please provide the screen name." ),
                width: 250,

                close: function( e )
                {
                    $( this ).empty();
                    $( this ).dialog( 'destroy' );
                    dev.screenLocked = locked;
                },

                buttons: [
                    {
                        text : TEXT.Dev( "OK" ),
                        click: function()
                        {
                            screenName = $.trim( $( this ).find( 'input' ).val() );
                            if ( "" == screenName )
                            {
                                $( this ).dialog( "close" );
                                return;
                            }

                            var screensTab = dev.ScreensTab;
                            screensTab.currentDef.name = screenName;
                            screensTab.saveDefinition();

                            terminal.matchedDefinition = AXES.Screens.findMatch( terminal );
                            terminal.updateWithMatch( terminal.matchedDefinition ); 

                            window.setTimeout(
                                function()
                                {
                                    terminal.fire(
                                        "ScreenArrived",
                                        [ { terminal: terminal } ]
                                    );
                                },
                                1
                            );

                            doSave();

                            $( this ).dialog( "close" );
                        }
                    },
                    {
                        text: TEXT.Dev( "Cancel" ),
                        click: function()
                        {
                            $( this ).dialog( "close" );
                        }
                    }
                ]
            }
        );

        return true;
    },
    /**
     * Handles the Reset button click event. Deletes the current definition and resets the screen to the default.
     * @param {Event} e
     */
    onResetClicked : function(e) {
        var delDialogOptions = {
            bgiframe : true,
            modal : true,
            position : "top",
            buttons : {},
            close: function()
            {
                $(this).dialog('destroy');
            }
        };
        delDialogOptions.buttons[TEXT.Dev("Yes")] = function() {
            $(this).dialog('close');
            AXES.Dev.ExtensionsTab._resetScreen();            
            var exTab = AXES.Dev.ExtensionsTab;        
            if (exTab.AppsDefNeedSave) exTab.AppsDefNeedSave = false;
            if (exTab.ScreenDefNeedSave) exTab.ScreenDefNeedSave = false; 
        };
        delDialogOptions.buttons[TEXT.Dev("No")] = function() {
            $(this).dialog('close');
        };

        $(document.createElement("div"))
        .attr("title", TEXT.Dev("Delete Screen Extensions"))
        .text(TEXT.Dev("Are you sure you want to reset this screen? All customizations will be lost."))
        .dialog(delDialogOptions);

        return false;
    },
    _resetScreen: function()
    {
        var sn = AXES.Terminal.symbolicName, deleteResult, delDialogOptions;
        if (!sn)
        {
            return;
        }
        deleteResult = AXES.Dev.deleteFileFromServer("screen_" + sn, "js");
        if (deleteResult == "") 
        {
            AXES.Extensions.deleteCurrentScreenDefinition();
            AXES.Terminal.screenLoaded();
        }
        else 
        {
            delDialogOptions = {
                bgiframe : true,
                modal : true,
                position : "top",
                buttons : {},
                close: function()
                {
                    $(this).dialog('destroy');
                }
            };
            delDialogOptions.buttons[TEXT.Dev("OK")] = function()
            {
                $(this).dialog('close');
            };

            $(document.createElement("div"))
            .attr("title", TEXT.Dev("Delete Screen Extensions"))
            .text(TEXT.Dev("Unable to delete screen definition: %s1.", deleteResult))
            .dialog(delDialogOptions);
        }
    },
    onCancelClicked: function(e) 
    {
        var exTab = AXES.Dev.ExtensionsTab;        
        if (exTab.AppsDefNeedSave)
        {
            // todo reset apps property changes
            exTab.AppsDefNeedSave = false;
        }
        if (exTab.ScreenDefNeedSave)
        {                     
            AXES.Extensions.deleteCurrentScreenDefinition();
            AXES.Terminal.screenLoaded();
            exTab.ScreenDefNeedSave = false;
        }               
    },
    _renderAddedExtension: function( exObject, ex, exIndex )
    {
        if ( exObject === AXES.Application )
        {
            if (AXES.Dev.screenLocked) { // do a refresh
                AXES.Dev.setScreenLocked(false);
                AXES.Dev.setScreenLocked(true);
            } else {
                AXES.Dev.doLayout();            
            }            
            this.selectFieldRow( $( "#ax-fldRow-application" ) );
            return;
        }

        if ( exObject === AXES.Layout )
        {
            if (AXES.Dev.screenLocked) { // do a refresh
                AXES.Dev.setScreenLocked(false);
                AXES.Dev.setScreenLocked(true);
            } else {
                AXES.Dev.doLayout();            
            }
            this.selectFieldRow( $( "#ax-fldRow-screen" ) );
            return;
        }

        if ( exIndex === undefined )
            exIndex = exObject.extensions.length;

        if ( exObject.extensions.length > exIndex )
        {
            AXES.Layout.redrawField( exObject );
            this.selectFieldRow( $( "#ax-fldRow-ex\\." + exObject.id ) );
            return;
        }

        AXES.Layout.drawFieldExtension( exObject, ex );
        if ( !$( "#ax-extensions-fieldDetail-exList table" )[ 0 ] || 2 <= $( "#ax-extensions-fieldDetail-exList table" )[ 0 ].childElementCount )
            this.renderFieldExtension( ex, exObject, exIndex, false );
        else
            $( "#ax-extensions-fieldDetail-exList table" )[ 0 ].appendChild( this.renderFieldExtension( ex, exObject, exIndex, false ) );
        
        this.redrawExTab();            
        this.selectFieldRow( $( "#ax-fldRow-ex\\." + exObject.id ) );
    },
    addExtensionToExObj : function(exName, exObj, name) {
        var propV = {};
        if (exName == "axsx") {
            propV = {
                "name" : {
                    "v" : name
                }
            };
        } else if (exName == "axButton") {
            var basePropStyle = exObj.extensions[0].properties["style"];
            propV.style = basePropStyle;
        }
        // do a check to prevent adding more axtrm and appstyles
        if ( (exObj.definition) && (exObj.definition.clsid == "axapp") ) {
            for (var i = 0, exInfo; exInfo = exObj.extensions[i]; i++) {
                if (exInfo.name == exName ) {
                    alert(TEXT.Dev("not allowed to add more than 1 %s1 extension", exName));
                    return; 
                }
            }
        }
        var ex = AXES.Extensions.createNewExtension(exName, propV, exObj, false);
               
        this._undo.addUndo(
        {
            undo: function()
            {
                // eXtension may have been recreated during redraw so we can't rely on the ex variable in the closure
                var $tbody = $("#ax-extensions-fieldDetail-exList tbody:last");
                ex = $tbody.children("tr:first").data("extension");
                if (exObj.removeExtension(ex))
                    $tbody.remove();
            },
            undoArgs : [],
            redo : function() {
                exObj.extensions.push(ex);
                this._renderAddedExtension(exObj, ex);
            },
            redoArgs : [],
            updateDirtyState : true,
            context : this
        });
        
        if (ex.type == "Application") this.AppsDefNeedSave = true;
        else this.ScreenDefNeedSave = true;

        this._renderAddedExtension(exObj, ex);        
    },
    removeExtensionFromExObj: function( exObj, exIndex )
    {
        var tbodySelector = "#ax-extensions-fieldDetail-exList tbody:eq(" + exIndex + ")";
        var ex = exObj.extensions[ exIndex ];

        exObj.removeExtension( ex );
        $( tbodySelector ).remove();

        if ( "Application" == ex.type )
            this.AppsDefNeedSave = true;
        else
            this.ScreenDefNeedSave = true;

        this._undo.addUndo(
            {
                context         : this,
                redoArgs        : [],
                undoArgs        : [],
                updateDirtyState: true,

                undo: function()
                {
                    exObj.extensions.splice( exIndex, 0, ex );
                    this._renderAddedExtension( exObj, ex, exIndex );
                },
                redo: function()
                {
                    ex = exObj.extensions[ exIndex ];
                    exObj.removeExtension( ex );
                    $( tbodySelector ).remove();
                }
            }
        );
    },
    isSaveNeeded : function() {
        return ($("#btnExSave:enabled", this.tabContent).length > 0);
    },
    checkIfSaveNeededAndSave : function( callBackToContinue )
    {
        var et = this;
        if ( !et.isSaveNeeded() )
        {
            return true;
        }

        var dialogOptions =
        {
            closeOnEscape: false,
            bgiframe     : true,
            modal        : true,
            position     : "top",
            buttons      : {}
        }
        dialogOptions.buttons[ TEXT.Dev( "Save" ) ] = function()
        {
            $( this ).dialog( 'close' ).dialog( 'destroy' );
            if (et.ScreenDefNeedSave)  
                et.saveScreen() ;
            if (et.AppsDefNeedSave)
                et.saveAppsDefinition();               
            callBackToContinue();
        };
        dialogOptions.buttons[ TEXT.Dev( "Discard" ) ] = function()
        {
            $( this ).dialog( 'close' ).dialog( 'destroy' );
            et.onCancelClicked();
            callBackToContinue();
        };
        dialogOptions.buttons[ TEXT.Dev( "Cancel" ) ] = function()
        {
            $( this ).dialog( 'close' ).dialog( 'destroy' );
        };

        $( document.createElement( "div" ) )
            .attr( "title", TEXT.Dev( "Customizations Changed" ) )
            .text( TEXT.Dev( "Do you want to save your definition changes first?" ) )
            .dialog( dialogOptions );
        return false;
    }
};

/**
 * Manages the AutoGUI tab.
 * <strong>Note:</strong> This namespace is only available in developer mode.
 * @namespace
 */
AXES.Dev.AutoGUITab = {
    initDone : false,
    init: function()
    {
        var self = this;
        this.initDone = true;

        this.tabContent = $("#ax-tab-autogui");
        TEXT.apply(window, this.tabContent[0]);

        $("#ax-autogui-accordian").accordion({
            fillSpace : true,
            active : 0
        });

        $("#ax-tab-autogui fieldset.rowColSearch input:radio").live("click", function(e)
        {
            var startEndInfo = $(this).parents(".displayLine").next();
            if (this.value == "true") 
                startEndInfo.slideUp("fast").find("input").attr("disabled", "disabled");
            else 
                startEndInfo.slideDown("fast").find("input").attr("disabled", false);
            $("#btnAGSave,#btnAGCancel").attr("disabled", false);
        });
        $("#ax-autoguirule-FKeyGUIPanel").live("click", function(e)
        {
            if ( $("#ax-autoguirule-FKeyGUIPanel").is(":checked") == true )
                $("#FKeyGUIPanelOptions").attr("disabled", false);
            else
                $("#FKeyGUIPanelOptions").attr("disabled", "disabled");
        });

        $("#fkeyformpanel").validate({
            rules : {
                FKeyGUIPanelHorizontalWidth : {
                    required : true,
                    max : 300,
                    min : 100
                },
                FKeyGUIPanelVerticalHeight : {
                    required : true,
                    max : 100,
                    min : 30
                },
                FKeyGUIPanelButtonWidth : {
                    required : true,
                    max : 290,
                    min : 10
                },
                FKeyGUIPanelButtonHeight : {
                    required : true,
                    max : 90,
                    min : 10
                },
                FKeyGUIPanelButtonGap : {
                    required : true,
                    max : 30,
                    min : 1
                }
            }
        });
        $("#ax-tab-autogui :text").bind("change", function(e)
        {
            $("#btnAGSave,#btnAGCancel").attr("disabled", false);
        });
        $("#ax-tab-autogui input:checkbox").live("click", function(e)
        {
            $("#btnAGSave,#btnAGCancel").attr("disabled", false);
        });
        $("#FKeyGUIPanelOptions select").bind("change", function(e)
        {
            $("#btnAGSave,#btnAGCancel").attr("disabled", false);
        });
        $("#btnAGSave").click(this.saveChanges);
        $("#btnAGCancel").click(this.cancelChanges);

        this.resize();
        this.updateDisplay();

    },
    onShow: function()
    {
        if (!this.initDone)
        {
            return;
        }
        this.resize();
    },
    resize : function() {
        if (this.initDone && $("#ax-tab-autogui").is(":visible")) {
            this.tabContent.height($("#ax-developerTab").innerHeight() - $("#ax-developerTab>ul").outerHeight());
            $("#ax-autogui-accordianWrapper").height(this.tabContent.innerHeight() - $("#ax-autogui-footer").outerHeight());
            $("#ax-autogui-accordian").accordion("resize");
        }
    },
    ruleIdToElId: function(ruleId)
    {
        return "ax-autoguirule-" + ruleId;
    },
    elIdToRuleId: function(elId)
    {
        return (elId.substring(0, 15) == "ax-autoguirule-") ? elId.substring(15) : "";
    },
    updateDisplay: function()
    {
        var self = this;
        var rules = AXES.Layout.AutoGUI.rules;

        $("#ax-autogui-accordian input:text").each(function()
        {
            $(this).val(self.getRuleValue(this.name));
        });
        $("#ax-autogui-accordian select").each(function()
        {
            $(this).val(self.getRuleValue(this.name));
        });
        $("#ax-autogui-accordian textarea").each(function()
        {
            $(this).text(self.getRuleValue(this.name));
        });
        $("#ax-autogui-accordian input:checkbox").each(function()
        {
            this.checked = self.getRuleValue(this.name);
        });
        $("#ax-autogui-accordian input:radio").each(function()
        {
            this.checked = (self.getRuleValue(this.name).toString() == this.value);
            if (this.value == "true") 
            {
                if (this.checked) 
                    $(this).parents(".displayLine").next().hide().find("input").attr("disabled", "disabled");
                else 
                    $(this).parents(".displayLine").next().show().find("input").attr("disabled", false);
            }
        });

        $("#btnAGSave,#btnAGCancel").attr("disabled", "disabled");
        AXES.Dev.doLayout();
    },
    getRuleValue: function(ruleID) 
    {
        var o = AXES.Layout.AutoGUI.rules, path = ruleID.split("_"), p, i;
        // Skip the first item as it is the ax-autoguirule prefix
        for (i=1; i<path.length; i++) 
        {
            p = path[i - 1];
            if (o[p] == undefined)
                o[p] = {};
            o = o[p];
        }
        return o[path[i - 1]];
    },
    updateRule: function(ruleID, value) 
    {
        var o = AXES.Layout.AutoGUI.rules, path = ruleID.split("_"), p, i;
        // Skip the first item as it is the ax-autoguirule prefix
        for (i=1; i<path.length; i++) 
        {
            p = path[i - 1];
            if (o[p] == undefined)
                o[p] = {};
            o = o[p];
        }
        o[path[i - 1]] = value;
    },
    saveChanges: function(e)
    {
        var self = AXES.Dev.AutoGUITab, saveResult;

        $("#ax-autogui-accordian :text").each(function()
        {
            self.updateRule(this.name, this.value);
        });
        $("#ax-autogui-accordian input:checkbox").each(function()
        {
            self.updateRule(this.name, this.checked);
        });
        $("#ax-autogui-accordian input:radio:checked").each(function()
        {
            self.updateRule(this.name, this.value);
        });
        $("#ax-autogui-accordian select").each(function()
        {
            self.updateRule(this.name, this.value);
        });

        saveResult = AXES.Dev.uploadFileToServer("axautoguirules", "json", JSON.stringify(AXES.Layout.AutoGUI.rules));
        if (saveResult == "") 
        {
            $("#btnAGSave,#btnAGCancel").attr("disabled", "disabled");
            AXES.Dev.doLayout();
        }
        else 
        {
            var saveDialogOptions = {
                bgiframe : true,
                modal : true,
                position : "top",
                buttons : {},
                close: function()
                {
                    $(this).dialog('destroy');
                }
            };
            saveDialogOptions.buttons[TEXT.Dev("OK")] = function()
            {
                $(this).dialog('close');
            };

            $(document.createElement("div")).attr("title", TEXT.Dev("Save AutoGUI Rules")).text(TEXT.Dev("Unable to save rules: %s1.", saveResult)).dialog(saveDialogOptions);
        }

    },
    cancelChanges: function(e)
    {
        var self = AXES.Dev.AutoGUITab;
        self.updateDisplay();
    }
}

/**
 * @namespace
 * Contains and manages editors for custom eXtension properties.
 *
 * Apart from the init method, this namespace is not accessed directly. It contains
 * a set of object properties that will be used to extend the corresponding property
 * object.
 *
 * <strong>Note:</strong> This namespace is only available in developer mode.
 */
AXES.Dev.CustomPropertyEditors =
{
    init: function()
    {
        var o = AXES.Objects;
        o.updateObjectClass( "axZoomValue", this.axZoomValue );
        o.updateObjectClass( "axAppStyle",  this.axAppStyle );
        o.updateObjectClass( "Style",       this.axStyle );
        o.updateObjectClass( "Html",        this.axBasicInput );
        o.updateObjectClass( "URL",         this.axBasicInput );
        o.updateObjectClass( "MString",     this.axBasicInput );
        o.updateObjectClass( "ScreenName",  this.axBasicInput );
        o.updateObjectClass( "String",      this.axBasicInput );
        o.updateObjectClass( "Table",       this.axBasicInput );
        o.updateObjectClass( "Integer",     this.axBasicInput );
        o.updateObjectClass( "Number",      this.axBasicInput );
        o.updateObjectClass( "Boolean",     this.axBoolean );
    },
    axZoomValue : {
        _renderEditor: function(setDisabled)
        {
            var property = this,
            propVal = property.getRawValue().v,
            slider, sel, opt, div, newVal, sliderVal,
            retVal = this._UIElement = document.createElement("div"),
            exTab = AXES.Dev.ExtensionsTab;

            sel = document.createElement("select");
            sel.style.display = "block";
            sel.style.width = "100%";

            opt = document.createElement("option");
            opt.setAttribute("value", 0);
            opt.appendChild(document.createTextNode("Use current"));
            if (propVal == 0)
                opt.setAttribute("selected", "selected");
            sel.appendChild(opt);

            opt = document.createElement("option");
            opt.setAttribute("value", -1);
            opt.appendChild(document.createTextNode("Auto"));
            if (propVal == -1)
                opt.setAttribute("selected", "selected");
            sel.appendChild(opt);

            opt = document.createElement("option");
            opt.setAttribute("value", -2);
            opt.appendChild(document.createTextNode("Fit"));
            if (propVal == -2)
                opt.setAttribute("selected", "selected");
            sel.appendChild(opt);

            opt = document.createElement("option");
            opt.setAttribute("value", 1);
            opt.appendChild(document.createTextNode("Custom"));
            if (propVal > 0)
                opt.setAttribute("selected", "selected");
            sel.appendChild(opt);

            sel.disabled = setDisabled;
            $(sel).change(
                function(e)
                {
                    newVal = $(this).val();
                    if ( newVal )
                    {
                        sliderVal = slider.slider("value");
                        slider.slider("enable").next().attr("disabled", false).next().text(sliderVal + "%");
                        exTab.updatePropertyValue(property, { v : sliderVal });
                    }
                    else 
                    {
                        slider.slider("disable").next().attr("disabled", "disabled");
                        exTab.updatePropertyValue( property, { v : newVal } );
                    }
                }
            );
            retVal.appendChild(sel);

            div = document.createElement("div");
            div.style["top"] = "4px";
            div.style["margin-bottom"] = "5px";

            slider = $(div).slider({
                min : 5,
                max : 200,
                step : 5,
                value : (propVal > 0) ? propVal : 100,
                disabled : setDisabled || (propVal <= 0),
                change: function(event, ui)
                {
                    exTab.updatePropertyValue(property, {
                        v : ui.value
                    });
                    $(this).next().text(ui.value + "%");
                },
                slide: function(event, ui)
                {
                    $(this).next().text(ui.value + "%");
                }
            });

            retVal.appendChild(div);
            div = document.createElement("div");
            div.style.textAlign = "center";
            div.appendChild(document.createTextNode((propVal > 0) ? propVal : "\u00a0"));
            retVal.appendChild(div);

            return retVal;
        },
        _refreshEditorDisplay: function() 
        {
            var newVal = this.getRawValue().v;
            $(this._UIElement).find("select").val(newVal < 0 ? newVal : 1);
            $(this._UIElement).find("div").slider("option", "disabled", (newVal <= 0)).slider("value", (newVal > 0) ? newVal : 100);
        }
    },
    axAppStyle : {
        _renderEditor: function(setDisabled) 
        {
            var retVal = document.createElement("button");
            retVal.className = "ax-extension-editScriptButton";
            retVal.appendChild(document.createTextNode( setDisabled ? "View Styles..." : "Edit Styles..."));
            return retVal;
        }
    },
    axStyle:
    {
        _renderEditor: function( setDisabled )
        {
            var interfaceVal = AXES.Application.getExtensions( "axAppStyles" )[ 0 ].properties.interface.v;
            if ( !interfaceVal )
                interfaceVal = "list";
            var btnText      = setDisabled ? TEXT.Dev( "View Styles..." ) : TEXT.Dev( "Edit Styles..." );
            var retVal       = document.createElement( "button" );
            retVal.className = "ax-extension-editStyleButton";
            retVal.appendChild( document.createTextNode( btnText ) );

            $( retVal ).bind(
                "click",
                {
                    isDisabled: setDisabled,
                    property  : this,
                    title     : btnText
                },
                "list" == interfaceVal ? this.editStyle2 : this.editStyle
            );

            this._UIElement = null;
            return retVal;
        }
    },
    axBasicInput : {
        _renderEditor: function( setDisabled )
        {
            var retVal       = document.createElement( "div" )
            retVal.className = "ax-css3box";
            var property     = this;
            var inp          = property._UIElement = document.createElement( "input" );
            inp.className    = "ax-css3FlexBox1";
            inp.value        = property.getRawValue().v;
            var $inp         = $( inp )
                .attr( "disabled", setDisabled ? "disabled" : false )
                .change(
                    function( e )
                    {
                        AXES.Dev.ExtensionsTab.updatePropertyValue(
                            property,
                            { v: "Integer" == property.type ? parseInt( this.value ) : ( "Number" == property.type ? Number( this.value ) : this.value ) }
                        );
                    }
                );
            switch ( property.type )
            {
            case "ScreenName":
                $inp.constrain( { allow: { regex : "[a-zA-Z0-9_-]" } } );
                break;
            case "Integer":
                $inp.numeric( { format: "0" } );
                break;
            case "Number":
                $inp.numeric( { format: "0.0" } );
                break;
            }
            retVal.appendChild( inp );
            return retVal;
        },
        _refreshEditorDisplay: function()
        {
            if (this._UIElement) this._UIElement.value = this.getRawValue().v;
        }
    },
    axBoolean : {
        _renderEditor: function(setDisabled)
        {
            var property = this, retVal = this._UIElement = document.createElement("input");
            retVal.type = "checkbox";
            retVal.checked = (property.getRawValue().v == true);
            $(retVal).attr("disabled", setDisabled ? "disabled" : false).click(function(e) {
                AXES.Dev.ExtensionsTab.updatePropertyValue(property, {
                    v : this.checked
                });
            });
            return retVal;
        },
        _refreshEditorDisplay: function() 
        {
            if (this._UIElement) this._UIElement.checked = (this.getRawValue().v == true);
        }
    },
    axScript:
    {
        _renderEditor: function( setDisabled )
        {
            return this.renderScriptEditor(setDisabled);
        }
    }
}
AXES.Dev.JSAutoComplete = {
    editor : null,
    startComplete: function()
    {
        // We want a single cursor position.
        var editor = this.editor, self = this;
        if (!editor || editor.somethingSelected())
        {
            return;
        }
        // Find the token at the cursor
        var cur = editor.getCursor(false), token = editor.getTokenAt(cur), tprop = token;
        // If it's not a 'word-style' token, ignore the token.
        if (!/^[\w$_]*$/.test(token.string)) 
        {
            token = tprop = {
                start : cur.ch,
                end : cur.ch,
                string : "",
                state : token.state,
                className : token.string == "." ? "property" : null
            };
        }
        // If it is a property, find out what it is a property of.
        while (tprop.className == "property") 
        {
            tprop = editor.getTokenAt({
                line : cur.line,
                ch : tprop.start
            });
            if (tprop.string != ".")
            {
                return;
            }
            tprop = editor.getTokenAt({
                line : cur.line,
                ch : tprop.start
            });
            if (!context)
                var context = [];
            context.push(tprop);
        }
        var completions = this.getCompletions(token, context);
        if (!completions.length)
        {
            return;
        }
        function insert(str)
        {
            editor.replaceRange(str, {
                line : cur.line,
                ch : token.start
            }, {
                line : cur.line,
                ch : token.end
            });
        }
        // When there is only one completion, use it directly.
        if (completions.length == 1) 
        {
            insert(completions[0]);
            return true;
        }

        // Build the select widget
        var complete = document.createElement("div");
        complete.className = "completions";
        var sel = complete.appendChild(document.createElement("select"));
        sel.multiple = true;
        for (var i = 0; i < completions.length; ++i) 
        {
            var opt = sel.appendChild(document.createElement("option"));
            opt.appendChild(document.createTextNode(completions[i]));
        }
        sel.firstChild.selected = true;
        sel.size = Math.min(10, completions.length);
        var pos = editor.cursorCoords(cur, "page");
        complete.style.left = pos.left + "px";
        complete.style.top = pos.bottom + "px";
        document.body.appendChild(complete);
        // Hack to hide the scrollbar.
        if (completions.length <= 10)
            complete.style.width = (sel.clientWidth - 1) + "px";

        var done = false;
        function close()
        {
            if (done)
            {
                return;
            }
            done = true;
            complete.parentNode.removeChild(complete);
            self.editor = null;
        }
        function pick()
        {
            insert(sel.options[sel.selectedIndex].value);
            close();
            setTimeout(function()
            {
                editor.focus();
            }, 50);
        }
        $(sel).blur(close).keydown(function(event) 
        {
            switch (event.which) 
            {
                case 13:
                case 32:
                    event.preventDefault();
                    event.stopPropagation();
                    pick();
                    break;
                case 27:
                    event.preventDefault();
                    event.stopPropagation();
                    close();
                    editor.focus();
                    break;
                case 38:
                case 40:
                    break;
                default:
                    close();
                    editor.focus();
                    setTimeout(this.startComplete, 50);
            }
        }).dblclick(pick);

        sel.focus();
        // Opera sometimes ignores focusing a freshly created node
        if (window.opera)
            setTimeout(function() {
                if (!done)
                    sel.focus();
            }, 100);
        return true;
    },

    stringProps: ("charAt charCodeAt indexOf lastIndexOf substring substr slice trim trimLeft trimRight " +
    "toUpperCase toLowerCase split concat match replace search").split(" "),
    arrayProps: ("length concat join splice push pop shift unshift slice reverse sort indexOf " +
    "lastIndexOf every some filter forEach map reduce reduceRight ").split(" "),
    funcProps : "prototype apply call bind".split(" "),
    keywords: ("break case catch continue debugger default delete do else false finally for function " +
    "if in instanceof new null return switch throw true try typeof var void while with").split(" "),

    getCompletions: function(token, context)
    {
        var found = [], start = token.string, props;
        function maybeAdd(key, str)
        {
            if (str.indexOf(start) == 0)
                found.push(str);
        }
        function gatherCompletions(obj)
        {
            if ( typeof obj == "string")
                props = this.stringProps;
            else if ( obj instanceof Array)
                props = this.arrayProps;
            else if ( obj instanceof Function)
                props = this.funcProps;
            if (props)
                jQuery.each(props, maybeAdd);
            for (var name in obj)
            if (name.charAt(0) != "_")
                maybeAdd(null, name);
        }

        if (context) 
        {
            // If this is a property, see if it belongs to some object we can
            // find in the current environment.
            var obj = context.pop(), base;
            if (obj.className == "variable") {
                if (obj.string == "ENV") {
                    base = this.ENV;
                } else {
                    base = window[obj.string];
                }
            } else if (obj.className == "string")
                base = "";
            else if (obj.className == "atom")
                base = 1;
            while (base != null && context.length)
            base = base[context.pop().string];
            if (base != null)
                gatherCompletions(base);
        }
        else 
        {
            // If not, just look in the window object and any local scope
            // (reading into JS mode internals to get at the local variables)
            for (var v = token.state.localVars; v; v = v.next)
                maybeAdd(null, v.name);
            gatherCompletions(window);
            jQuery.each(this.keywords, maybeAdd);
        }
        return found;
    }
}

AXES.Dev.CSSAutoComplete = {
    editor : null,
    startComplete: function()
    {
        // We want a single cursor position.
        var editor = this.editor, self = this;
        if (!editor || editor.somethingSelected())
        {
            return;
        }
        // Find the token at the cursor
        var cur = editor.getCursor(false), token = editor.getTokenAt(cur), context;
        // If it's not a 'word-style' token, ignore the token.
        if (!/^[\w-]*$/.test(token.string)) 
        {
            token = {
                start : cur.ch,
                end : cur.ch,
                string : "",
                state : token.state,
                className : "variable"
            };
        }

        var completions = this.getCompletions(token, context);
        if (!completions.length)
        {
            return;
        }
        function insert(str)
        {
            editor.replaceRange(str, {
                line : cur.line,
                ch : token.start
            }, {
                line : cur.line,
                ch : token.end
            });
        }

        // When there is only one completion, use it directly.
        if (completions.length == 1) 
        {
            insert(completions[0]);
            return true;
        }

        // Build the select widget
        var complete = document.createElement("div");
        complete.className = "completions";
        var sel = complete.appendChild(document.createElement("select"));
        sel.multiple = true;
        for (var i = 0; i < completions.length; ++i) 
        {
            var opt = sel.appendChild(document.createElement("option"));
            opt.appendChild(document.createTextNode(completions[i]));
        }
        sel.firstChild.selected = true;
        sel.size = Math.min(10, completions.length);
        var pos = editor.cursorCoords(cur, "page");
        complete.style.left = pos.left + "px";
        complete.style.top = pos.bottom + "px";
        document.body.appendChild(complete);
        // Hack to hide the scrollbar.
        if (completions.length <= 10)
            complete.style.width = (sel.clientWidth - 1) + "px";

        var done = false;
        function close()
        {
            if (done)
            {
                return;
            }
            done = true;
            complete.parentNode.removeChild(complete);
            self.editor = null;
        }
        function pick()
        {
            insert(sel.options[sel.selectedIndex].value);
            close();
            setTimeout(function()
            {
                editor.focus();
            }, 50);
        }
        $(sel).blur(close).keydown(function(event) 
        {
            switch (event.which) 
            {
                case 13:
                case 32:
                    event.preventDefault();
                    event.stopPropagation();
                    pick();
                    break;
                case 27:
                    event.preventDefault();
                    event.stopPropagation();
                    close();
                    editor.focus();
                    break;
                case 38:
                case 40:
                    break;
                default:
                    close();
                    editor.focus();
                    setTimeout(this.startComplete, 50);
            }
        }).dblclick(pick);

        sel.focus();
        // Opera sometimes ignores focusing a freshly created node
        if (window.opera)
            setTimeout(function()
            {
                if (!done)
                    sel.focus();
            }, 100);
        return true;
    },

    cssDictionary : {
        "background" : 1,
        "background-color" : {
            "transparent" : 2,
            "fixed" : 2
        },
        "background-image" : {
            "url('/|')" : 2
        },
        "background-repeat" : 1,
        "background-position" : 1,
        "background-attachment" : {
            "scroll" : 2,
            "fixed" : 2
        },
        "border" : {
            "solid |" : 2,
            "dashed |" : 2,
            "dotted |" : 2
        },
        "border-top" : 1,
        "border-right" : 1,
        "border-bottom" : 1,
        "border-left" : 1,
        "border-color" : 1,
        "border-width" : 1,
        "border-style" : 1,
        "border-spacing" : 1,
        "border-collapse" : {
            "collapse" : 2,
            "separate" : 2
        },
        "bottom" : {
            "px" : 2,
            "em" : 2,
            "%" : 2
        },
        "clear" : {
            "left" : 2,
            "right" : 2,
            "both" : 2,
            "none" : 2
        },
        "clip" : 1,
        "color" : {
            "#|" : 2,
            "rgb(#|0,0,0)" : 2
        },
        "content" : 1,
        "cursor" : {
            "default" : 2,
            "pointer" : 2,
            "move" : 2,
            "text" : 2,
            "wait" : 2,
            "help" : 2,
            "progress" : 2,
            "n-resize" : 2,
            "ne-resize" : 2,
            "e-resize" : 2,
            "se-resize" : 2,
            "s-resize" : 2,
            "sw-resize" : 2,
            "w-resize" : 2,
            "nw-resize" : 2
        },
        "display" : {
            "none" : 2,
            "block" : 2,
            "inline" : 2,
            "inline-block" : 2,
            "table-cell" : 2
        },
        "empty-cells" : {
            "show" : 2,
            "hide" : 2
        },
        "float" : {
            "left" : 2,
            "right" : 2,
            "none" : 2
        },
        "font-family" : {
            "Arial" : 2,
            "Comic Sans MS" : 2,
            "Consolas" : 2,
            "Courier New" : 2,
            "Courier" : 2,
            "Georgia" : 2,
            "Monospace" : 2,
            "Sans-Serif" : 2,
            "Segoe UI" : 2,
            "Tahoma" : 2,
            "Times New Roman" : 2,
            "Trebuchet MS" : 2,
            "Verdana" : 2
        },
        "font-size" : {
            "px" : 2,
            "em" : 2,
            "%" : 2
        },
        "font-weight" : {
            "bold" : 2,
            "normal" : 2
        },
        "font-style" : {
            "italic" : 2,
            "normal" : 2
        },
        "font-variant" : {
            "normal" : 2,
            "small-caps" : 2
        },
        "font" : 1,
        "height" : {
            "px" : 2,
            "em" : 2,
            "%" : 2
        },
        "left" : {
            "px" : 2,
            "em" : 2,
            "%" : 2
        },
        "letter-spacing" : {
            "normal" : 2
        },
        "line-height" : {
            "normal" : 2
        },
        "list-style" : 1,
        "list-style-image" : 1,
        "list-style-position" : 1,
        "list-style-type" : {
            "none" : 2,
            "disc" : 2,
            "circle" : 2,
            "square" : 2,
            "decimal" : 2,
            "decimal-leading-zero" : 2,
            "lower-roman" : 2,
            "upper-roman" : 2,
            "lower-greek" : 2,
            "lower-latin" : 2,
            "upper-latin" : 2,
            "georgian" : 2,
            "lower-alpha" : 2,
            "upper-alpha" : 2
        },
        "margin" : {
            "px" : 2,
            "em" : 2,
            "%" : 2
        },
        "margin-right" : {
            "px" : 2,
            "em" : 2,
            "%" : 2
        },
        "margin-left" : {
            "px" : 2,
            "em" : 2,
            "%" : 2
        },
        "margin-top" : {
            "px" : 2,
            "em" : 2,
            "%" : 2
        },
        "margin-bottom" : {
            "px" : 2,
            "em" : 2,
            "%" : 2
        },
        "max-height" : {
            "px" : 2,
            "em" : 2,
            "%" : 2
        },
        "max-width" : {
            "px" : 2,
            "em" : 2,
            "%" : 2
        },
        "min-height" : {
            "px" : 2,
            "em" : 2,
            "%" : 2
        },
        "min-width" : {
            "px" : 2,
            "em" : 2,
            "%" : 2
        },
        "outline" : 1,
        "outline-color" : 1,
        "outline-style" : 1,
        "outline-width" : 1,
        "overflow" : {
            "hidden" : 2,
            "visible" : 2,
            "auto" : 2,
            "scroll" : 2
        },
        "overflow-x" : {
            "hidden" : 2,
            "visible" : 2,
            "auto" : 2,
            "scroll" : 2
        },
        "overflow-y" : {
            "hidden" : 2,
            "visible" : 2,
            "auto" : 2,
            "scroll" : 2
        },
        "padding" : {
            "px" : 2,
            "em" : 2,
            "%" : 2
        },
        "padding-top" : {
            "px" : 2,
            "em" : 2,
            "%" : 2
        },
        "padding-right" : {
            "px" : 2,
            "em" : 2,
            "%" : 2
        },
        "padding-bottom" : {
            "px" : 2,
            "em" : 2,
            "%" : 2
        },
        "padding-left" : {
            "px" : 2,
            "em" : 2,
            "%" : 2
        },
        "page-break-after" : {
            "auto" : 2,
            "always" : 2,
            "avoid" : 2,
            "left" : 2,
            "right" : 2
        },
        "page-break-before" : {
            "auto" : 2,
            "always" : 2,
            "avoid" : 2,
            "left" : 2,
            "right" : 2
        },
        "page-break-inside" : 1,
        "position" : {
            "absolute" : 2,
            "relative" : 2,
            "fixed" : 2,
            "static" : 2
        },
        "right" : {
            "px" : 2,
            "em" : 2,
            "%" : 2
        },
        "table-layout" : {
            "fixed" : 2,
            "auto" : 2
        },
        "text-decoration" : {
            "none" : 2,
            "underline" : 2,
            "line-through" : 2,
            "blink" : 2
        },
        "text-align" : {
            "left" : 2,
            "right" : 2,
            "center" : 2,
            "justify" : 2
        },
        "text-indent" : 1,
        "text-transform" : {
            "capitalize" : 2,
            "uppercase" : 2,
            "lowercase" : 2,
            "none" : 2
        },
        "top" : {
            "px" : 2,
            "em" : 2,
            "%" : 2
        },
        "vertical-align" : {
            "top" : 2,
            "bottom" : 2
        },
        "visibility" : {
            "hidden" : 2,
            "visible" : 2
        },
        "white-space" : {
            "nowrap" : 2,
            "normal" : 2,
            "pre" : 2,
            "pre-line" : 2,
            "pre-wrap" : 2
        },
        "width" : {
            "px" : 2,
            "em" : 2,
            "%" : 2
        },
        "word-spacing" : {
            "normal" : 2
        },
        "z-index" : 1,

        // opacity
        "opacity" : 1,
        "filter" : {
            "alpha(opacity=|100)" : 2
        },

        "text-shadow" : {
            "|2px 2px 2px #777" : 2
        },
        "text-overflow" : {
            "ellipsis-word" : 2,
            "clip" : 2,
            "ellipsis" : 2
        },

        // border radius
        "border-radius" : 1,
        "-moz-border-radius" : 1,
        "-moz-border-radius-topright" : 1,
        "-moz-border-radius-bottomright" : 1,
        "-moz-border-radius-topleft" : 1,
        "-moz-border-radius-bottomleft" : 1,
        "-webkit-border-radius" : 1,
        "-webkit-border-top-right-radius" : 1,
        "-webkit-border-top-left-radius" : 1,
        "-webkit-border-bottom-right-radius" : 1,
        "-webkit-border-bottom-left-radius" : 1,

        // dropshadows
        "-moz-box-shadow" : 1,
        "-webkit-box-shadow" : 1,

        // transformations
        "transform" : {
            "rotate(|0deg)" : 2,
            "skew(|0deg)" : 2
        },
        "-moz-transform" : {
            "rotate(|0deg)" : 2,
            "skew(|0deg)" : 2
        },
        "-webkit-transform" : {
            "rotate(|0deg)" : 2,
            "skew(|0deg)" : 2
        }
    },

    getCompletions: function(token, context)
    {
        var found = [], start = token.string, props;
        function maybeAdd(key, val)
        {
            if (key.indexOf(start) == 0)
                found.push(key);
        }

        //if (token.className == "variable") // deprecated in codemirror 4.2.1
        jQuery.each(this.cssDictionary, maybeAdd);

        return found;
    }
}

AXES.Dev.ListPropertyEditor =
{
    editList: function( property, title, readOnly )
    {
        var $tr = $(
            '<tr class="ui-widget-header" id="labels">\
                <th></th>\
            </tr>'
        );

        var def        = property.definition;
        var labelAdded = false;

        function addLabels()
        {
            for ( var propName in def )
                $tr.append( '<th>' + propName + '</th>' );
            labelAdded = true;
        }

        var propVal   = AXES.Extensions._deserialisePropertyValue( property._owner, property._name, property.getJSONObject( false ) );
        var propValue = propVal.v;
        if ( propValue.length )
            addLabels();

        var rowCount   = 0;
        var $dialogDiv = $( "#ax-extensions-editListDialog" );
        var $tbody     = $dialogDiv.find( "tbody" ).empty();
        var $thead     = $dialogDiv.find( "thead" ).empty();
        $thead.append( $tr[ 0 ] );

        function appendRow( item )
        {
            var tr       = document.createElement( "tr" );
            tr.styleData = item;
            var td       = document.createElement( "td" );
            var itemProp = item.properties;
            td.innerHTML = '<span class="ui-icon ui-icon-carat-2-n-s"></span>' +
                            (
                                readOnly ?
                                '' : 
                                '<span class="ui-icon ui-icon-circle-close" style="background-image:url( ./css/smoothness/images/ui-icons_cd0a0a_256x240.png )"></span>'
                            );
            tr.appendChild( td );
            for ( var propName in def )
            {
                td = document.createElement( "td" );
                var propEditor = itemProp[ propName ].renderEditor( readOnly );
                if ( propEditor )
                    td.appendChild( propEditor );
                tr.appendChild( td );
            }
            $tbody.append( tr );
            rowCount++;
        };

        function addNote()
        {
            $( ".ui-dialog-buttonpane:first" )
                .append( "<div id='note' style='height:10px; margin-top:20px'>" + TEXT.Dev( "To change the order of an item, hold-click on it, then drag it around." ) + "</div>" );
        }

        for ( var i = 0, item; item = propValue[ i ]; i++ )
            appendRow( item );

        $tbody
            .sortable(
                {
                    axis                : "y",
                    forcePlaceholderSize: true,
                    handle              : ".ui-icon-carat-2-n-s",
                    placeholder         : "sortable-placeholder",
                    revert              : true,

                    helper: function( e, tr )
                    {
                        var $originals = tr.children();
                        var $helper = tr.clone();
                        $helper.addClass( "sortable-helper" ).children().each(
                            function( index )
                            {
                                $( this ).css( "width", $originals.eq( index ).innerWidth() + "px" );
                            }
                        );
                        return $helper;
                    },
                    start: function( e, ui )
                    {
                        ui.placeholder.height( ui.item.height() );
                    }
                }
            )
            .on(
                "click",
                ".ui-icon-circle-close",
                function( event )
                {
                    var tr = $( this ).closest( "tr" );
                    tr.remove();
                    rowCount--;
                    if ( !rowCount )
                        $( "#labels" ).remove();
                    else if ( 1 == rowCount )
                        $( ".ui-dialog-buttonpane:first" ).find( "#note" ).remove();
                }
            );

        $dialogDiv
            .find( ".ax-extensions-addListItem" )
            .button(
                {
                    icons: { primary: "ui-icon-plus" },
                    text : false
                }
            )
            .click(
                function( event )
                {
                    if ( !rowCount && !labelAdded )
                        addLabels();
                    else if ( 1 == rowCount )
                        addNote();
                    $thead.append( $tr[ 0 ] );
                    appendRow(
                        AXES.Extensions._deserialisePropertyValue(
                            property._owner,
                            property._name,
                            { clsid : property.type }
                        )
                    );
                }
            );

        if ( readOnly )
            $dialogDiv.find( ".ax-extensions-addListItem" ).hide();
        else
            $dialogDiv.find( ".ax-extensions-addListItem" ).show();

        var dialogOptions =
        {
            bgiframe : true,
            buttons  : {},
            height   : 400,
            minHeight: 200,
            minWidth : 200,
            modal    : true,
            position : "top",
            title    : title,
            width    : 700,

            close: function()
            {
                $( this ).find( ".ax-extensions-addListItem" ).button( "destroy" ).off( "click" );
                $tbody.sortable( "destroy" ).off( "click" );
                $( this ).off( "keydown keypress keyup" ).dialog( 'destroy' );
                $thead.empty();
                $tbody.empty();
            }
        }

        if ( !readOnly )
            dialogOptions.buttons[ TEXT.Core( "OK" ) ] = function()
            {
                var rows  = $tbody.find( "tr" );
                var vals  = propVal.v = [];
                var index = 0;

                if ( !rows.length || !rows[ 0 ].styleData.properties.name )
                {
                    for ( var row, $nameElement; row = rows[ index ]; index++ )
                        vals.push( row.styleData );
                    $( this ).dialog( "close" );
                    AXES.Dev.ExtensionsTab.updatePropertyValue( property, propVal );

                    return;
                }

                var name;
                var errorType  = 0;
                var styleNames = {};

                function isValidName( name )
                {
                    var test = '\\\/:*?"<>|';
                    var len = name.length;
                    var len2 = test.length;
                    for ( var i = 0; i < len; i++ )
                        for ( var j = 0; j < len2; j++ )
                            if ( name.charAt( i ) == test.charAt( j ) )
                            {
                                return false;
                            }

                    return true;
                }

                for ( var row, $nameElement; row = rows[ index ]; index++ )
                {
                    $nameElement = $( $( row ).children().eq( 1 ) ).find( "input" );

                    name = $nameElement.val().trim();
                    if ( "" == name )
                    {
                        errorType = 1;
                        break;
                    }

                    if ( !isValidName( name ) )
                    {
                        errorType = 2;
                        break;
                    }

                    if ( styleNames[ name ] )
                    {
                        errorType = 3;
                        break;
                    }

                    styleNames[ name ] = {};

                    $nameElement.val( name );
                    row.styleData.properties.name.v = name;
                    vals.push( row.styleData );
                }
                delete styleNames;

                switch ( errorType )
                {
                case 1:
                    alert( TEXT.dev( "axAppStyles.err1", index + 1 ) );
                    break;

                case 2:
                    alert( TEXT.dev( "axAppStyles.err2", name ) );
                    break;

                case 3:
                    alert( TEXT.dev( "axAppStyles.err3", name ) );
                    break;

                default:
                    $( this ).dialog( "close" );
                    AXES.Dev.ExtensionsTab.updatePropertyValue( property, propVal );
                }
            };

        dialogOptions.buttons[ TEXT.Core( "Cancel" ) ] = function()
        {
            $( this ).dialog( "close" );
        };

        $dialogDiv
            .dialog( dialogOptions )
            .on(
                "keydown keypress keyup",
                function( e )
                {
                    e.stopPropagation();
                }
            );

        if ( 1 < rowCount )
            addNote();
    }
}

/*
 * Add some developer mode functions to existing namespaces
 */

/**
 * Renders an edit/view script button for the property.
 * <strong>Note:</strong> This method is only available in developer mode.
 * @param {Boolean} setDisabled
 * @returns {Object} Returns a DOM element to be inserted in the display
 */
AXES.ExtensionProperty.prototype.renderScriptEditor = function( setDisabled )
{
    var btnText      = setDisabled ? TEXT.Dev( "View Script..." ) : TEXT.Dev( "Edit Script..." );
    var retVal       = document.createElement( "button" );
    retVal.className = "ax-extension-editScriptButton";
    retVal.appendChild( document.createTextNode( btnText ) );
    $( retVal ).bind(
        "click",
        {
            property  : this,
            title     : btnText,
            isDisabled: setDisabled
        },
        this.editScript
    );
    this._UIElement = null;
    return retVal;
}

AXES.ExtensionProperty.prototype.renderListEditor = function( setDisabled )
{
    var btnText      = setDisabled ? TEXT.Dev( "View Items..." ) : TEXT.Dev( "Edit Items..." );
    var retVal       = document.createElement( "button" );
    retVal.className = "ax-extension-editListButton";
    retVal.appendChild( document.createTextNode( btnText ) );
    $( retVal ).bind(
        "click",
        {
            property  : this,
            title     : btnText,
            isDisabled: setDisabled
        },
        this.editList
    );
    this._UIElement = null;

    return retVal;
}

AXES.ExtensionProperty.prototype.renderListEditor2 = function(setDisabled)
{
    if ( !this.v )
        this.v = [];

    var property     = this;
    var propVal      = property.v;
    var vals         = {};
    var btnText      = setDisabled ? TEXT.Dev( "View Items..." ) : TEXT.Dev( "Edit Items..." );
    var retVal       = document.createElement( "button" );
    retVal.className = "ax-extension-editListButton";
    retVal.appendChild( document.createTextNode( btnText ) );
    var $dlg, $tbody, $thead;

    $( retVal ).click(
        function()
        {            
            var dlgOptions = {
                    width: 500, 
                    height: 300, 
                    minHeight: 200,                    
                    modal: true,
                    title: btnText,
                    buttons: {},
                    close: function( e )
                    {
                        $( this ).find( ".ax-extensions-addListItem" ).button( "destroy" ).off( "click" );
                        $( this ).off( "keydown keypress keyup" ).dialog( 'destroy' );
                        $thead.empty();
                        $tbody.empty();
                    }
            };
            
            if (!setDisabled)
            {
                dlgOptions.buttons[ TEXT.Core( "OK" ) ] = function()
                {
                    propVal = property.v = [];
                    var rows = $tbody.find( "tr" );
                    for ( var i = 0, item; item = rows[ i ]; i++ )
                        propVal.push( $( item ).find( "select" ).val() );
                    $( this ).dialog( "close" );         
                };
            }
                
            dlgOptions.buttons[ TEXT.Core( "Cancel" ) ] = function()
            {
                $( this ).dialog( "close" );
            };

            $dlg = $('<div id="ax-extensions-editListDialog2">\
                    <table>\
                        <thead></thead>\
                        <tbody class="ax-extensions-listItems"></tbody>\
                    </table>\
                    <button type="button" class="ax-extensions-addListItem" ax_txtscope="Dev" ax_txt="Add Item">Add Item</button>\
                </div>'
            ).dialog(dlgOptions);
   
            $tbody = $dlg.find( "tbody" ).empty()
            $thead = $dlg.find( "thead" ).empty()
            $thead.append( document.createElement( "tr" ) );
            
            if (setDisabled) {
                $dlg.find( ".ax-extensions-addListItem" ).hide();
            } else {
                $dlg.find( ".ax-extensions-addListItem" ).show();
            }                
            
            function append5250Row( val )
            {
                var tr = document.createElement( "tr" );
                if ( !$tbody[ 0 ].childElementCount )
                    tr.style.borderTop = "1px solid gray";
                var td = document.createElement( "td" );
                td.innerHTML = (setDisabled ? '<span class="ui-icon ui-icon-carat-2-n-s"></span>&nbsp;&nbsp;&nbsp;&nbsp;' : 
                            '<span class="ui-icon ui-icon-carat-2-n-s"></span><span class="ui-icon ui-icon-circle-close" style="background-image:url( ./css/smoothness/images/ui-icons_cd0a0a_256x240.png )"></span>');
                tr.appendChild( td );

                var el = document.createElement( "select" );
                el.style.width = "90%";
                var items = property.items;
                for ( var i = 0, item; item = items[ i ]; i++ )
                {
                    var opt = document.createElement( "option" );
                    opt.setAttribute( "value", item.value );
                    if ( "" != val && val == item.value )
                        opt.setAttribute( "selected", "selected" );
                    opt.appendChild( document.createTextNode( item.text ) );
                    el.appendChild( opt );
                }
                $(el).attr( "disabled", setDisabled ? "disabled" : false );
                td.appendChild( el );
                tr.appendChild( td );
                $tbody.append( tr );
            }

            $tbody
                .sortable()
                .on(
                    "click",
                    ".ui-icon-circle-close",
                    function( event )
                    {
                        var tr = $( this ).closest( "tr" );
                        tr.remove();
                    }
                );

            if ( property.v && property.v.length )
                for ( var i in propVal )
                    append5250Row( propVal[ i ] );

            $dlg.find( ".ax-extensions-addListItem" )
                .button( { icons: { primary : "ui-icon-plus" }, text: false } )
                .click(
                    function()
                    {
                        append5250Row( "" );
                    }
                );

            $dlg.on(
                "keydown keypress keyup",
                function( e )
                {
                    e.stopPropagation();
                }
            );
        }
    );
    this._UIElement = null;
    return retVal;
}

/**
 * Returns the property as a simple object suitable for serialization
 */
AXES.ExtensionProperty.prototype.getJSONObject = function( forSave )
{
    // Only save the value if it is not the default value
    // TODO: this is a bit crude. A proper object comparison should be faster and might be a useful utility
    if ( forSave && this.v === undefined && JSON.stringify( this.getRawDefaultValue() ) == JSON.stringify( this.getRawValue() ) )
    {
        return null;
    }
    if ( this.isList )
    {
        var retVal = { v : [] };
        if ( this.v )
        {
            var v = this.v;
            if ( this.definition )
                for ( var i = 0, item, jsonObj; item = v[ i ]; i++ )
                {
                    jsonObj = item.getJSONObject ? item.getJSONObject( false ) : item;
                    if ( jsonObj )
                        retVal.v.push( jsonObj );
                }
            else
                retVal.v = v;
        }
        return retVal;
    }
    if ( this.definition )
    {
        var retVal = { clsid : this.type };
        var jsonObj;
        var prop = this.properties;
        for ( var p in prop )
        {
            jsonObj = prop[ p ].getJSONObject( false );
            if ( jsonObj )
                retVal[ p ] = jsonObj;
        }
        return retVal;
    }
    return this.getRawValue();
}

/**
 * A jQuery event handler that opens the script editor. Bind this to an event to trigger the editor.
 * You should set a data object in the bind method that contains:
 *  title: A string to use for the title of the dialog
 *  ex: The extension the property belongs to
 *  isDisabled: A boolean indicating if editing is enabled/disabled
 * <strong>Note:</strong> This method is only available in developer mode.
 * @param {Object} e jQuery event object
 */
AXES.ExtensionProperty.prototype.editScript = function( e ) 
{
    var d = e.data;
    d.property._openCodeMirror( d.property, d.title, "javascript", d.isDisabled );
}

/**
 * A jQuery event handler that opens the style editor. Bind this to an event to trigger the editor.
 * You should set a data object in the bind method that contains:
 *  title: A string to use for the title of the dialog
 *  ex: The extension the property belongs to
 *  isDisabled: A boolean indicating if editing is enabled/disabled
 * <strong>Note:</strong> This method is only available in developer mode.
 * @param {Object} e jQuery event object
 */
AXES.ExtensionProperty.prototype.editStyle = function( e )
{
    var d = e.data;
    d.property._openCodeMirror( d.property, d.title, "cssattr", d.isDisabled );
}

AXES.ExtensionProperty.prototype.editStyle2 = function( e )
{
    var d             = e.data;
    var isDisabled    = d.isDisabled;
    var property      = d.property;
    var dev           = AXES.Dev;
    var extensionsTab = dev.ExtensionsTab;
    var locked        = dev.screenLocked;
    if ( locked )
        dev.screenLocked = false;

    var dlgButtons = [
        {
            text : TEXT.Dev( "OK" ),
            click: function()
            {
                if ( isDisabled )
                {
                    $( this ).dialog( "close" );
                    return;
                }

                if ( $row )
                    updateData();
                if ( $baseClassList.val() )
                    propertyVal.v[ "_base" ] = $baseClassList.val();
                else
                    delete propertyVal.v[ "_base" ];

                    var fld = property._owner ? property._owner._field : null; //get the field
                    if (fld && fld.recordFormat && fld.recordFormat.cols) {
                        var tmp = property._displayRowId.split("_");
                        var extIndex = tmp[4]; //get the extension index
                        if ( 0 == extIndex ) {
                            //if its a basic property apply to the selected field in a subfile
                            //not all, see how it works on TS developer as reference.
                            AXES.Dev.ExtensionsTab.updatePropertyValue( property, propertyVal );
                        } else {
                            // its an extension property then apply to all field in a subfile
                            var col = fld.recordFormat.getSubfileColumn(fld);
                            var colKey = Object.keys(col)[0];
                            var sbColumn = col[colKey];
                            for (var cfld in sbColumn.items) {
                                if (sbColumn.items[cfld]) {
                                    property = sbColumn.items[cfld].extensions[extIndex].getPropertyByName(property._name);
                                    AXES.Dev.ExtensionsTab.updatePropertyValue( property, propertyVal );
                                }
                            }
                        }
                    } else {
                        AXES.Dev.ExtensionsTab.updatePropertyValue( property, propertyVal );
                    }

                $( this ).dialog( "close" );
            }
        }
    ];
    if ( !isDisabled )
        dlgButtons.push(
            {
                text: TEXT.Dev( "Cancel" ),
                click: function()
                {
                    $( this ).dialog( "close" );
                }
            }
        );

    var $dlg = $(
        '<div id="style-dialog"> \
            <div id="baseClassContainer" style="background:rgb( 245, 245, 245 ); border:solid 1px black; display:block;"> \
                <div style="background:black; color:white; padding:8px; font-weight:bold;" ax_txtscope="dev" ax_txt="designer.styleEditor.baseStyle">Base Style</div> \
                <div style="padding:7px" ax_txtscope="dev" ax_txt="designer.styleEditor.baseStyleInstruction">Select a style that acts as a base for this style:</div>\
                <div id="baseClass" style="padding:1px 7px 9px;"> \
                    <select id="baseClassList" disabled:readOnly style="width:100%"> \
                        <option value="" ax_txtscope="dev" ax_txt="designer.styleEditor.noBaseStyle">[No base style]</option> \
                    </select> \
                </div> \
            </div> \
            <div id="vspace1" style="height:5px; display:block;"></div> \
            <div id="attrListContainer" style="border:solid 1px black; height:310px;"> \
                <div style="background:black; color:white; padding:8px; font-weight:bold;" ax_txtscope="dev" ax_txt="designer.styleEditor.attributes">Attributes</div> \
                <div id="buttonPane" style="background-color:rgb( 210, 210, 210 ); padding:10px 5px; border-bottom: solid 1px rgb( 100, 100, 100 );"> \
                    <input id="displayOnly" type="checkbox"/> \
                    <label style="padding:5px 5px 5px 4px;" ax_txtscope="dev" ax_txt="designer.styleEditor.displayUsedOnly">Display Used Attributes Only</label> \
                </div> \
                <div id="attrList" style="overflow-x:none; overflow-y:scroll; height:240px"> \
                    <table id="styleEditor" cellpadding="0" cellspacing="0" border="0"> \
                        <tbody id="styleEditorBody"> \
                        </tbody> \
                    </table> \
                </div> \
            </div> \
        </div>'
    ).dialog(
        {
            buttons: dlgButtons,
            height : 500,
            hide   : 'clip',
            modal  : true,
            show   : 'clip',
            title  : TEXT.Dev( "Edit Styles..." ),
            width  : 500,

            close: function( e )
            {
                $( this ).empty();
                $( this ).dialog( 'destroy' );
                dev.screenLocked = locked;
            },
            resize: function( e )
            {
                $( "#attrListContainer" ).height( $dlg.height() - hDif );
                $( "#attrList" ).height( $( "#attrListContainer" ).height() - attrHDif );
            }
        }
    );

    var hDif     = $dlg.height() - $( "#attrListContainer" ).height();
    var attrHDif = $( "#attrListContainer" ).height() - $( "#attrList" ).height();

    var $baseClassList = $( "#baseClassList" );
    var appStyles      = $( "#ax-extensions-editListDialog" ).find( "tbody" ).children();
    if ( appStyles.length )
    {
        var appStyleName = $( e.srcElement ).closest( "tr" ).children().eq( 1 ).find( "input" ).val();
        for ( var i = 0, style, name; style = appStyles[ i ]; i++ )
        {
            name = $( style ).children().eq( 1 ).find( "input" ).val();
            if ( "" == name || appStyleName == name )
                continue;
            $baseClassList.append(
                '<option value="' + name + '" ax_txtscope="dev">' + name + '</option>'
            );
        }
    }
    else
    {
        appStyles = extensionsTab.getFieldRowFromRowId( "ax-fldRow-application" ).data( "ExtendableObject" ).getExtensionProperty( "styles" ).v;
        for ( var i = 0, style, name; style = appStyles[ i ]; i++ )
        {
            name = style.properties ? style.properties.name.v : style.name.v;
            if ( "" == name )
                continue;
            $baseClassList.append(
                '<option value="' + name + '" ax_txtscope="dev">' + name + '</option>'
            );
        }
    }

    function toCcsString( str )
    {
        for( var i = 0, ch; ch = str.charAt( i ); i++ )
            if ( ch >= 'A' && ch <= 'Z' )
                str = str.replace( ch, "-" + ch.toLowerCase() );

        return str;
    }

    var styles = $( "#style-dialog" )[ 0 ].style;
    var cssProperties = [];
    for ( var i in styles )
        if ( typeof styles[ i ] == "string" && isNaN( i ) )
            cssProperties.push( toCcsString( i ) );

    var styleValues = property.v || property.defaultStatic;
    for ( var i in styleValues )
    {
        if ( cssProperties[ i ] )
            continue;
        cssProperties.push( i );
    }
    cssProperties.sort();

    var styleProperties = {};
    for ( var i = 0, style, val; style = cssProperties[ i ]; i++ )
        styleProperties[ style ] = "";

    var origValues = {};
    for ( var i in styleValues )
        origValues[ i ] = styleProperties[ i ] = styleValues[ i ];

    if ( styleValues && styleValues._base )
    {
        $baseClassList.val( styleValues._base );
        delete styleProperties._base;
    }

    var $styleEditorBody = $( "#styleEditorBody" );
    for ( var i in styleProperties )
        $styleEditorBody.append(
            '<tr> \
                <td style="white-space:nowrap; background-color:rgb( 240, 240, 240 ); border-right:solid 1px rgb( 200, 200, 200 ); border-bottom: solid 1px rgb( 200, 200, 200 ); padding:5px 8px 5px 5px; width:100px; overflow:hidden;">' + i + '</td> \
                <td style="white-space:nowrap; width:100%; overflow:hidden; border-bottom: solid 1px rgb( 200, 200, 200 ); border-right:solid 1px rgb( 200, 200, 200 );"> \
                    <label style="padding:5px;">' + styleProperties[ i ] + '</label> \
                </td> \
            </tr>'
        );

    var $row        = null;
    var $label      = null;
    var styleName   = "";
    var $input      = $( '<input style="border: none; width: 95%; padding-left: 5px"/>' );
    var propertyVal = { v: origValues };

    function updateData()
    {
        $input.remove();

        var value = $input.val();
        if ( "" == value )
            delete propertyVal.v[ styleName ];
        else
            propertyVal.v[ styleName ] = value;

        $label.css( "display", "" );
        $label.text( value );
    }

    function updateElements()
    {
        styleName = $row.children()[ 0 ].innerText;
        var $val  = $( $row.children()[ 1 ] );
        $label    = $val.find( "label" );
        $label.css( "display", "none" );

        $input.val( $label.text() );
        $val.append( $input );
        $input.focus();
    }

    function firstInputClick( e )
    {
        $row = $( e.srcElement ).closest( "tr" );
        if ( !$row.length )
        {
            return;
        }

        updateElements();
        $styleEditorBody.unbind( "click", firstInputClick );
        $styleEditorBody.bind( "click", inputClick );
    }

    function inputClick( e )
    {
        var $row_ = $( e.srcElement ).closest( "tr" );
        if ( $row[ 0 ] == $row_[ 0 ] )
        {
            return;
        }

        updateData();

        if ( !$row_.length )
        {
            return;
        }

        $row = $row_;
        updateElements();
    }

    if ( isDisabled )
        $baseClassList.prop( "disabled", true );
    else
    {
        $input.blur(
            function ()
            {
                updateData();
                $row = null;
                $styleEditorBody.unbind( "click", inputClick );
                $styleEditorBody.bind( "click", firstInputClick );
            }
        );
        $styleEditorBody.bind( "click", firstInputClick );
    }

    $( "#displayOnly" ).click(
        function ( e )
        {
            var $children = $styleEditorBody.children();
            if ( e.srcElement.checked )
            {
                for ( var i = 0, $row; ( $row = $( $children[ i ] ) ) && $row[ 0 ]; i++ )
                    if ( $row.find( "label" ).text() == "" )
                        $row.css( "display", "none" )
            }
            else
                for ( var i = 0, $row; ( $row = $( $children[ i ] ) ) && $row[ 0 ]; i++ )
                    $row.css( "display", "" )
        }
    );
}

AXES.ExtensionProperty.prototype.editList = function( e )
{
    var d = e.data;
    AXES.Dev.ListPropertyEditor.editList( d.property, d.title, d.isDisabled );
}

AXES.ExtensionProperty.prototype._openCodeMirror = function( property, title, editorMode, readOnly )
{
    var propertyValue = property.getRawValue();
    var editorOptions =
    {
        gutters    : [ "CodeMirror-lint-markers" ],
        lineNumbers: true,
        lint       : editorMode == "javascript" ? true: false,
        mode       : editorMode,
        readOnly   : readOnly,
        theme      : "default",
        value      : propertyValue.v
    }

    var autoComplete, convertResult;
    var ENV = null;
    switch ( editorMode )
    {
    case "javascript":
        autoComplete = AXES.Dev.JSAutoComplete;
        ENV          = new AXES.ExtensionPropertyScriptEnvironment( property, {} );
        break;

    case "cssattr":
        editorOptions.mode      = "css";
        editorOptions.enterMode = "keep";
        autoComplete            = AXES.Dev.CSSAutoComplete;
        var cssVal              = editorOptions.value;
        var cssText             = "";
        for ( var p in cssVal )
            cssText += p + ": " + cssVal[ p ] + ";\n";
        editorOptions.value = "{\n" + cssText + "}";

        convertResult = function( str )
        {
            cssText = str.trim().replace( /(^\n*\{?\n*)|(\n*\}?\n*$)/g, "" ).replace( "\n", "" );
            cssVal  = {};

            var val, pos;
            for ( var startPos = 0, endPos; -1 != ( endPos = cssText.indexOf( ";", startPos ) ); startPos = endPos + 1 )
            {
                val = cssText.substring( startPos, endPos ).trim();
                pos = val.indexOf( ":" );
                cssVal[ val.substring( 0, pos ).trim() ] = val.substring( pos + 1 ).trim();
            }
            return cssVal;
        };
        break;

    default:
        autoComplete = null;
    }

    var dialogOptions =
    {
        bgiframe: true,
        buttons : {},
        height  : 400,
        modal   : true,
        position: "top",
        width   : 600,

        close : function()
        {
            $( this ).dialog( 'destroy' );
            if ( autoComplete )
                delete autoComplete.ENV;
        }
    }

    if ( !readOnly )
    {
        dialogOptions.buttons[ TEXT.Dev( "OK" ) ] = function()
        {
            propertyValue.v = convertResult ? convertResult( editor.getValue() ) : editor.getValue();
            $( this ).dialog( 'close' );
              
            var fld = property._owner ? property._owner._field : null; //get the field            
            if (fld && fld.recordFormat && fld.recordFormat.cols) {
                var tmp = property._displayRowId.split("_");
                var extIndex = tmp[4]; //get the extension index
                if ( 0 == extIndex ) {
                    //if its a basic property apply to the selected field in a subfile
                    //not all, see how it works on TS developer as reference.
                    AXES.Dev.ExtensionsTab.updatePropertyValue( property, propertyValue );
                } else {
                    // its an extension property then apply to all field in a subfile
                    var col = fld.recordFormat.getSubfileColumn(fld);
                    var colKey = Object.keys(col)[0];
                    var sbColumn = col[colKey];                        
                    for (var cfld in sbColumn.items) {
                        if (sbColumn.items[cfld]) {
                            property = sbColumn.items[cfld].extensions[extIndex].getPropertyByName(property._name);
                            AXES.Dev.ExtensionsTab.updatePropertyValue( property, propertyValue );
                        }
                    }
                }
            } else {
                AXES.Dev.ExtensionsTab.updatePropertyValue( property, propertyValue );
            }
        };
    }

    dialogOptions.buttons[ readOnly ? TEXT.Dev( "OK" ) : TEXT.Dev( "Cancel" ) ] = function()
    {
        $( this ).dialog( 'close' );
    };

    var div = document.createElement( "div" );
    $( div )
        .attr( "title", title )
        .bind(
            "keydown keypress keyup",
            function( e )
            {
                e.stopPropagation();
            }
        )
        .dialog( dialogOptions );

    editor = CodeMirror( div, editorOptions );
    editor.on(
        "keypress",
        function( i, e )
        {
            // Hook into ctrl-space
            if ( e.keyCode != 32 || !e.ctrlKey || e.altKey )
            {
                return;
            }
            e.preventDefault();
            autoComplete.editor = i;
            autoComplete.ENV = ENV;
            return autoComplete.startComplete();
        }
    );

    $( div ).append(
        $( '<span style="color:gray;font-size:smaller;"></span>' ).text( TEXT.Dev( "Press Ctrl-Space for autocomplete." ) )
    );
}

AXES.ExtensionProperty.prototype.renderEditor = function( setDisabled )
{
    var rawVal = this.getRawValue();
    if ( rawVal.s )
    {
        return this.renderScriptEditor( setDisabled );
    }
    if ( this.isList )
    {
        if ( this.items )
        {
            return this.renderListEditor2( setDisabled );
        }
        return this.renderListEditor( setDisabled );
    }
    if ( this.items )
    {
        var retVal = this._UIElement = document.createElement( "select" );
        retVal.style.width = "100%";
        var property = this;
        rawVal = this.getRawValue();
        var items = this.items;
        for ( var i = 0, item; item = items[ i ]; i++ )
        {
            var opt = document.createElement( "option" );
            opt.setAttribute( "value", item.value );
            if ( rawVal.v == item.value )
                opt.setAttribute( "selected", "selected" );
            opt.appendChild( document.createTextNode( item.text ) );
            retVal.appendChild( opt );
        }
        retVal.disabled = setDisabled;
        $( retVal ).attr( "disabled", setDisabled ? "disabled" : false ).change(
            function( e )
            {
                AXES.Dev.ExtensionsTab.updatePropertyValue( property, { v : this.options[ this.selectedIndex ].value } );
            }
        );
        return retVal;
    }
    return this._renderEditor( setDisabled )
}

AXES.ExtensionProperty.prototype._renderEditor = function(setDisabled) 
{
    return (this._UIElement = document.createTextNode(this.getRawValue().v));
}

AXES.ExtensionProperty.prototype.refreshEditorDisplay = function() 
{
    if (this.items && this._UIElement) 
    {
        var newVal = this.getRawValue(), i, item, l = this.items.length;
        for (i = 0; i < l; i++) 
        {
            item = this.items[i];
            if (newVal.v == item.value) 
            {
                this._UIElement.selectedIndex = i;
                return;
            }
            this._UIElement.selectedIndex = -1;
        }
    }
    else 
    {
        this._refreshEditorDisplay();
    }
}
AXES.ExtensionProperty.prototype._refreshEditorDisplay = function() 
{
    if (this._UIElement)
        $(this._UIElement).text(this.getRawValue().v);
}
/*
* Add some screen name indexing functions to the AXES.Screens namespace
*/
/**
 * Create an index of screen names.
 * <strong>Note:</strong> This method is only available in developer mode.
 */
AXES.Screens.indexScreenNames = function() {
    // Only if it doesn't already exist.
    if (!AXES.Screens.namesHash) {
        var nh = AXES.Screens.namesHash = {};
        var keysHash = AXES.Screens.keysHash;

        for (var key in keysHash) {
            var kh = keysHash[key];
            for (var id in kh) {
                var def = kh[id];
                this.addToScreenNameIndex(def.name, def.variantName);
            }
        }
    }
}
/**
 * Determine if a screen and variant exists.
 * <strong>Note:</strong> This method is only available in developer mode.
 * @param {String} name
 * @param {String} variant
 * @returns {Boolean}
 */
AXES.Screens.isScreenNameInIndex = function(name, variant) {
    if (this.namesHash) {
        var key = name + "#variant#" + variant;
        if (this.namesHash[key])
        {
            return true;
        }
    }
    return false;
}
/**
 * Removes a name and variant from the Screen Name index.
 * <strong>Note:</strong> This method is only available in developer mode.
 * @param {String} name
 * @param {String} variant
 */
AXES.Screens.removeFromScreenNameIndex = function(name, variant) {
    if (this.namesHash) {
        var key = name + "#variant#" + variant;
        this.namesHash[key] = false;
    }
}
/**
 * Add a Screen name and variant to the Screen Name index.
 * <strong>Note:</strong> This method is only available in developer mode.
 * @param {String} name
 * @param {String} variant
 */
AXES.Screens.addToScreenNameIndex = function(name, variant) {
    if (this.namesHash) {
        var key = name + "#variant#" + variant;
        this.namesHash[key] = true;
    }
}
/**
 * Delete a screen definition.
 * <strong>Note:</strong> This method is only available in developer mode.
 * @param {Object} definition
 */
AXES.Screens.deleteScreenDefinition = function(definition) {
    var kh = this.keysHash;
    var id = definition.id;
    var key = definition.key;
    if (kh[key] && kh[key][id])
        delete kh[key][id];
}
/**
 * Sets and/or returns the layout design mode.
 * <strong>Note:</strong> This method is only available in developer mode.
 * @param {Boolean} [mode] Set to true or false to set the design mode. Leave empty to get the current mode.
 * @returns {Boolean} The design mode
 */
AXES.Layout.designMode = function(mode) {
    if ( typeof mode == "boolean") {
        this._designMode = mode;
        if (mode) {
            AXES.Layout.Cursor.hide();
        } else {
            AXES.Layout.Cursor.show();
        }
    }
    return this._designMode;
}
AXES.Layout._designMode = false;

/**
 * Replace the cached screen definition for the current screen with the supplied one.
 * <strong>Note:</strong> This method is only available in developer mode.
 * @param {Object} newDef The new definition
 */
AXES.Extensions.replaceCurrentScreenDefinition = function(newDef)
{
    var screenName = AXES.Terminal.symbolicName;
    if (screenName) 
    {
        this._screenDefinitions[this._getScreenKey(screenName)] = newDef;
    }
}
/**
 * Delete the cached definition for the current screen. This will force the definition to be
 * reloaded next time the screen is processed.
 * <strong>Note:</strong> This method is only available in developer mode.
 */
AXES.Extensions.deleteCurrentScreenDefinition = function()
{
    var screenName = AXES.Terminal.symbolicName;
    if (screenName) 
    {
        delete this._screenDefinitions[this._getScreenKey(screenName)];
    }
}

